@@ -20,7 +20,8 @@
 ;;; The CLASSOID structure is a supertype of all classoid types.
 ;;; Its definition occurs in 'early-classoid.lisp'
 
-(defun classoid-make-load-form-fun (self)
+(defmethod make-load-form ((self classoid) &optional env)
+  (declare (ignore env))
   (let ((name (classoid-name self)))
     (if (and name (eq (find-classoid name nil) self))
         `(find-classoid ',name)
@@ -57,7 +58,7 @@
 (declaim (inline layout-for-std-class-p))
 (defun layout-for-std-class-p (x) (not (zerop (layout-%for-std-class-b x))))
 
-(def!method print-object ((layout layout) stream)
+(defmethod print-object ((layout layout) stream)
   (print-unreadable-object (layout stream :type t :identity t)
     (format stream
             "for ~S~@[, INVALID=~S~]"
@@ -116,27 +117,25 @@
                   (make-layout :classoid (or classoid (make-undefined-classoid name)))))))))
 
 ;;; If LAYOUT is uninitialized, initialize it with CLASSOID, LENGTH,
-;;; INHERITS, DEPTHOID, and RAW-SLOT-METADATA. Otherwise require
-;;; that it be consistent with the existing values.
+;;; INHERITS, DEPTHOID, and BITMAP.
+;;; Otherwise require that it be consistent with the existing values.
 ;;;
 ;;; UNDEFINED-CLASS values are interpreted specially as "we don't know
 ;;; anything about the class", so if LAYOUT is initialized, any
 ;;; preexisting class slot value is OK, and if it's not initialized,
 ;;; its class slot value is set to an UNDEFINED-CLASS. -- FIXME: This
 ;;; is no longer true, :UNINITIALIZED used instead.
-(declaim (ftype (function (layout classoid index simple-vector layout-depthoid
-                                  layout-raw-slot-metadata-type)
+(declaim (ftype (function (layout classoid index simple-vector layout-depthoid layout-bitmap)
                           layout)
                 %init-or-check-layout))
-(defun %init-or-check-layout
-    (layout classoid length inherits depthoid raw-slot-metadata)
+(defun %init-or-check-layout (layout classoid length inherits depthoid bitmap)
   (cond ((eq (layout-invalid layout) :uninitialized)
          ;; There was no layout before, we just created one which
          ;; we'll now initialize with our information.
          (setf (layout-length layout) length
                (layout-inherits layout) inherits
                (layout-depthoid layout) depthoid
-               (layout-raw-slot-metadata layout) raw-slot-metadata
+               (layout-bitmap layout) bitmap
                (layout-classoid layout) classoid
                (layout-invalid layout) nil))
         ;; FIXME: Now that LAYOUTs are born :UNINITIALIZED, maybe this
@@ -148,8 +147,7 @@
          ;; information, and we'll now check that old information
          ;; which was known with certainty is consistent with current
          ;; information which is known with certainty.
-         (check-layout layout classoid length inherits depthoid
-                       raw-slot-metadata)))
+         (check-layout layout classoid length inherits depthoid bitmap)))
   layout)
 
 ;;; In code for the target Lisp, we don't dump LAYOUTs using the
@@ -159,22 +157,15 @@
 ;;; to special non-ANSI low-level things like special fops, and we
 ;;; don't need to do that anyway because our code isn't going to be
 ;;; cold loaded, so we use the ordinary load form system.
-;;;
-;;; KLUDGE: A special hack causes this not to be called when we are
-;;; building code for the target Lisp. It would be tidier to just not
-;;; have it in place when we're building the target Lisp, but it
-;;; wasn't clear how to do that without rethinking DEF!STRUCT quite a
-;;; bit, so I punted. -- WHN 19990914
 #+sb-xc-host
-(defun make-load-form-for-layout (layout &optional env)
-  (declare (type layout layout))
+(defmethod make-load-form ((layout layout) &optional env)
   (declare (ignore env))
   (when (layout-invalid layout)
-    (compiler-error "can't dump reference to obsolete class: ~S"
-                    (layout-classoid layout)))
+    (sb!c::compiler-error "can't dump reference to obsolete class: ~S"
+                          (layout-classoid layout)))
   (let ((name (classoid-name (layout-classoid layout))))
     (unless name
-      (compiler-error "can't dump anonymous LAYOUT: ~S" layout))
+      (sb!c::compiler-error "can't dump anonymous LAYOUT: ~S" layout))
     ;; Since LAYOUT refers to a class which refers back to the LAYOUT,
     ;; we have to do this in two stages, like the TREE-WITH-PARENT
     ;; example in the MAKE-LOAD-FORM entry in the ANSI spec.
@@ -189,7 +180,8 @@
                              ',(layout-length layout)
                              ',(layout-inherits layout)
                              ',(layout-depthoid layout)
-                             ',(layout-raw-slot-metadata layout)))))
+                             ',(layout-bitmap layout)))))
+(!set-load-form-method layout (:xc :target) :ignore-it)
 
 ;;; If LAYOUT's slot values differ from the specified slot values in
 ;;; any interesting way, then give a warning and return T.
@@ -199,11 +191,10 @@
                            index
                            simple-vector
                            layout-depthoid
-                           layout-raw-slot-metadata-type))
+                           layout-bitmap))
                 redefine-layout-warning))
 (defun redefine-layout-warning (old-context old-layout
-                                context length inherits depthoid
-                                raw-slot-metadata)
+                                context length inherits depthoid bitmap)
   (declare (type layout old-layout) (type simple-string old-context context))
   (let ((name (layout-proper-name old-layout))
         (old-inherits (layout-inherits old-layout)))
@@ -236,16 +227,8 @@
                   old-context old-length
                   context length)
             t))
-        (let ((old-metadata (layout-raw-slot-metadata old-layout)))
-          (unless (= old-metadata raw-slot-metadata)
-            #!-interleaved-raw-slots
-            (warn "change in instance layout of class ~S:~%  ~
-                   ~A untagged slots: ~W~%  ~
-                   ~A untagged slots: ~W"
-                  name
-                  old-context old-metadata
-                  context raw-slot-metadata)
-            #!+interleaved-raw-slots
+        (let ((old-bitmap (layout-bitmap old-layout)))
+          (unless (= old-bitmap bitmap)
             (warn "change in placement of raw slots of class ~S ~
 between the ~A definition and the ~A definition"
                   name old-context context)
@@ -257,16 +240,13 @@ between the ~A definition and the ~A definition"
           t))))
 
 ;;; Require that LAYOUT data be consistent with CLASSOID, LENGTH,
-;;; INHERITS, DEPTHOID, and RAW-SLOT-METADATA.
-(declaim (ftype (function
-                 (layout classoid index simple-vector layout-depthoid
-                         layout-raw-slot-metadata-type))
+;;; INHERITS, DEPTHOID, and BITMAP.
+(declaim (ftype (function (layout classoid index simple-vector layout-depthoid layout-bitmap))
                 check-layout))
-(defun check-layout (layout classoid length inherits depthoid raw-slot-metadata)
+(defun check-layout (layout classoid length inherits depthoid bitmap)
   (aver (eq (layout-classoid layout) classoid))
   (when (redefine-layout-warning "current" layout
-                                 "compile time" length inherits depthoid
-                                 raw-slot-metadata)
+                                 "compile time" length inherits depthoid bitmap)
     ;; Classic CMU CL had more options here. There are several reasons
     ;; why they might want more options which are less appropriate for
     ;; us: (1) It's hard to fit the classic CMU CL flexible approach
@@ -290,11 +270,10 @@ between the ~A definition and the ~A definition"
 ;;; Used by the loader to forward-reference layouts for classes whose
 ;;; definitions may not have been loaded yet. This allows type tests
 ;;; to be loaded when the type definition hasn't been loaded yet.
-(declaim (ftype (function (symbol index simple-vector layout-depthoid
-                                  layout-raw-slot-metadata-type)
+(declaim (ftype (function (symbol index simple-vector layout-depthoid layout-bitmap)
                           layout)
                 find-and-init-or-check-layout))
-(defun find-and-init-or-check-layout (name length inherits depthoid metadata)
+(defun find-and-init-or-check-layout (name length inherits depthoid bitmap)
   (truly-the ; avoid an "assertion too complex to check" optimizer note
    (values layout &optional)
    (with-world-lock ()
@@ -305,7 +284,7 @@ between the ~A definition and the ~A definition"
                              length
                              inherits
                              depthoid
-                             metadata)))))
+                             bitmap)))))
 
 ;;; Record LAYOUT as the layout for its class, adding it as a subtype
 ;;; of all superclasses. This is the operation that "installs" a
@@ -352,8 +331,7 @@ between the ~A definition and the ~A definition"
                 (layout-inherits destruct-layout) (layout-inherits layout)
                 (layout-depthoid destruct-layout) (layout-depthoid layout)
                 (layout-length destruct-layout) (layout-length layout)
-                (layout-raw-slot-metadata destruct-layout)
-                (layout-raw-slot-metadata layout)
+                (layout-bitmap destruct-layout) (layout-bitmap layout)
                 (layout-info destruct-layout) (layout-info layout)
                 (classoid-layout classoid) destruct-layout)
           (setf (layout-invalid layout) nil
@@ -562,10 +540,11 @@ between the ~A definition and the ~A definition"
                    (warn "Changing meta-class of ~S from ~S to ~S."
                          name (classoid-name old) (classoid-name new))))))
             (:primitive
-             (error "Cannot redefine standard type ~S." name))
+             (error "Cannot redefine standard type ~
+                     ~/sb!impl:print-type-specifier/." name))
             (:defined
              (warn "redefining DEFTYPE type to be a class: ~
-                    ~/sb-impl::print-symbol-with-prefix/" name)
+                    ~/sb!impl::print-symbol-with-prefix/" name)
              (clear-info :type :expander name)
              (clear-info :type :source-location name)))
 
@@ -601,7 +580,8 @@ between the ~A definition and the ~A definition"
       ((nil))
       (:defined)
       (:primitive
-       (error "Attempt to remove :PRIMITIVE type: ~S" name))
+       (error "Attempt to remove :PRIMITIVE type: ~
+              ~/sb!impl:print-type-specifier/" name))
       ((:forthcoming-defclass-type :instance)
        (when cell
          ;; Note: We cannot remove the classoid cell from the table,
@@ -620,7 +600,8 @@ between the ~A definition and the ~A definition"
                (classoid-cell-pcl-class cell) nil))
        (clear-info :type :kind name)
        (clear-info :type :documentation name)
-       (clear-info :type :compiler-layout name)))))
+       (clear-info :type :compiler-layout name)
+       (values-specifier-type-cache-clear)))))
 
 ;;; Called when we are about to define NAME as a class meeting some
 ;;; predicate (such as a meta-class type test.) The first result is
@@ -671,12 +652,25 @@ between the ~A definition and the ~A definition"
 (!define-type-class classoid :enumerable #'classoid-enumerable-p
                     :might-contain-other-types nil)
 
+(defun classoid-inherits-from (sub super-or-name)
+  (declare (type classoid sub)
+           (type (or symbol classoid) super-or-name))
+  (let ((super (if (symbolp super-or-name)
+                   (find-classoid super-or-name)
+                   super-or-name)))
+    (find (classoid-layout super)
+          (layout-inherits (classoid-layout sub)))))
+
 ;;; We might be passed classoids with invalid layouts; in any pairwise
 ;;; class comparison, we must ensure that both are valid before
 ;;; proceeding.
 (defun %ensure-classoid-valid (classoid layout error-context)
+  (declare (ignorable error-context)) ; not used on host
   (aver (eq classoid (layout-classoid layout)))
   (or (not (layout-invalid layout))
+      ;; Avoid accidentally reaching code that can't work.
+      #+sb-xc-host (bug "(TYPEP x 'STANDARD-CLASSOID) can't be tested")
+      #-sb-xc-host
       (if (typep classoid 'standard-classoid)
           (let ((class (classoid-pcl-class classoid)))
             (cond
@@ -709,6 +703,7 @@ between the ~A definition and the ~A definition"
                  (%ensure-classoid-valid class2 layout2 errorp))
       (return-from %ensure-both-classoids-valid nil))))
 
+#-sb-xc-host ; No such thing as LAYOUT-OF, never mind the rest
 (defun update-object-layout-or-invalid (object layout)
   ;; FIXME: explain why this isn't (LAYOUT-FOR-STD-CLASS-P LAYOUT).
   (if (layout-for-std-class-p (layout-of object))
@@ -785,6 +780,15 @@ between the ~A definition and the ~A definition"
           ((eq (classoid-state class2) :sealed)
            ;; checking whether a subclass of both can be defined:
            (sealed-class-intersection2 class2 class1))
+          ;; If exactly one of CLASS{1,2} is a CONDITION-CLASSOID,
+          ;; there can be no intersection: sub-/superclass relations
+          ;; between CONDITION-CLASSOIDs and other CLASSOIDs are not
+          ;; possible and a CONDITION-CLASSOIDs cannot be changed into
+          ;; different CLASSOIDs.
+          ((let ((c1 (condition-classoid-p class1))
+                 (c2 (condition-classoid-p class2)))
+             (or (and c1 (not c2)) (and (not c1) c2)))
+           *empty-type*)
           (t
            ;; uncertain, since a subclass of both might be defined
            nil))))
@@ -793,9 +797,16 @@ between the ~A definition and the ~A definition"
 ;;; FUNCALLABLE-INSTANCE types (which used to be CLASSOIDs until CSR
 ;;; discovered that this was incompatible with the MOP class
 ;;; hierarchy).  See NAMED :COMPLEX-SUBTYPEP-ARG2
-(defvar *non-instance-classoid-types*
+(declaim (type cons **non-instance-classoid-types**))
+(defglobal **non-instance-classoid-types**
   '(symbol system-area-pointer weak-pointer code-component
-    lra fdefn random-class))
+    #!-(or x86 x86-64) lra
+    fdefn random-class))
+
+(defun classoid-non-instance-p (classoid)
+  (declare (type classoid classoid))
+  (member classoid **non-instance-classoid-types**
+          :key #'find-classoid))
 
 ;;; KLUDGE: we need this because of the need to represent
 ;;; intersections of two classes, even when empty at a given time, as
@@ -812,8 +823,7 @@ between the ~A definition and the ~A definition"
       (values nil nil)
       (invoke-complex-subtypep-arg1-method type1 class2 nil t)))
 
-(!define-type-method (classoid :negate) (type)
-  (make-negation-type :type type))
+(!define-type-method (classoid :negate) (type) (make-negation-type type))
 
 (!define-type-method (classoid :unparse) (type)
   (classoid-proper-name type))
@@ -858,7 +868,7 @@ between the ~A definition and the ~A definition"
      (character :codes (#.sb!vm:character-widetag)
                 :translation (character-set)
                 :prototype-form (code-char 42))
-     (symbol :codes (#.sb!vm:symbol-header-widetag)
+     (symbol :codes (#.sb!vm:symbol-widetag)
              :prototype-form '#:mu)
 
      (system-area-pointer :codes (#.sb!vm:sap-widetag)
@@ -866,15 +876,14 @@ between the ~A definition and the ~A definition"
      (weak-pointer :codes (#.sb!vm:weak-pointer-widetag)
       :prototype-form (make-weak-pointer (find-package "CL")))
      (code-component :codes (#.sb!vm:code-header-widetag))
-     ;; should this be #!-(or x86 x86-64) ?
-     (lra :codes (#.sb!vm:return-pc-header-widetag))
+     #!-(or x86 x86-64) (lra :codes (#.sb!vm:return-pc-widetag))
      (fdefn :codes (#.sb!vm:fdefn-widetag)
             :prototype-form (make-fdefn "42"))
      (random-class) ; used for unknown type codes
 
      (function
-      :codes (#.sb!vm:closure-header-widetag
-              #.sb!vm:simple-fun-header-widetag)
+      :codes (#.sb!vm:closure-widetag
+              #.sb!vm:simple-fun-widetag)
       :state :read-only
       :prototype-form (function (lambda () 42)))
 
@@ -1219,16 +1228,25 @@ between the ~A definition and the ~A definition"
       (let ((inherits-list (if (eq name t)
                                ()
                                (cons t (reverse inherits))))
-            (classoid (make-built-in-classoid
-                       :name name
-                       :translation (if trans-p :initializing nil)
-                       :direct-superclasses
-                       (if (eq name t)
-                           nil
-                           (mapcar #'find-classoid direct-superclasses)))))
-        (mark-ctype-interned classoid)
-        (setf (info :type :kind name) :primitive
-              (classoid-cell-classoid (find-classoid-cell name :create t)) classoid)
+            (classoid
+             (acond #+sb-xc ; genesis dumps some classoid literals
+                    ((find-classoid name nil)
+                     ;; Unseal it so that REGISTER-LAYOUT doesn't warn
+                     (setf (classoid-state it) nil)
+                     it)
+                    (t
+                     (setf (classoid-cell-classoid
+                            (find-classoid-cell name :create t))
+                           (mark-ctype-interned
+                            (make-built-in-classoid
+                             :name name
+                             :translation (if trans-p :initializing nil)
+                             :direct-superclasses
+                             (if (eq name t)
+                                 nil
+                                 (mapcar #'find-classoid
+                                         direct-superclasses)))))))))
+        (setf (info :type :kind name) :primitive)
         (unless trans-p
           (setf (info :type :builtin name) classoid))
         (let* ((inherits-vector
@@ -1248,56 +1266,10 @@ between the ~A definition and the ~A definition"
                                           0
                                           inherits-vector
                                           depthoid
-                                          0)
+                                          +layout-all-tagged+)
            :invalidate nil)))))
   (/show0 "done with loop over *!BUILT-IN-CLASSES*"))
 
-;;; Define temporary PCL STANDARD-CLASSes. These will be set up
-;;; correctly and the Lisp layout replaced by a PCL wrapper after PCL
-;;; is loaded and the class defined.
-(!cold-init-forms
-  (/show0 "about to define temporary STANDARD-CLASSes")
-  ;; You'd think with all the pedantic explanation in here it would at least
-  ;; be right, but it isn't: layout-inherits for FUNDAMENTAL-STREAM
-  ;; ends up as (T SLOT-OBJECT STREAM STANDARD-OBJECT)
-  (dolist (x '(;; Why is STREAM duplicated in this list? Because, when
-               ;; the inherits-vector of FUNDAMENTAL-STREAM is set up,
-               ;; a vector containing the elements of the list below,
-               ;; i.e. '(T STREAM STREAM), is created, and
-               ;; this is what the function ORDER-LAYOUT-INHERITS
-               ;; would do, too.
-               ;;
-               ;; So, the purpose is to guarantee a valid layout for
-               ;; the FUNDAMENTAL-STREAM class, matching what
-               ;; ORDER-LAYOUT-INHERITS would do.
-               ;; ORDER-LAYOUT-INHERITS would place STREAM at index 2
-               ;; in the INHERITS(-VECTOR). Index 1 would not be
-               ;; filled, so STREAM is duplicated there (as
-               ;; ORDER-LAYOUTS-INHERITS would do). Maybe the
-               ;; duplicate definition could be removed (removing a
-               ;; STREAM element), because FUNDAMENTAL-STREAM is
-               ;; redefined after PCL is set up, anyway. But to play
-               ;; it safely, we define the class with a valid INHERITS
-               ;; vector.
-               (fundamental-stream (t stream stream))))
-    (/show0 "defining temporary STANDARD-CLASS")
-    (let* ((name (first x))
-           (inherits-list (second x))
-           (classoid (make-standard-classoid :name name))
-           (classoid-cell (find-classoid-cell name :create t)))
-      ;; Needed to open-code the MAP, below
-      (declare (type list inherits-list))
-      (setf (classoid-cell-classoid classoid-cell) classoid
-            (info :type :kind name) :instance)
-      (let ((inherits (map 'simple-vector
-                           (lambda (x)
-                             (classoid-layout (find-classoid x)))
-                           inherits-list)))
-        #-sb-xc-host (/show0 "INHERITS=..") #-sb-xc-host (/hexstr inherits)
-        (register-layout (find-and-init-or-check-layout name 0 inherits -1 0)
-                         :invalidate nil))))
-  (/show0 "done defining temporary STANDARD-CLASSes"))
-
 ;;; Now that we have set up the class heterarchy, seal the sealed
 ;;; classes. This must be done after the subclasses have been set up.
 (!cold-init-forms
@@ -1308,13 +1280,14 @@ between the ~A definition and the ~A definition"
 ;;;; class definition/redefinition
 
 ;;; This is to be called whenever we are altering a class.
+#+sb-xc-host
+(defun %modify-classoid (classoid) (bug "MODIFY-CLASSOID ~S" classoid))
+#-sb-xc-host
 (defun %modify-classoid (classoid)
   (clear-type-caches)
-  (when (member (classoid-state classoid) '(:read-only :frozen))
+  (awhen (classoid-state classoid)
     ;; FIXME: This should probably be CERROR.
-    (warn "making ~(~A~) class ~S writable"
-          (classoid-state classoid)
-          (classoid-name classoid))
+    (warn "making ~(~A~) class ~S writable" it (classoid-name classoid))
     (setf (classoid-state classoid) nil)))
 
 ;;; Mark LAYOUT as invalid. Setting DEPTHOID -1 helps cause unsafe
