@@ -21,7 +21,6 @@
 ;;; is declared to hold a DEFSTRUCT-DESCRIPTION.
 (def!struct (defstruct-description
              (:conc-name dd-)
-             (:make-load-form-fun just-dump-it-normally)
              #-sb-xc-host (:pure t)
              (:constructor make-defstruct-description (null-lexenv-p name)))
   ;; name of the structure
@@ -52,8 +51,7 @@
   (slots () :type list)
   ;; a list of (NAME . INDEX) pairs for accessors of included structures
   (inherited-accessor-alist () :type list)
-  ;; number of elements we've allocated (See also RAW-LENGTH, which is not
-  ;; included in LENGTH.)
+  ;; number of elements including the layout itself (minimum=1)
   (length 0 :type index)
   ;; General kind of implementation.
   (type 'structure :type (member structure vector list
@@ -75,12 +73,11 @@
   ;; NIL if the option was given with no argument.
   (printer-fname nil :type (or cons symbol))
 
-  ;; The number of untagged slots at the end.
-  #!-interleaved-raw-slots (raw-length 0 :type index)
   ;; the value of the :PURE option, or :UNSPECIFIED. This is only
   ;; meaningful if DD-CLASS-P = T.
   (pure :unspecified :type (member t nil :unspecified)))
 #!-sb-fluid (declaim (freeze-type defstruct-description))
+(!set-load-form-method defstruct-description (:host :xc :target))
 
 ;;;; basic LAYOUT stuff
 
@@ -88,7 +85,6 @@
 ;;; in order to guarantee that several hash values can be added without
 ;;; overflowing into a bignum.
 (defconstant layout-clos-hash-limit (1+ (ash sb!xc:most-positive-fixnum -3))
-  #!+sb-doc
   "the exclusive upper bound on LAYOUT-CLOS-HASH values")
 ;; This must be DEF!TYPE and not just DEFTYPE because access to slots
 ;; of a layout occur "before" the structure definition is made in the
@@ -110,29 +106,7 @@
 ;;; well, since the initialization of layout slots is hardcoded there.
 ;;;
 ;;; FIXME: ...it would be better to automate this, of course...
-(def!struct (layout
-             ;; KLUDGE: A special hack keeps this from being
-             ;; called when building code for the
-             ;; cross-compiler. See comments at the DEFUN for
-             ;; this. -- WHN 19990914
-             (:make-load-form-fun #-sb-xc-host ignore-it
-                                  ;; KLUDGE: DEF!STRUCT at #+SB-XC-HOST
-                                  ;; time controls both the
-                                  ;; build-the-cross-compiler behavior
-                                  ;; and the run-the-cross-compiler
-                                  ;; behavior. The value below only
-                                  ;; works for build-the-cross-compiler.
-                                  ;; There's a special hack in
-                                  ;; EMIT-MAKE-LOAD-FORM which gives
-                                  ;; effectively IGNORE-IT behavior for
-                                  ;; LAYOUT at run-the-cross-compiler
-                                  ;; time. It would be cleaner to
-                                  ;; actually have an IGNORE-IT value
-                                  ;; stored, but it's hard to see how to
-                                  ;; do that concisely with the current
-                                  ;; DEF!STRUCT setup. -- WHN 19990930
-                                  #+sb-xc-host
-                                  make-load-form-for-layout))
+(def!struct (layout #-sb-xc-host (:constructor #!+immobile-space nil))
   ;; a pseudo-random hash value for use by CLOS.
   (clos-hash (random-layout-clos-hash) :type layout-clos-hash)
   ;; the class that this is a layout for
@@ -165,6 +139,8 @@
   ;;      a depth when it isn't quite.
   (depthoid -1 :type layout-depthoid)
   ;; the number of top level descriptor cells in each instance
+  ;; For [FUNCALLABLE-]STANDARD-OBJECT instances, this is the slot vector
+  ;; length, not the primitive object length.
   (length 0 :type index)
   ;; If this layout has some kind of compiler meta-info, then this is
   ;; it. If a structure, then we store the DEFSTRUCT-DESCRIPTION here.
@@ -176,14 +152,10 @@
   ;;
   ;; This slot is known to the C runtime support code.
   (pure nil :type (member t nil 0))
-  ;; Number of raw words at the end.
-  ;; This slot is known to the C runtime support code.
-  ;; It counts the number of untagged cells, not user-visible slots.
-  ;; e.g. on 32-bit machines, each (COMPLEX DOUBLE-FLOAT) counts as 4.
-  #!-interleaved-raw-slots (n-untagged-slots 0 :type index)
-  ;; Metadata
-  #!+interleaved-raw-slots (untagged-bitmap 0 :type unsigned-byte)
-  #!+interleaved-raw-slots (equalp-tests #() :type simple-vector)
+  ;; Map of raw slot indices.
+  (bitmap +layout-all-tagged+ :type layout-bitmap)
+  ;; Per-slot comparator for implementing EQUALP.
+  (equalp-tests #() :type simple-vector)
   ;; Definition location
   (source-location nil)
   ;; If this layout is for an object of metatype STANDARD-CLASS,
@@ -205,13 +177,28 @@
   (%for-std-class-b 0 :type bit :read-only t))
 (declaim (freeze-type layout))
 
+#!+(and immobile-space (host-feature sb-xc))
+(macrolet ((def-layout-maker ()
+             (let ((slots (dd-slots (find-defstruct-description 'layout))))
+               `(defun make-layout
+                    (&key ,@(mapcar (lambda (s) `(,(dsd-name s) ,(dsd-default s)))
+                                    slots))
+                  (declare ,@(mapcar (lambda (s) `(type ,(dsd-type s) ,(dsd-name s)))
+                                     slots))
+                  ;; After calling into C, registers are trashed,
+                  ;; so we pass everything as a single vector,
+                  ;; and don't rely on Lisp to write the slots of the layout.
+                  (dx-let ((data (vector ,@(mapcar #'dsd-name slots))))
+                    (truly-the layout
+                     (values (%primitive sb!vm::alloc-immobile-layout data))))))))
+  (def-layout-maker))
+
 ;;; The CLASSOID structure is a supertype of all classoid types.  A
 ;;; CLASSOID is also a CTYPE structure as recognized by the type
 ;;; system.  (FIXME: It's also a type specifier, though this might go
 ;;; away as with the merger of SB-PCL:CLASS and CL:CLASS it's no
 ;;; longer necessary)
 (def!struct (classoid
-             (:make-load-form-fun classoid-make-load-form-fun)
              (:include ctype
                        (class-info (type-class-or-lose 'classoid)))
              (:constructor nil)
@@ -309,10 +296,6 @@
 ;;; definitions with load-time resolution.
 (def!struct (classoid-cell
              (:constructor make-classoid-cell (name &optional classoid))
-             (:make-load-form-fun (lambda (c)
-                                    `(find-classoid-cell
-                                      ',(classoid-cell-name c)
-                                      :create t)))
              #-no-ansi-print-object
              (:print-object (lambda (s stream)
                               (print-unreadable-object (s stream :type t)
@@ -324,10 +307,30 @@
   ;; PCL class, if any
   (pcl-class nil))
 (declaim (freeze-type classoid-cell))
+(!set-load-form-method classoid-cell (:xc :target)
+  (lambda (self env)
+    (declare (ignore env))
+    `(find-classoid-cell ',(classoid-cell-name self) :create t)))
 
-;;; This would be a logical place to define FIND-CLASSOID-CELL,
-;;; but since 'globaldb' occurs later in the build order,
-;;; you'd have to go out of your way to declare INFO notinline.
+(defun find-classoid-cell (name &key create)
+  (let ((real-name (uncross name)))
+    (cond ((info :type :classoid-cell real-name))
+          (create
+           (get-info-value-initializing :type :classoid-cell real-name
+                                        (make-classoid-cell real-name))))))
+
+;;; Return the classoid with the specified NAME. If ERRORP is false,
+;;; then NIL is returned when no such class exists.
+(defun find-classoid (name &optional (errorp t))
+  (declare (type symbol name))
+  (let ((cell (find-classoid-cell name)))
+    (cond ((and cell (classoid-cell-classoid cell)))
+          (errorp
+           (error 'simple-type-error
+                  :datum nil
+                  :expected-type 'class
+                  :format-control "Class not yet defined: ~S"
+                  :format-arguments (list name))))))
 
 ;;;; PCL stuff
 
@@ -344,3 +347,41 @@
 
 (declaim (freeze-type built-in-classoid condition-classoid
                       standard-classoid static-classoid))
+
+(in-package "SB!C")
+
+;;; layout for this type being used by the compiler
+(define-info-type (:type :compiler-layout)
+  :type-spec (or layout null)
+  :default (lambda (name)
+             (let ((class (find-classoid name nil)))
+               (when class (classoid-layout class)))))
+
+(eval-when (#-sb-xc :compile-toplevel :load-toplevel :execute)
+(defun ftype-from-fdefn (name)
+  (declare (ignorable name))
+  ;; Again [as in (DEFINE-INFO-TYPE (:FUNCTION :TYPE) ...)] it's
+  ;; not clear how to generalize the FBOUNDP expression to the
+  ;; cross-compiler. -- WHN 19990330
+  #+sb-xc-host
+  (specifier-type 'function)
+  #-sb-xc-host
+  (let* ((fdefn (sb!kernel::find-fdefn name))
+         (fun (and fdefn (fdefn-fun fdefn))))
+    (if fun
+        (handler-bind ((style-warning #'muffle-warning))
+          (specifier-type (sb!impl::%fun-type fun)))
+        (specifier-type 'function)))))
+
+;;; The type specifier for this function, or a DEFSTRUCT-DESCRIPTION
+;;; or the symbol :GENERIC-FUNTION.
+;;; If a DD, it must contain a constructor whose name is
+;;; the one being sought in globaldb, which is used to derive the type.
+;;; If :GENERIC-FUNCTION, the info is recomputed from existing methods
+;;; and stored back into globaldb.
+(define-info-type (:function :type)
+  :type-spec (or ctype defstruct-description (member :generic-function))
+  :default #'ftype-from-fdefn)
+
+;;; A random place for this :-(
+#+sb-xc-host (setq *info-environment* (make-info-hashtable))
