@@ -8,59 +8,71 @@
 ;;; to elide bounds checking.
 (defmacro with-fop-stack (((stack-var &optional stack-expr) ptr-var count)
                           &body body)
-  `(let* (,@(when stack-expr
-              (list `(,stack-var (the simple-vector ,stack-expr))))
-          (,ptr-var (truly-the index (fop-stack-pop-n ,stack-var ,count))))
-     (macrolet ((fop-stack-ref (i)
-                  `(locally
-                       #-sb-xc-host
-                       (declare (optimize (sb!c::insert-array-bounds-checks 0)))
-                     (svref ,',stack-var (truly-the index ,i)))))
+  `(macrolet ((fop-stack-ref (i)
+                `(locally
+                     #-sb-xc-host
+                     (declare (optimize (sb!c::insert-array-bounds-checks 0)))
+                   (svref ,',stack-var (truly-the index ,i)))))
+     (let* (,@(when stack-expr
+                (list `(,stack-var (the simple-vector ,stack-expr))))
+            (,ptr-var (truly-the index (fop-stack-pop-n ,stack-var ,count))))
        ,@body)))
 
-;;; Define NAME as a fasl operation, with op-code FOP-CODE. PUSHP
-;;; describes what the body does to the fop stack:
-;;;   T
-;;;     The body might pop the fop stack. The result of the body is
-;;;     pushed on the fop stack.
-;;;   NIL
-;;;     The body might pop the fop stack. The result of the body is
-;;;     discarded.
+;;; Define NAME as a fasl operation, with op-code FOP-CODE.
+;;; PUSHP describes what the body does to the fop stack:
+;;;   T   - The result of the body is pushed on the fop stack.
+;;;   NIL - The result of the body is discarded.
+;;; In either case, the body is permitted to pop the stack.
 ;;;
-(defmacro !define-fop (fop-code (name &optional arglist (pushp t)) &body forms)
-  (aver (member pushp '(nil t)))
-  (multiple-value-bind (operands stack-args)
-      (if (consp (car arglist))
-          (ecase (caar arglist)
-            (:operands (values (cdar arglist) (cdr arglist))))
-          (values nil arglist))
-    (assert (<= (length operands) 2))
+(defmacro !define-fop (fop-code &rest stuff)
+  (multiple-value-bind (name allowp operands stack-args pushp forms)
+      (let ((allowp (if (eq (car stuff) :not-host)
+                        (progn (pop stuff) (or #+sb-xc t))
+                        t)))
+        (destructuring-bind ((name &optional arglist (pushp t)) . forms) stuff
+          (aver (member pushp '(nil t)))
+          (multiple-value-bind (operands stack-args)
+              (if (atom (car arglist))
+                  (values nil arglist)
+                  (ecase (caar arglist)
+                    (:operands (values (cdar arglist) (cdr arglist)))))
+            (assert (<= (length operands) 3))
+            (values name allowp operands stack-args pushp forms))))
     `(progn
        (defun ,name (.fasl-input. ,@operands)
          (declare (ignorable .fasl-input.))
-         (macrolet ((fasl-input () '(truly-the fasl-input .fasl-input.))
+         ,@(if allowp
+               `((macrolet
+                   ((fasl-input () '(truly-the fasl-input .fasl-input.))
                     (fasl-input-stream () '(%fasl-input-stream (fasl-input)))
                     (operand-stack () '(%fasl-input-stack (fasl-input)))
                     (skip-until () '(%fasl-input-skip-until (fasl-input))))
-           ,@(if (null stack-args)
-                 forms
-                 (with-unique-names (stack ptr)
-                   `((with-fop-stack ((,stack (operand-stack))
-                                      ,ptr ,(length stack-args))
-                       (multiple-value-bind ,stack-args
-                           (values ,@(loop for i below (length stack-args)
-                                           collect `(fop-stack-ref (+ ,ptr ,i))))
-                         ,@forms)))))))
+                  ,@(if (null stack-args)
+                        forms
+                        (with-unique-names (stack ptr)
+                         `((with-fop-stack ((,stack (operand-stack))
+                                            ,ptr ,(length stack-args))
+                             (multiple-value-bind ,stack-args
+                                 (values ,@(loop for i below (length stack-args)
+                                                 collect `(fop-stack-ref (+ ,ptr ,i))))
+                               ,@forms)))))))
+               `((declare (ignore ,@operands))
+                 (error ,(format nil "Not-host fop invoked: ~A" name)))))
        (!%define-fop ',name ,fop-code ,(length operands) ,(if pushp 1 0)))))
 
 (defun !%define-fop (name base-opcode n-operands pushp)
-  (declare (type (mod 3) n-operands)) ; 0, 1, or 2 are allowed
-  (let ((n-slots (expt 4 n-operands)))
+  (declare (type (mod 4) n-operands))
+  ;; If at least one non-stack operand is present, the same fop function
+  ;; appears in 4 consecutive cells in the fop table, with the low 2 bits
+  ;; of the fopcode determining the number of additional bytes to read
+  ;; for the first operand. The second and third operands are varint-encoded.
+  (let ((n-slots (if (plusp n-operands) 4 1)))
     (unless (zerop (mod base-opcode n-slots))
       (error "Opcode for fop ~S must be a multiple of ~D" name n-slots))
     (loop for opcode from base-opcode below (+ base-opcode n-slots)
-          when (functionp (svref **fop-funs** opcode))
-          do (let ((oname (svref **fop-names** opcode)))
+          for function = (svref **fop-funs** opcode)
+          when (functionp function)
+          do (let ((oname (nth-value 2 (function-lambda-expression function))))
                (when (and oname (not (eq oname name)))
                  (error "fop ~S with opcode ~D conflicts with fop ~S."
                         name opcode oname))))
@@ -70,13 +82,12 @@
                name base-opcode existing-opcode)))
     (setf (get name 'opcode) base-opcode)
     ;; The low 2 bits of the opcode comprise the length modifier if there is
-    ;; exactly one operand. Such opcodes are aligned in blocks of 4.
-    ;; 2-operand fops occupy 16 slots in a reserved range of the function table.
+    ;; at least one non-stack integer operand.
+    ;; If there is more than 1, they follow, using varint encoding.
     (dotimes (j n-slots)
       (let ((opcode (+ base-opcode j)))
-        (setf (svref **fop-names** opcode) name
-              (svref **fop-funs** opcode) (symbol-function name)
-              (sbit (car **fop-signatures**) (ash opcode -2)) (signum n-operands)
+        (setf (svref **fop-funs** opcode) (symbol-function name)
+              (aref (car **fop-signatures**) opcode) n-operands
               (sbit (cdr **fop-signatures**) opcode) pushp))))
   name)
 
@@ -102,13 +113,12 @@
       (setf (aref string i)
             (sb!xc:code-char (fast-read-byte)))))
   string)
-#!+sb-unicode
+#!+(and sb-unicode (host-feature sb-xc))
 (defun read-string-as-unsigned-byte-32
     (stream string &optional (length (length string)))
   (declare (type (simple-array character (*)) string)
            (type index length)
            (optimize speed))
-  #+sb-xc-host (bug "READ-STRING-AS-UNSIGNED-BYTE-32 called")
   (with-fast-read-byte ((unsigned-byte 8) stream)
     (dotimes (i length)
       (setf (aref string i)
@@ -146,56 +156,39 @@
   (push-fop-table x (fasl-input))
   x)
 
-;;; CMU CL had FOP-POP-FOR-EFFECT as fop 65, but it was never used and seemed
-;;; to have no possible use.
-(!define-fop 66 (fop-misc-trap)
-  #+sb-xc-host ; since xc host doesn't know how to compile %PRIMITIVE
-  (error "FOP-MISC-TRAP can't be defined without %PRIMITIVE.")
-  #-sb-xc-host
+(!define-fop 66 :not-host (fop-misc-trap)
   (%primitive sb!c:make-unbound-marker))
 
 (!define-fop 76 (fop-character ((:operands char-code)))
   (code-char char-code))
 
-(!define-fop 48 (fop-struct ((:operands size) layout))
-  (let* ((res (%make-instance size)) ; number of words excluding header
-         ;; Compute count of elements to pop from stack, sans layout.
-         ;; If instance-data-start is 0, then size is the count,
-         ;; otherwise subtract 1 because the layout consumes a slot.
-         (n-data-words (- size sb!vm:instance-data-start)))
-    (declare (type index size))
+;; %MAKE-INSTANCE does not exist on the host.
+(!define-fop 48 :not-host (fop-struct ((:operands size) layout))
+  (let ((res (%make-instance size)) ; number of words excluding header
+        ;; Discount the layout from number of user-visible words.
+        (n-data-words (- size sb!vm:instance-data-start)))
+    (setf (%instance-layout res) layout)
     (with-fop-stack ((stack (operand-stack)) ptr n-data-words)
-      (let ((ptr (+ ptr n-data-words)))
-        (declare (type index ptr))
-        (setf (%instance-layout res) layout)
-        #!-interleaved-raw-slots
-        (let* ((nuntagged (layout-n-untagged-slots layout))
-               (ntagged (- size nuntagged)))
-          (dotimes (n (1- ntagged))
-            (declare (type index n))
-            (setf (%instance-ref res (1+ n)) (fop-stack-ref (decf ptr))))
-          (dotimes (n nuntagged)
-            (declare (type index n))
-            (setf (%raw-instance-ref/word res (- nuntagged n 1))
-                  (fop-stack-ref (decf ptr)))))
-        #!+interleaved-raw-slots
-        (let ((metadata (layout-untagged-bitmap layout)))
-          (do ((i sb!vm:instance-data-start (1+ i)))
-              ((>= i size))
-            (declare (type index i))
-            (let ((val (fop-stack-ref (decf ptr))))
-              (if (logbitp i metadata)
-                  (setf (%raw-instance-ref/word res i) val)
-                  (setf (%instance-ref res i) val)))))))
+      (declare (type index ptr))
+      (let ((bitmap (layout-bitmap layout)))
+        ;; Values on the stack are in the same order as in the structure itself.
+        (do ((i sb!vm:instance-data-start (1+ i)))
+            ((>= i size))
+          (declare (type index i))
+          (let ((val (fop-stack-ref ptr)))
+            (if (logbitp i bitmap)
+                (setf (%instance-ref res i) val)
+                (setf (%raw-instance-ref/word res i) val))
+            (incf ptr)))))
     res))
 
-(!define-fop 45 (fop-layout (name inherits depthoid length metadata))
-  (find-and-init-or-check-layout name length inherits depthoid metadata))
+(!define-fop 45 (fop-layout (name inherits depthoid length bitmap))
+  (find-and-init-or-check-layout name length inherits depthoid bitmap))
 
 ;; Allocate a CLOS object. This is used when the compiler detects that
 ;; MAKE-LOAD-FORM returned a simple use of MAKE-LOAD-FORM-SAVING-SLOTS,
 ;; or possibly a hand-written equivalent (however unlikely).
-(!define-fop 68 (fop-allocate-instance (name) nil)
+(!define-fop 68 :not-host (fop-allocate-instance (name) nil)
   (let ((instance (allocate-instance (find-class (the symbol name)))))
     (push-fop-table instance (fasl-input))))
 
@@ -204,18 +197,17 @@
 ;; This wants a 'count' as the first item in the SLOT-NAMES argument
 ;; rather than using read-arg because many calls of this might share
 ;; the list, which must be constructed into the fop-table no matter what.
-(!define-fop 69 (fop-initialize-instance (slot-names obj) nil)
+(!define-fop 69 :not-host (fop-set-slot-values (slot-names obj) nil)
   (let* ((n-slots (pop slot-names))
          (stack (operand-stack))
          (ptr (fop-stack-pop-n stack n-slots)))
       (dotimes (i n-slots)
         (let ((val (svref stack (+ ptr i)))
               (slot-name (pop slot-names)))
-          (if (eq val 'sb!pcl::..slot-unbound..)
+          (if (eq val sb!pcl:+slot-unbound+)
               ;; SLOT-MAKUNBOUND-USING-CLASS might do something nonstandard.
               (slot-makunbound obj slot-name)
-              ;; FIXME: the DEFSETF for this isn't defined until warm load
-              (sb!pcl::set-slot-value obj slot-name val))))))
+              (setf (slot-value obj slot-name) val))))))
 
 (!define-fop 64 (fop-end-group () nil)
   (/show0 "THROWing FASL-GROUP-END")
@@ -235,74 +227,64 @@
 
 ;;;; fops for loading symbols
 
-(defstruct (undefined-package
-            (:copier nil))
-  error)
-
+(defstruct (undefined-package (:copier nil))
+  (error nil :read-only t))
 (declaim (freeze-type undefined-package))
 
-;; Not the same as the real definition, but nobody cares.
-;; We'd just better not get this error.
-#+sb-xc-host
-(deftype simple-package-error () 'simple-error)
+;; Cold load has its own implementation of all symbol fops,
+;; but we have to execute define-fop now to assign their numbers.
+(labels #+sb-xc-host ()
+        #-sb-xc-host
+        ((read-symbol-name (length+flag fasl-input)
+           (let* ((namelen (ash (the fixnum length+flag) -1))
+                  (base-p (logand length+flag 1))
+                  (elt-type (if (eql base-p 1) 'base-char 'character))
+                  (buffer (%fasl-input-name-buffer fasl-input))
+                  (string (the string (svref buffer base-p))))
+             (when (< (length string) namelen) ; grow
+               (setf string (make-string namelen :element-type elt-type)
+                     (svref buffer base-p) string))
+             (funcall (if (eql base-p 1)
+                          'read-base-string-as-bytes
+                          'read-string-as-unsigned-byte-32)
+                      (%fasl-input-stream fasl-input) string namelen)
+             (values string namelen elt-type)))
+         (aux-fop-intern (length+flag package fasl-input)
+           (multiple-value-bind (name length elt-type)
+               (read-symbol-name length+flag fasl-input)
+             (if (undefined-package-p package)
+                 (error 'simple-package-error
+                        :format-control "Error finding package for symbol ~s:~% ~a"
+                        :format-arguments
+                        (list (subseq name 0 length)
+                              (undefined-package-error package)))
+                 (push-fop-table (without-package-locks
+                                  (%intern name length package elt-type))
+                                 fasl-input))))
+         ;; Symbol-hash is usually computed lazily and memoized into a symbol.
+         ;; Laziness slightly improves the speed of allocation.
+         ;; But when loading fasls, the time spent in the loader totally swamps
+         ;; any time savings of not precomputing symbol-hash.
+         ;; INTERN hashes everything anyway, so let's be consistent
+         ;; and precompute the hashes of uninterned symbols too.
+         (ensure-hashed (symbol)
+           (ensure-symbol-hash symbol)
+           symbol))
 
-(defun aux-fop-intern (size package fasl-input)
-  (declare (optimize speed))
-  (let ((input-stream (%fasl-input-stream fasl-input))
-        (buffer (make-string size)))
-    #+sb-xc-host
-    (read-string-as-bytes input-stream buffer size)
-    #-sb-xc-host
-    (progn
-      #!+sb-unicode
-      (read-string-as-unsigned-byte-32 input-stream buffer size)
-      #!-sb-unicode
-      (read-string-as-bytes input-stream buffer size))
-    (if (undefined-package-p package)
-        (error 'simple-package-error
-               :format-control "Error finding package for symbol ~s:~% ~a"
-               :format-arguments
-               (list (subseq buffer 0 size)
-                     (undefined-package-error package)))
-        (push-fop-table (without-package-locks
-                          (%intern buffer size package nil))
-                        fasl-input))))
+  #-sb-xc-host (declare (inline ensure-hashed))
+  (!define-fop 80 :not-host (fop-lisp-symbol-save ((:operands length+flag)))
+    (aux-fop-intern length+flag *cl-package* (fasl-input)))
+  (!define-fop 84 :not-host (fop-keyword-symbol-save ((:operands length+flag)))
+    (aux-fop-intern length+flag *keyword-package* (fasl-input)))
+  (!define-fop #xF0 :not-host (fop-symbol-in-package-save ((:operands length+flag pkg-index)))
+    (aux-fop-intern length+flag (ref-fop-table (fasl-input) pkg-index) (fasl-input)))
 
-(!define-fop 80 (fop-lisp-symbol-save ((:operands namelen)))
-  (aux-fop-intern namelen *cl-package* (fasl-input)))
-(!define-fop 84 (fop-keyword-symbol-save ((:operands namelen)))
-  (aux-fop-intern namelen *keyword-package* (fasl-input)))
-
-;; But srsly? Most of the space is wasted by UCS4 encoding of ASCII.
-;; An extra word per symbol for the package is nothing by comparison.
-  ;; FIXME: Because we don't have FOP-SYMBOL-SAVE any more, an
-  ;; enormous number of symbols will fall through to this case,
-  ;; probably resulting in bloated fasl files. A new
-  ;; FOP-SYMBOL-IN-LAST-PACKAGE-SAVE/FOP-SMALL-SYMBOL-IN-LAST-PACKAGE-SAVE
-  ;; cloned fop pair could undo some of this bloat.
-(!define-fop #xF0 (fop-symbol-in-package-save ((:operands pkg-index namelen)))
-  (aux-fop-intern namelen (ref-fop-table (fasl-input) pkg-index) (fasl-input)))
-
-;;; Symbol-hash is usually computed lazily and memoized into a symbol.
-;;; Laziness slightly improves the speed of allocation.
-;;; But when loading fasls, the time spent in the loader totally swamps
-;;; any time savings of not precomputing symbol-hash.
-;;; INTERN hashes everything anyway, so let's be consistent
-;;; and precompute the hashes of uninterned symbols too.
-(macrolet ((ensure-hashed (symbol-form)
-             `(let ((symbol ,symbol-form))
-                (ensure-symbol-hash symbol)
-                symbol)))
-  (!define-fop 96 (fop-uninterned-symbol-save ((:operands namelen)))
-    (let ((res (make-string namelen)))
-      #!-sb-unicode
-      (read-string-as-bytes (fasl-input-stream) res)
-      #!+sb-unicode
-      (read-string-as-unsigned-byte-32 (fasl-input-stream) res)
-      (push-fop-table (ensure-hashed (make-symbol res))
+  (!define-fop 96 :not-host (fop-uninterned-symbol-save ((:operands length+flag)))
+    (multiple-value-bind (name len) (read-symbol-name length+flag (fasl-input))
+      (push-fop-table (ensure-hashed (make-symbol (subseq name 0 len)))
                       (fasl-input))))
 
-  (!define-fop 104 (fop-copy-symbol-save ((:operands table-index)))
+  (!define-fop 104 :not-host (fop-copy-symbol-save ((:operands table-index)))
     (push-fop-table (ensure-hashed
                      (copy-symbol (ref-fop-table (fasl-input) table-index)))
                     (fasl-input))))
@@ -310,16 +292,10 @@
 (!define-fop 44 (fop-package (pkg-designator))
   (find-undeleted-package-or-lose pkg-designator))
 
-(!define-fop 156 (fop-named-package-save ((:operands length)) nil)
+(!define-fop 156 :not-host (fop-named-package-save ((:operands length)) nil)
   (let ((package-name (make-string length)))
-    #+sb-xc-host
-    (read-string-as-bytes (fasl-input-stream) package-name)
-    #-sb-xc-host
-    (progn
-      #!-sb-unicode
-      (read-string-as-bytes (fasl-input-stream) package-name)
-      #!+sb-unicode
-      (read-string-as-unsigned-byte-32 (fasl-input-stream) package-name))
+    #!-sb-unicode (read-string-as-bytes (fasl-input-stream) package-name)
+    #!+sb-unicode (read-string-as-unsigned-byte-32 (fasl-input-stream) package-name)
     (push-fop-table
      (handler-case (find-undeleted-package-or-lose package-name)
        (simple-package-error (c)
@@ -356,9 +332,11 @@
   (with-fast-read-byte ((unsigned-byte 8) (fasl-input-stream))
     (fast-read-s-integer 1)))
 
-(!define-fop 70 (fop-ratio (num den)) (%make-ratio num den))
+;; No %MAKE-RATIO on host
+(!define-fop 70 :not-host (fop-ratio (num den)) (%make-ratio num den))
 
-(!define-fop 71 (fop-complex (realpart imagpart))
+;; No %MAKE-COMPLEX on host
+(!define-fop 71 :not-host (fop-complex (realpart imagpart))
   (%make-complex realpart imagpart))
 
 (macrolet ((fast-read-single-float ()
@@ -441,36 +419,31 @@
 
 ;;;; fops for loading arrays
 
-(!define-fop 100 (fop-base-string ((:operands length)))
-  (let ((res (make-string length :element-type 'base-char)))
-    (read-base-string-as-bytes (fasl-input-stream) res)
-    res))
+(!define-fop 100 :not-host (fop-base-string ((:operands length)))
+  (logically-readonlyize
+   (read-base-string-as-bytes (fasl-input-stream)
+                              (make-string length :element-type 'base-char))))
 
-#!+sb-unicode
 ;; FIXME: can save space by UTF-8 encoding, or use 1 bit to indicate pure ASCII
 ;; in the fasl even though the result will be a non-base string.
-(progn
-  #+sb-xc-host
-  (!define-fop 160 (fop-character-string ((:operands length)))
-    length ; touch the argument to avoid style-warning
-    (bug "CHARACTER-STRING FOP encountered"))
-
-  #-sb-xc-host
-  (!define-fop 160 (fop-character-string ((:operands length)))
-    (let ((res (make-string length)))
-      (read-string-as-unsigned-byte-32 (fasl-input-stream) res)
-      res)))
+#!+sb-unicode
+(!define-fop 160 :not-host (fop-character-string ((:operands length)))
+  (logically-readonlyize
+   (read-string-as-unsigned-byte-32 (fasl-input-stream)
+                                    (make-string length))))
 
 (!define-fop 92 (fop-vector ((:operands size)))
-  (let ((res (make-array size))
-        (stack (operand-stack)))
-    (declare (fixnum size))
-    (unless (zerop size)
-      (let ((ptr (fop-stack-pop-n stack size)))
-        (replace res stack :start2 ptr)))
-    res))
+  (if (zerop size)
+      #()
+      (let ((res (make-array size))
+            (stack (operand-stack)))
+        (declare (fixnum size))
+        (let ((ptr (fop-stack-pop-n stack size)))
+          (replace res stack :start2 ptr))
+        res)))
 
-(!define-fop 89 (fop-array (vec))
+;; No MAKE-ARRAY-HEADER on host
+(!define-fop 89 :not-host (fop-array (vec))
   (let* ((rank (read-word-arg (fasl-input-stream)))
          (length (length vec))
          (res (make-array-header sb!vm:simple-array-widetag rank)))
@@ -490,12 +463,12 @@
             (setf (aref array (sb!vm:saetp-typecode saetp))
                   (sb!vm:saetp-n-bits saetp)))
       array)
-    #!+sb-doc
     "255 means bad entry.")
 (declaim (type (simple-array (unsigned-byte 8) (#.(1+ sb!vm:widetag-mask)))
                **saetp-bits-per-length**))
 
-(!define-fop 43 (fop-spec-vector)
+;; No ALLOCATE-VECTOR on host (nor READ-N-BYTES)
+(!define-fop 43 :not-host (fop-spec-vector)
   (let* ((length (read-word-arg (fasl-input-stream)))
          (widetag (read-byte-arg (fasl-input-stream)))
          (bits-per-length (aref **saetp-bits-per-length** widetag))
@@ -503,7 +476,10 @@
                       (* length bits-per-length)))
          (bytes (ceiling bits sb!vm:n-byte-bits))
          (words (ceiling bytes sb!vm:n-word-bytes))
-         (vector (allocate-vector widetag length words)))
+         (vector (if (and (= widetag sb!vm:simple-vector-widetag)
+                          (= words 0))
+                     #()
+                     (allocate-vector widetag length words))))
     (declare (type index length bytes words)
              (type word bits))
     (read-n-bytes (fasl-input-stream) vector 0 bytes)
@@ -535,6 +511,10 @@
   (fop-funcall* (fasl-input-stream) (operand-stack) (skip-until)))
 (!define-fop 56 (fop-funcall-for-effect () nil)
   (fop-funcall* (fasl-input-stream) (operand-stack) (skip-until)))
+
+;;; For LOAD-TIME-VALUE which is used for MAKE-LOAD-FORM
+(!define-fop 57 (fop-funcall-no-skip)
+  (fop-funcall* (fasl-input-stream) (operand-stack) nil))
 
 ;;;; fops for fixing up circularities
 
@@ -553,10 +533,11 @@
          (obj (ref-fop-table (fasl-input) obi))
          (idx (read-word-arg (fasl-input-stream))))
     (if (%instancep obj) ; suspicious. should have been FOP-STRUCTSET
-        (setf (%instance-ref obj idx) val)
+        #+sb-xc (setf (%instance-ref obj idx) val)
+        #-sb-xc (bug "%instance-set?")
         (setf (svref obj idx) val))))
 
-(!define-fop 204 (fop-structset (val) nil)
+(!define-fop 204 :not-host (fop-structset (val) nil)
   (setf (%instance-ref (ref-fop-table (fasl-input)
                                       (read-word-arg (fasl-input-stream)))
                        (read-word-arg (fasl-input-stream)))
@@ -576,57 +557,41 @@
 ;;; putting the implementation and version in required fields in the
 ;;; fasl file header.)
 
-(!define-fop #xE0 (fop-code ((:operands n-boxed-words n-unboxed-bytes)))
-  (with-fop-stack ((stack (operand-stack)) ptr (1+ n-boxed-words))
-    (load-code n-boxed-words n-unboxed-bytes stack ptr (fasl-input-stream))))
+;; Cold-load calls COLD-LOAD-CODE instead
+(!define-fop #xE0 :not-host (fop-code ((:operands n-code-bytes n-boxed-words nfuns)))
+  ;; add 1 word for the toplevel-p flag and one for the debug-info
+  (with-fop-stack ((stack (operand-stack)) ptr (+ n-boxed-words 2))
+    (load-code nfuns n-boxed-words n-code-bytes stack ptr (fasl-input))))
 
 ;; this gets you an #<fdefn> object, not the result of (FDEFINITION x)
-(!define-fop 60 (fop-fdefn (name))
+;; cold-loader uses COLD-FDEFINITION-OBJECT instead.
+(!define-fop 60 :not-host (fop-fdefn (name))
   (awhen (deprecated-thing-p 'function name) ; returns the stage of deprecation
     (pushnew (list* it name :function)
              (%fasl-input-deprecated-stuff (fasl-input)) :test 'equal))
   (find-or-create-fdefn name))
 
-(!define-fop 65 (fop-known-fun (name))
+(!define-fop 65 :not-host (fop-known-fun (name))
   (%coerce-name-to-fun name))
 
-#!-(or x86 x86-64)
-(!define-fop 61 (fop-sanctify-for-execution (component))
+#!-(or x86 (and x86-64 (not immobile-space)))
+(!define-fop 61 :not-host (fop-sanctify-for-execution (component))
   (sb!vm:sanctify-for-execution component)
   component)
 
-(!define-fop 174 (fop-note-debug-source (debug-source) nil)
-  (warn "~@<FOP-NOTE-DEBUG-SOURCE seen in ordinary load (not cold load) -- ~
-very strange!  If you didn't do something to cause this, please report it as ~
-a bug.~@:>")
-  ;; we are going to be lenient with coming across this fop in a warm SBCL.
-  (setf (sb!c::debug-source-compiled debug-source) (get-universal-time)
-        (sb!c::debug-source-created debug-source)
-        (file-write-date (sb!c::debug-source-namestring debug-source))))
-
 ;;; Modify a slot in a CONSTANTS object.
-(!define-fop 140 (fop-alter-code ((:operands index) code value) nil)
+(!define-fop 140 :not-host (fop-alter-code ((:operands index) code value) nil)
   (setf (code-header-ref code index) value)
   (values))
 
-(!define-fop 139 (fop-fun-entry (code-object name arglist type info))
-  #+sb-xc-host ; since xc host doesn't know how to compile %PRIMITIVE
-  (progn code-object name arglist type info ; touch the arguments
-         (error "FOP-FUN-ENTRY can't be defined without %PRIMITIVE."))
-  #-sb-xc-host
-  (let ((offset (read-word-arg (fasl-input-stream))))
-    (declare (type index offset))
-    (unless (zerop (logand offset sb!vm:lowtag-mask))
-      (bug "unaligned function object, offset = #X~X" offset))
-    (let ((fun (%primitive sb!c:compute-fun code-object offset)))
-      (setf (%simple-fun-self fun) fun)
-      (setf (%simple-fun-next fun) (%code-entry-points code-object))
-      (setf (%code-entry-points code-object) fun)
-      (setf (%simple-fun-name fun) name)
-      (setf (%simple-fun-arglist fun) arglist)
-      (setf (%simple-fun-type fun) type)
-      (setf (%simple-fun-info fun) info)
-      fun)))
+(!define-fop #xFC :not-host (fop-fun-entry ((:operands fun-index)
+                                            code-object name arglist type info))
+  (let ((fun (%code-entry-point code-object fun-index)))
+    (setf (%simple-fun-name fun) name)
+    (setf (%simple-fun-arglist fun) arglist)
+    (setf (%simple-fun-type fun) type)
+    (setf (%simple-fun-info fun) info)
+    fun))
 
 ;;;; Some Dylan FOPs used to live here. By 1 November 1998 the code
 ;;;; was sufficiently stale that the functions it called were no
@@ -648,14 +613,38 @@ a bug.~@:>")
 (!define-fop 145 (fop-assembler-routine)
   (error "cannot load assembler code except at cold load"))
 
-(!define-fop 146 (fop-symbol-tls-fixup (code-object kind symbol))
+(!define-fop 146 :not-host (fop-symbol-tls-fixup (code-object kind symbol))
   (sb!vm:fixup-code-object code-object
                            (read-word-arg (fasl-input-stream))
                            (ensure-symbol-tls-index symbol)
                            kind)
   code-object)
 
-(!define-fop 147 (fop-foreign-fixup (code-object kind))
+#!+immobile-space
+(!define-fop 134 :not-host (fop-immobile-obj-fixup (code-object kind obj))
+  (sb!vm:fixup-code-object code-object
+                           (read-word-arg (fasl-input-stream))
+                           (get-lisp-obj-address obj) ; OBJ can't move
+                           kind :immobile-object)
+  code-object)
+
+#!+immobile-code
+(progn
+  (!define-fop 133 :not-host (fop-named-call-fixup (code-object kind name))
+    (sb!vm:fixup-code-object code-object
+                             (read-word-arg (fasl-input-stream))
+                             (sb!vm::fdefn-entry-address name)
+                             kind :named-call)
+    code-object)
+
+  (!define-fop 135 :not-host (fop-static-call-fixup (code-object kind name))
+    (sb!vm:fixup-code-object code-object
+                             (read-word-arg (fasl-input-stream))
+                             (sb!vm::function-raw-address name)
+                             kind)
+    code-object))
+
+(!define-fop 147 :not-host (fop-foreign-fixup (code-object kind))
   (let* ((len (read-byte-arg (fasl-input-stream)))
          (sym (make-string len :element-type 'base-char)))
     (read-n-bytes (fasl-input-stream) sym 0 len)
@@ -665,7 +654,7 @@ a bug.~@:>")
                              kind)
     code-object))
 
-(!define-fop 148 (fop-assembler-fixup (code-object kind routine))
+(!define-fop 148 :not-host (fop-assembler-fixup (code-object kind routine))
   (multiple-value-bind (value found) (gethash routine *assembler-routines*)
     (unless found
       (error "undefined assembler routine: ~S" routine))
@@ -673,7 +662,7 @@ a bug.~@:>")
                              value kind))
   code-object)
 
-(!define-fop 149 (fop-code-object-fixup (code-object kind))
+(!define-fop 149 :not-host (fop-code-object-fixup (code-object kind))
     ;; Note: We don't have to worry about GC moving the code-object after
     ;; the GET-LISP-OBJ-ADDRESS and before that value is deposited, because
     ;; we can only use code-object fixups when code-objects don't move.
@@ -682,7 +671,7 @@ a bug.~@:>")
   code-object)
 
 #!+linkage-table
-(!define-fop 150 (fop-foreign-dataref-fixup (code-object kind))
+(!define-fop 150 :not-host (fop-foreign-dataref-fixup (code-object kind))
   (let* ((len (read-byte-arg (fasl-input-stream)))
          (sym (make-string len :element-type 'base-char)))
     (read-n-bytes (fasl-input-stream) sym 0 len)
@@ -731,8 +720,6 @@ a bug.~@:>")
   (values))
 
 ;;; Primordial layouts.
-;;; At the rate the opcode space is filling up, it might be wise to
-;;; rethink using 16 each for FOP-CODE and -SYMBOL-IN-PACKAGE-SAVE.
 (macrolet ((frob (&rest specs)
              `(progn
                 (defun known-layout-fop (name)
@@ -747,7 +734,6 @@ a bug.~@:>")
                           specs))))
   (frob (#x6c t)
         (#x6d structure-object)
-        (#x6e structure!object)
         (#x6f definition-source-location)
         (#x70 sb!c::debug-fun)
         (#x71 sb!c::compiled-debug-fun)
