@@ -18,7 +18,7 @@
 ;;;; -- WHN 20000127
 
 (declaim (maybe-inline
-          tree-equal nth %setnth nthcdr make-list
+          tree-equal %setnth nthcdr
           tailp union
           nunion intersection nintersection set-difference nset-difference
           set-exclusive-or nset-exclusive-or subsetp acons
@@ -27,90 +27,90 @@
           subst-if-not nsubst nsubst-if nsubst-if-not sublis nsublis))
 
 ;;; These functions perform basic list operations.
-(defun car (list) #!+sb-doc "Return the 1st object in a list." (car list))
+(defun car (list) "Return the 1st object in a list." (car list))
 (defun cdr (list)
-  #!+sb-doc "Return all but the first object in a list."
+  "Return all but the first object in a list."
   (cdr list))
-(defun cadr (list) #!+sb-doc "Return the 2nd object in a list." (cadr list))
-(defun cdar (list) #!+sb-doc "Return the cdr of the 1st sublist." (cdar list))
-(defun caar (list) #!+sb-doc "Return the car of the 1st sublist." (caar list))
+(defun cadr (list) "Return the 2nd object in a list." (cadr list))
+(defun cdar (list) "Return the cdr of the 1st sublist." (cdar list))
+(defun caar (list) "Return the car of the 1st sublist." (caar list))
 (defun cddr (list)
-  #!+sb-doc "Return all but the 1st two objects of a list."
+  "Return all but the 1st two objects of a list."
   (cddr list))
 (defun caddr (list)
-  #!+sb-doc "Return the 1st object in the cddr of a list."
+  "Return the 1st object in the cddr of a list."
   (caddr list))
 (defun caadr (list)
-  #!+sb-doc "Return the 1st object in the cadr of a list."
+  "Return the 1st object in the cadr of a list."
   (caadr list))
 (defun caaar (list)
-  #!+sb-doc "Return the 1st object in the caar of a list."
+  "Return the 1st object in the caar of a list."
   (caaar list))
 (defun cdaar (list)
-  #!+sb-doc "Return the cdr of the caar of a list."
+  "Return the cdr of the caar of a list."
   (cdaar list))
 (defun cddar (list)
-  #!+sb-doc "Return the cdr of the cdar of a list."
+  "Return the cdr of the cdar of a list."
   (cddar list))
 (defun cdddr (list)
-  #!+sb-doc "Return the cdr of the cddr of a list."
+  "Return the cdr of the cddr of a list."
   (cdddr list))
 (defun cadar (list)
-  #!+sb-doc "Return the car of the cdar of a list."
+  "Return the car of the cdar of a list."
   (cadar list))
 (defun cdadr (list)
-  #!+sb-doc "Return the cdr of the cadr of a list."
+  "Return the cdr of the cadr of a list."
   (cdadr list))
 (defun caaaar (list)
-  #!+sb-doc "Return the car of the caaar of a list."
+  "Return the car of the caaar of a list."
   (caaaar list))
 (defun caaadr (list)
-  #!+sb-doc "Return the car of the caadr of a list."
+  "Return the car of the caadr of a list."
   (caaadr list))
 (defun caaddr (list)
-  #!+sb-doc "Return the car of the caddr of a list."
+  "Return the car of the caddr of a list."
   (caaddr list))
 (defun cadddr (list)
-  #!+sb-doc "Return the car of the cdddr of a list."
+  "Return the car of the cdddr of a list."
   (cadddr list))
 (defun cddddr (list)
-  #!+sb-doc "Return the cdr of the cdddr of a list."
+  "Return the cdr of the cdddr of a list."
   (cddddr list))
 (defun cdaaar (list)
-  #!+sb-doc "Return the cdr of the caaar of a list."
+  "Return the cdr of the caaar of a list."
   (cdaaar list))
 (defun cddaar (list)
-  #!+sb-doc "Return the cdr of the cdaar of a list."
+  "Return the cdr of the cdaar of a list."
   (cddaar list))
 (defun cdddar (list)
-  #!+sb-doc "Return the cdr of the cddar of a list."
+  "Return the cdr of the cddar of a list."
   (cdddar list))
 (defun caadar (list)
-  #!+sb-doc "Return the car of the cadar of a list."
+  "Return the car of the cadar of a list."
   (caadar list))
 (defun cadaar (list)
-  #!+sb-doc "Return the car of the cdaar of a list."
+  "Return the car of the cdaar of a list."
   (cadaar list))
 (defun cadadr (list)
-  #!+sb-doc "Return the car of the cdadr of a list."
+  "Return the car of the cdadr of a list."
   (cadadr list))
 (defun caddar (list)
-  #!+sb-doc "Return the car of the cddar of a list."
+  "Return the car of the cddar of a list."
   (caddar list))
 (defun cdaadr (list)
-  #!+sb-doc "Return the cdr of the caadr of a list."
+  "Return the cdr of the caadr of a list."
   (cdaadr list))
 (defun cdadar (list)
-  #!+sb-doc "Return the cdr of the cadar of a list."
+  "Return the cdr of the cadar of a list."
   (cdadar list))
 (defun cdaddr (list)
-  #!+sb-doc "Return the cdr of the caddr of a list."
+  "Return the cdr of the caddr of a list."
   (cdaddr list))
 (defun cddadr (list)
-  #!+sb-doc "Return the cdr of the cdadr of a list."
+  "Return the cdr of the cdadr of a list."
   (cddadr list))
 (defun cons (se1 se2)
-  #!+sb-doc "Return a list with SE1 as the CAR and SE2 as the CDR."
+  "Return a list with SE1 as the CAR and SE2 as the CDR."
   (cons se1 se2))
 
 (declaim (maybe-inline tree-equal-test tree-equal-test-not))
@@ -135,24 +135,44 @@
         ((funcall test x y) t)
         (t ())))
 
-(defun tree-equal (x y &key (test #'eql testp) (test-not nil notp))
-  #!+sb-doc
+(defun tree-equal-eql (x y)
+  (labels ((recurse (x y)
+             (if (eq x y)
+                 t
+                 (do ((x x (cdr x))
+                      (y y (cdr y)))
+                     ((or (atom x)
+                          (atom y))
+                      (eql x y))
+                   (cond ((consp (car x))
+                          (when (atom (car y))
+                            (return-from tree-equal-eql))
+                          (recurse (car x) (car y)))
+                         ((not (eql (car x) (car y)))
+                          (return-from tree-equal-eql)))))))
+    (recurse x y)))
+
+(defun tree-equal (x y &key (test nil testp) (test-not nil notp))
   "Return T if X and Y are isomorphic trees with identical leaves."
-  (when (and testp notp)
-    (error ":TEST and :TEST-NOT were both supplied."))
-  (if test-not
-      (tree-equal-test-not x y (%coerce-callable-to-fun test-not))
-      (tree-equal-test x y (%coerce-callable-to-fun test))))
+  (declare (explicit-check))
+  (cond (notp
+         (when testp
+           (error ":TEST and :TEST-NOT were both supplied."))
+         (tree-equal-test-not x y (%coerce-callable-to-fun test-not)))
+        ((or (not test)
+             (eql test #'eql)
+             (eql test 'eql))
+         (tree-equal-eql x y))
+        (t
+         (tree-equal-test x y (%coerce-callable-to-fun test)))))
 
 (defun endp (object)
-  #!+sb-doc
   "This is the recommended way to test for the end of a proper list. It
    returns true if OBJECT is NIL, false if OBJECT is a CONS, and an error
    for any other type of OBJECT."
   (endp object))
 
 (defun list-length (list)
-  #!+sb-doc
   "Return the length of the given List, or Nil if the List is circular."
   (do ((n 0 (+ n 2))
        (y list (cddr y))
@@ -165,57 +185,45 @@
     (when (and (eq y z) (> n 0)) (return nil))))
 
 (defun nth (n list)
-  #!+sb-doc
   "Return the nth object in a list where the car is the zero-th element."
+  (declare (explicit-check))
   (car (nthcdr n list)))
 
 (defun first (list)
-  #!+sb-doc
   "Return the 1st object in a list or NIL if the list is empty."
   (car list))
 (defun second (list)
-  #!+sb-doc
   "Return the 2nd object in a list or NIL if there is no 2nd object."
   (cadr list))
 (defun third (list)
-  #!+sb-doc
   "Return the 3rd object in a list or NIL if there is no 3rd object."
   (caddr list))
 (defun fourth (list)
-  #!+sb-doc
   "Return the 4th object in a list or NIL if there is no 4th object."
   (cadddr list))
 (defun fifth (list)
-  #!+sb-doc
   "Return the 5th object in a list or NIL if there is no 5th object."
   (car (cddddr list)))
 (defun sixth (list)
-  #!+sb-doc
   "Return the 6th object in a list or NIL if there is no 6th object."
   (cadr (cddddr list)))
 (defun seventh (list)
-  #!+sb-doc
   "Return the 7th object in a list or NIL if there is no 7th object."
   (caddr (cddddr list)))
 (defun eighth (list)
-  #!+sb-doc
   "Return the 8th object in a list or NIL if there is no 8th object."
   (cadddr (cddddr list)))
 (defun ninth (list)
-  #!+sb-doc
   "Return the 9th object in a list or NIL if there is no 9th object."
   (car (cddddr (cddddr list))))
 (defun tenth (list)
-  #!+sb-doc
   "Return the 10th object in a list or NIL if there is no 10th object."
   (cadr (cddddr (cddddr list))))
 (defun rest (list)
-  #!+sb-doc
   "Means the same as the cdr of a list."
   (cdr list))
 
 (defun nthcdr (n list)
-  #!+sb-doc
   "Performs the cdr function n times on a list."
   (flet ((fast-nthcdr (n list)
            (declare (type index n))
@@ -292,7 +300,6 @@
     (lastn-macro unsigned-byte))
 
   (defun last (list &optional (n 1))
-    #!+sb-doc
     "Return the last N conses (not the last element!) of a list."
     (case n
       (1 (last1-macro))
@@ -313,7 +320,6 @@
       form))
 
 (defun list (&rest args)
-  #!+sb-doc
   "Return constructs and returns a list of its arguments."
   args)
 
@@ -321,19 +327,34 @@
 ;;; a dotted pair.
 
 (defun list* (arg &rest others)
-  #!+sb-doc
   "Return a list of the arguments with last cons a dotted pair."
-  ;; We know the &REST is a proper list.
-  (declare (optimize (sb!c::type-check 0)))
-  (cond ((atom others) arg)
-        ((atom (cdr others)) (cons arg (car others)))
-        (t (do ((x others (cdr x)))
-               ((null (cddr x)) (rplacd x (cadr x))))
-           (cons arg others))))
+  (let ((length (length others)))
+    (cond ((= length 0) arg)
+          ((= length 1)
+           (cons arg (fast-&rest-nth 0 others)))
+          (t
+           (let* ((cons (list arg))
+                  (result cons)
+                  (index 0)
+                  (1-length (1- length)))
+             (loop
+              (cond
+                ((< index 1-length)
+                 (setf cons
+                       (setf (cdr cons)
+                             (list (fast-&rest-nth index others))))
+                 (incf index))
+                (t (return nil))))
+             (setf (cdr cons) (fast-&rest-nth index others))
+             result)))))
 
 (defun make-list (size &key initial-element)
-  #!+sb-doc
   "Constructs a list with size elements each set to value"
+  (declare (explicit-check))
+  (%make-list size initial-element))
+;;; This entry point is to be preferred, irrespective of
+;;; whether or not the backend has vops for %MAKE-LIST.
+(defun %make-list (size initial-element)
   (declare (type index size))
   (do ((count size (1- count))
        (result '() (cons initial-element result)))
@@ -341,56 +362,42 @@
     (declare (type index count))))
 
 (defun append (&rest lists)
-  #!+sb-doc
   "Construct a new list by concatenating the list arguments"
-  (declare (truly-dynamic-extent lists) (optimize speed))
-  (labels ((fail (object)
-             (error 'type-error
-                    :datum object
-                    :expected-type 'list))
-           (append-into (last-cons current rest)
-             ;; Set (CDR LAST-CONS) to (APPLY #'APPEND CURRENT REST).
-             (declare (cons last-cons rest))
-             (if (listp current)
-                 (if (consp current)
-                     ;; normal case, cdr down the list
-                     (append-into (setf (cdr last-cons) (list (car current)))
-                                  (cdr current)
-                                  rest)
-                     ;; empty list
-                     (let ((more (cdr rest)))
-                       (if (null more)
-                           (setf (cdr last-cons) (car rest))
-                           (append-into last-cons (car rest) more))))
-                 (fail current)))
-           (append1 (lists)
-             (let ((current (car lists))
-                   (rest (cdr lists)))
-               (cond ((null rest)
-                      current)
-                     ((consp current)
-                      (let ((result (truly-the cons (list (car current)))))
-                        (append-into result
-                                     (cdr current)
-                                     rest)
-                        result))
-                     ((null current)
-                      (append1 rest))
-                     (t
-                      (fail current))))))
-    (append1 lists)))
+  (let* ((result (list nil))
+         (tail result)
+         (index 0)
+         (length (length lists))
+         (last (1- length)))
+    (declare (truly-dynamic-extent result))
+    (loop
+     (cond
+       ((< (truly-the index index) last)
+        (let ((list (fast-&rest-nth (truly-the index index) lists)))
+          (dolist (elt list)
+            (setf (cdr (truly-the cons tail)) (list elt)
+                  tail (cdr tail))))
+        (incf index))
+       (t (return nil))))
+    (cond
+      ((zerop length) nil)
+      ((null (cdr result))
+       (fast-&rest-nth (truly-the index last) lists))
+      (t
+       (setf (cdr (truly-the cons tail))
+             (fast-&rest-nth (truly-the index last) lists))
+       (cdr result)))))
 
 (defun append2 (x y)
-  (declare (optimize speed (sb!c::verify-arg-count 0)))
+  (declare (optimize (sb!c::verify-arg-count 0)))
   (if (null x)
       y
       (let ((result (list (car x))))
         (do ((more (cdr x) (cdr more))
              (tail result (cdr tail)))
             ((null more)
-             (rplacd tail y)
+             (rplacd (truly-the cons tail) y)
              result)
-          (rplacd tail (list (car more)))))))
+                    (rplacd (truly-the cons tail) (list (car more)))))))
 
 
 ;;;; list copying functions
@@ -412,12 +419,10 @@
               copy))))))
 
 (defun copy-list (list)
-  #!+sb-doc
   "Return a new list which is EQUAL to LIST. LIST may be improper."
   (!copy-list-macro list))
 
 (defun copy-alist (alist)
-  #!+sb-doc
   "Return a new association list which is EQUAL to ALIST."
   (if (endp alist)
       alist
@@ -438,7 +443,6 @@
         result)))
 
 (defun copy-tree (object)
-  #!+sb-doc
   "Recursively copy trees of conses."
   (if (consp object)
       (let ((result (list (if (consp (car object))
@@ -460,7 +464,6 @@
 ;;;; more commonly-used list functions
 
 (defun revappend (x y)
-  #!+sb-doc
   "Return (append (reverse x) y)."
   (do ((top x (cdr top))
        (result y (cons (car top) result)))
@@ -478,7 +481,6 @@
 ;;; and it avoids running down the last argument to NCONC which allows
 ;;; the last argument to be circular.
 (defun nconc (&rest lists)
-   #!+sb-doc
    "Concatenates the lists given as arguments (by changing them)"
    (declare (optimize speed))
    (flet ((fail (object)
@@ -505,7 +507,6 @@
                 (return result)))))))
 
 (defun nreconc (x y)
-  #!+sb-doc
   "Return (NCONC (NREVERSE X) Y)."
   (do ((1st (cdr x) (if (endp 1st) 1st (cdr 1st)))
        (2nd x 1st)              ;2nd follows first down the list.
@@ -549,7 +550,6 @@
                      list)))))))
 
 (defun ldiff (list object)
-  #!+sb-doc
   "Return a new list, whose elements are those of LIST that appear before
    OBJECT. If OBJECT is not a tail of LIST, a copy of LIST is returned.
    LIST must be a proper list or a dotted list."
@@ -567,12 +567,10 @@
 ;;;; functions to alter list structure
 
 (defun rplaca (cons x)
-  #!+sb-doc
   "Change the CAR of CONS to X and return the CONS."
   (rplaca cons x))
 
 (defun rplacd (cons x)
-  #!+sb-doc
   "Change the CDR of CONS to X and return the CONS."
   (rplacd cons x))
 
@@ -609,6 +607,11 @@
   `(if ,key
        (funcall ,key ,element)
        ,element))
+
+(defmacro apply-key-function (key element)
+  `(if ,key
+       (funcall (truly-the function ,key) ,element)
+       ,element))
 
 ;;;; macros for (&KEY (KEY #'IDENTITY) (TEST #'EQL TESTP) (TEST-NOT NIL NOTP))
 
@@ -628,7 +631,6 @@
 ;;;; substitution of expressions
 
 (defun subst (new old tree &key key (test #'eql testp) (test-not #'eql notp))
-  #!+sb-doc
   "Substitutes new for subtrees matching old."
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
@@ -648,7 +650,6 @@
       (s tree))))
 
 (defun subst-if (new test tree &key key)
-  #!+sb-doc
   "Substitutes new for subtrees for which test is true."
   (let ((test (%coerce-callable-to-fun test))
         (key (and key (%coerce-callable-to-fun key))))
@@ -664,7 +665,6 @@
       (s tree))))
 
 (defun subst-if-not (new test tree &key key)
-  #!+sb-doc
   "Substitutes new for subtrees for which test is false."
   (let ((test (%coerce-callable-to-fun test))
         (key (and key (%coerce-callable-to-fun key))))
@@ -680,7 +680,6 @@
       (s tree))))
 
 (defun nsubst (new old tree &key key (test #'eql testp) (test-not #'eql notp))
-  #!+sb-doc
   "Substitute NEW for subtrees matching OLD."
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
@@ -703,7 +702,6 @@
       (s tree))))
 
 (defun nsubst-if (new test tree &key key)
-  #!+sb-doc
   "Substitute NEW for subtrees of TREE for which TEST is true."
   (let ((test (%coerce-callable-to-fun test))
         (key (and key (%coerce-callable-to-fun key))))
@@ -722,7 +720,6 @@
       (s tree))))
 
 (defun nsubst-if-not (new test tree &key key)
-  #!+sb-doc
   "Substitute NEW for subtrees of TREE for which TEST is false."
   (let ((test (%coerce-callable-to-fun test))
         (key (and key (%coerce-callable-to-fun key))))
@@ -741,7 +738,6 @@
       (s tree))))
 
 (defun sublis (alist tree &key key (test #'eql testp) (test-not #'eql notp))
-  #!+sb-doc
   "Substitute from ALIST into TREE nondestructively."
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
@@ -775,7 +771,6 @@
           (assoc ,key-tmp alist :test test)))))
 
 (defun nsublis (alist tree &key key (test #'eql testp) (test-not #'eql notp))
-  #!+sb-doc
   "Substitute from ALIST into TRUE destructively."
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
@@ -802,9 +797,9 @@
 ;;;; functions for using lists as sets
 
 (defun member (item list &key key (test nil testp) (test-not nil notp))
-  #!+sb-doc
   "Return the tail of LIST beginning with first element satisfying EQLity,
    :TEST, or :TEST-NOT with the given ITEM."
+  (declare (explicit-check))
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
   (let ((key (and key (%coerce-callable-to-fun key)))
@@ -824,8 +819,8 @@
                (%member item list))))))
 
 (defun member-if (test list &key key)
-  #!+sb-doc
   "Return tail of LIST beginning with first element satisfying TEST."
+  (declare (explicit-check))
   (let ((test (%coerce-callable-to-fun test))
         (key (and key (%coerce-callable-to-fun key))))
     (if key
@@ -833,8 +828,8 @@
         (%member-if test list))))
 
 (defun member-if-not (test list &key key)
-  #!+sb-doc
   "Return tail of LIST beginning with first element not satisfying TEST."
+  (declare (explicit-check))
   (let ((test (%coerce-callable-to-fun test))
         (key (and key (%coerce-callable-to-fun key))))
     (if key
@@ -842,7 +837,6 @@
         (%member-if-not test list))))
 
 (defun tailp (object list)
-  #!+sb-doc
   "Return true if OBJECT is the same as some tail of LIST, otherwise
    returns false. LIST must be a proper list or a dotted list."
   (do ((list list (cdr list)))
@@ -851,8 +845,8 @@
         (return t))))
 
 (defun adjoin (item list &key key (test #'eql testp) (test-not nil notp))
-  #!+sb-doc
   "Add ITEM to LIST unless it is already a member"
+  (declare (explicit-check))
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
   (let ((key (and key (%coerce-callable-to-fun key)))
@@ -871,36 +865,83 @@
                (%adjoin-key item list key)
                (%adjoin item list))))))
 
+;;; For cases where MEMBER is called in a loop this allows to perform
+;;; the dispatch that the MEMBER function does only once.
+(defmacro with-member-test ((test-var &optional first-clause) &body body)
+  `(let* ((key (and key (%coerce-callable-to-fun key)))
+          (,test-var (cond ,@(and first-clause ; used by LIST-REMOVE-DUPLICATES*
+                                  `(,first-clause))
+                           (notp
+                            (if key
+                                (lambda (x list2 key test)
+                                  (%member-key-test-not (funcall (truly-the function key) x)
+                                                        list2 key test))
+                                (lambda (x list2 key test)
+                                  (declare (ignore key))
+                                  (%member-test-not x list2 test))))
+                           (testp
+                            (if key
+                                (lambda (x list2 key test)
+                                  (%member-key-test (funcall (truly-the function key) x)
+                                                    list2 key test))
+                                (lambda (x list2 key test)
+                                  (declare (ignore key))
+                                  (%member-test x list2 test))))
+                           (key
+                            (lambda (x list2 key test)
+                              (declare (ignore test))
+                              (%member-key (funcall (truly-the function key) x) list2 key)))
+                           (t
+                            (lambda (x list2 key test)
+                              (declare (ignore key test))
+                              (%member x list2)))))
+          (test (cond (notp
+                       (%coerce-callable-to-fun test-not))
+                      (testp
+                       (%coerce-callable-to-fun test)))))
+
+     ,@body))
+
 (defconstant +list-based-union-limit+ 80)
 
-(defun union (list1 list2 &key key (test #'eql testp) (test-not nil notp))
-  #!+sb-doc
+(defun hash-table-test-p (fun)
+  (or (eq fun #'eq)
+      (eq fun #'eql)
+      (eq fun #'equal)
+      (eq fun #'equalp)
+      (eq fun 'eq)
+      (eq fun 'eql)
+      (eq fun 'equal)
+      (eq fun 'equalp)))
+
+(defun union (list1 list2 &key key (test nil testp) (test-not nil notp))
   "Return the union of LIST1 and LIST2."
-  (declare (inline member))
+  (declare (explicit-check))
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
   ;; We have two possibilities here: for shortish lists we pick up the
   ;; shorter one as the result, and add the other one to it. For long
   ;; lists we use a hash-table when possible.
   (let ((n1 (length list1))
-        (n2 (length list2))
-        (key (and key (%coerce-callable-to-fun key)))
-        (test (if notp
-                  (let ((test-not-fun (%coerce-callable-to-fun test-not)))
-                    (lambda (x y) (not (funcall test-not-fun x y))))
-                  (%coerce-callable-to-fun test))))
+        (n2 (length list2)))
     (multiple-value-bind (short long n-short)
         (if (< n1 n2)
             (values list1 list2 n1)
             (values list2 list1 n2))
       (if (or (< n-short +list-based-union-limit+)
-              (not (member test (list #'eq #'eql #'equal #'equalp))))
-          (let ((orig short))
-            (dolist (elt long)
-              (unless (member (apply-key key elt) orig :key key :test test)
-                (push elt short)))
-            short)
-          (let ((table (make-hash-table :test test :size (+ n1 n2)))
+              notp
+              (and testp
+                   (not (hash-table-test-p test))))
+          (with-member-test (member-test)
+            (let ((orig short))
+              (dolist (elt long)
+                (unless (funcall member-test elt orig key test)
+                  (push elt short)))
+              short))
+          (let ((table (make-hash-table :test (if testp
+                                                  test
+                                                  #'eql) :size (+ n1 n2)))
+                (key (and key (%coerce-callable-to-fun key)))
                 (union nil))
             (dolist (elt long)
               (setf (gethash (apply-key key elt) table) elt))
@@ -912,46 +953,36 @@
                      table)
             union)))))
 
-;;; Destination and source are SETF-able and many-evaluable. Set the
-;;; SOURCE to the CDR, and "cons" the 1st elt of source to DESTINATION.
-;;;
-;;; FIXME: needs a more mnemonic name
-(defmacro steve-splice (source destination)
-  `(let ((temp ,source))
-     (setf ,source (cdr ,source)
-           (cdr temp) ,destination
-           ,destination temp)))
-
-(defun nunion (list1 list2 &key key (test #'eql testp) (test-not nil notp))
-  #!+sb-doc
+(defun nunion (list1 list2 &key key (test nil testp) (test-not nil notp))
   "Destructively return the union of LIST1 and LIST2."
-  (declare (inline member))
+  (declare (explicit-check))
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
   ;; We have two possibilities here: for shortish lists we pick up the
   ;; shorter one as the result, and add the other one to it. For long
   ;; lists we use a hash-table when possible.
   (let ((n1 (length list1))
-        (n2 (length list2))
-        (key (and key (%coerce-callable-to-fun key)))
-        (test (if notp
-                  (let ((test-not-fun (%coerce-callable-to-fun test-not)))
-                    (lambda (x y) (not (funcall test-not-fun x y))))
-                  (%coerce-callable-to-fun test))))
+        (n2 (length list2)))
     (multiple-value-bind (short long n-short)
         (if (< n1 n2)
             (values list1 list2 n1)
             (values list2 list1 n2))
       (if (or (< n-short +list-based-union-limit+)
-              (not (member test (list #'eq #'eql #'equal #'equalp))))
-          (let ((orig short))
-            (do ((elt (car long) (car long)))
+              notp
+              (and testp
+                   (not (hash-table-test-p test))))
+          (with-member-test (member-test)
+            (do ((orig short)
+                 (elt (car long) (car long)))
                 ((endp long))
-              (if (not (member (apply-key key elt) orig :key key :test test))
-                  (steve-splice long short)
-                  (setf long (cdr long))))
+              (if (funcall member-test elt orig key test)
+                  (pop long)
+                  (shiftf long (cdr long) short long)))
             short)
-          (let ((table (make-hash-table :test test :size (+ n1 n2))))
+          (let ((table (make-hash-table :test (if testp
+                                                  test
+                                                  #'eql) :size (+ n1 n2)))
+                (key (and key (%coerce-callable-to-fun key))))
             (dolist (elt long)
               (setf (gethash (apply-key key elt) table) elt))
             (dolist (elt short)
@@ -968,98 +999,89 @@
               union))))))
 
 (defun intersection (list1 list2
-                     &key key (test #'eql testp) (test-not nil notp))
-  #!+sb-doc
+                     &key key (test nil testp) (test-not nil notp))
   "Return the intersection of LIST1 and LIST2."
-  (declare (inline member))
+  (declare (explicit-check))
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
-  (let ((key (and key (%coerce-callable-to-fun key))))
-    (let ((res nil))
-      (dolist (elt list1)
-        (if (with-set-keys (member (apply-key key elt) list2))
+  (when (and list1 list2)
+    (with-member-test (member-test)
+      (let ((res nil))
+        (dolist (elt list1)
+          (when (funcall member-test elt list2 key test)
             (push elt res)))
-      res)))
+        res))))
 
 (defun nintersection (list1 list2
-                      &key key (test #'eql testp) (test-not nil notp))
-  #!+sb-doc
+                      &key key (test nil testp) (test-not nil notp))
   "Destructively return the intersection of LIST1 and LIST2."
-  (declare (inline member))
+  (declare (explicit-check))
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
-  (let ((key (and key (%coerce-callable-to-fun key))))
-    (let ((res nil)
-          (list1 list1))
-      (do () ((endp list1))
-        (if (with-set-keys (member (apply-key key (car list1)) list2))
-            (steve-splice list1 res)
-            (setq list1 (cdr list1))))
-      res)))
+  (when (and list1 list2)
+    (with-member-test (member-test)
+      (let ((res nil)
+            (list1 list1))
+        (do () ((endp list1))
+          (if (funcall member-test (car list1) list2 key test)
+              (shiftf list1 (cdr list1) res list1)
+              (setf list1 (cdr list1))))
+        res))))
 
 (defun set-difference (list1 list2
-                       &key key (test #'eql testp) (test-not nil notp))
-  #!+sb-doc
+                       &key key (test nil testp) (test-not nil notp))
   "Return the elements of LIST1 which are not in LIST2."
-  (declare (inline member))
+  (declare (explicit-check))
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
-  (let ((key (and key (%coerce-callable-to-fun key))))
-    (if (null list2)
-        list1
+  (if list2
+      (with-member-test (member-test)
         (let ((res nil))
           (dolist (elt list1)
-            (if (not (with-set-keys (member (apply-key key elt) list2)))
-                (push elt res)))
-          res))))
+            (unless (funcall member-test elt list2 key test)
+              (push elt res)))
+          res))
+      list1))
 
 (defun nset-difference (list1 list2
-                        &key key (test #'eql testp) (test-not nil notp))
-  #!+sb-doc
+                        &key key (test nil testp) (test-not nil notp))
   "Destructively return the elements of LIST1 which are not in LIST2."
-  (declare (inline member))
+  (declare (explicit-check))
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
-  (let ((key (and key (%coerce-callable-to-fun key))))
-    (let ((res nil)
-          (list1 list1))
-      (do () ((endp list1))
-        (if (not (with-set-keys (member (apply-key key (car list1)) list2)))
-            (steve-splice list1 res)
-            (setq list1 (cdr list1))))
-      res)))
+  (if list2
+      (with-member-test (member-test)
+        (let ((res nil)
+              (list1 list1))
+          (do () ((endp list1))
+            (if (funcall member-test (car list1) list2 key test)
+                (setf list1 (cdr list1))
+                (shiftf list1 (cdr list1) res list1)))
+          res))
+      list1))
 
 (defun set-exclusive-or (list1 list2
-                         &key key (test #'eql testp) (test-not #'eql notp))
-  #!+sb-doc
+                         &key key (test nil testp) (test-not nil notp))
   "Return new list of elements appearing exactly once in LIST1 and LIST2."
-  (declare (inline member))
+  (declare (explicit-check))
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
-  (let ((result nil)
-        (key (and key (%coerce-callable-to-fun key)))
-        (test (if testp (%coerce-callable-to-fun test) test))
-        (test-not (if notp (%coerce-callable-to-fun test-not) test-not)))
-    (declare (type function test test-not))
-    (dolist (elt list1)
-      (unless (with-set-keys (member (apply-key key elt) list2))
-        (setq result (cons elt result))))
-    (let ((test (if testp
-                    (lambda (x y) (funcall test y x))
-                    test))
-          (test-not (if notp
-                        (lambda (x y) (funcall test-not y x))
-                        test-not)))
-      (dolist (elt list2)
-        (unless (with-set-keys (member (apply-key key elt) list1))
-          (setq result (cons elt result)))))
+  (let ((result nil))
+    (with-member-test (member-test)
+      (dolist (elt list1)
+        (unless (funcall member-test elt list2 key test)
+          (push elt result)))
+      (dx-flet ((test (x y) (funcall (truly-the function test) y x)))
+        (dolist (elt list2)
+          (unless (funcall member-test elt list1 key #'test)
+            (push elt result)))))
     result))
 
 (defun nset-exclusive-or (list1 list2
                           &key key (test #'eql testp) (test-not #'eql notp))
-  #!+sb-doc
   "Destructively return a list with elements which appear but once in LIST1
    and LIST2."
+  (declare (explicit-check))
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
   (let ((key (and key (%coerce-callable-to-fun key)))
@@ -1116,26 +1138,23 @@
             (setq splicex x))))))
 
 (defun subsetp (list1 list2 &key key (test #'eql testp) (test-not nil notp))
-  #!+sb-doc
   "Return T if every element in LIST1 is also in LIST2."
-  (declare (inline member))
+  (declare (explicit-check))
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
-  (let ((key (and key (%coerce-callable-to-fun key))))
+  (with-member-test (member-test)
     (dolist (elt list1)
-      (unless (with-set-keys (member (apply-key key elt) list2))
+      (unless (funcall member-test elt list2 key test)
         (return-from subsetp nil)))
     t))
 
 ;;;; functions that operate on association lists
 
 (defun acons (key datum alist)
-  #!+sb-doc
   "Construct a new alist by adding the pair (KEY . DATUM) to ALIST."
   (cons (cons key datum) alist))
 
 (defun pairlis (keys data &optional (alist '()))
-  #!+sb-doc
   "Construct an association list from KEYS and DATA (adding to ALIST)."
   (do ((x keys (cdr x))
        (y data (cdr y)))
@@ -1145,9 +1164,9 @@
     (setq alist (acons (car x) (car y) alist))))
 
 (defun assoc (item alist &key key (test nil testp) (test-not nil notp))
-  #!+sb-doc
   "Return the cons in ALIST whose car is equal (by a given test or EQL) to
    the ITEM."
+  (declare (explicit-check))
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
   (let ((key (and key (%coerce-callable-to-fun key)))
@@ -1167,9 +1186,9 @@
                (%assoc item alist))))))
 
 (defun assoc-if (predicate alist &key key)
-  #!+sb-doc
   "Return the first cons in ALIST whose CAR satisfies PREDICATE. If
    KEY is supplied, apply it to the CAR of each cons before testing."
+  (declare (explicit-check))
   (let ((predicate (%coerce-callable-to-fun predicate))
         (key (and key (%coerce-callable-to-fun key))))
     (if key
@@ -1177,9 +1196,9 @@
         (%assoc-if predicate alist))))
 
 (defun assoc-if-not (predicate alist &key key)
-  #!+sb-doc
   "Return the first cons in ALIST whose CAR does not satisfy PREDICATE.
   If KEY is supplied, apply it to the CAR of each cons before testing."
+  (declare (explicit-check))
   (let ((predicate (%coerce-callable-to-fun predicate))
         (key (and key (%coerce-callable-to-fun key))))
     (if key
@@ -1187,10 +1206,9 @@
         (%assoc-if-not predicate alist))))
 
 (defun rassoc (item alist &key key (test nil testp) (test-not nil notp))
-  (declare (list alist))
-  #!+sb-doc
   "Return the cons in ALIST whose CDR is equal (by a given test or EQL) to
    the ITEM."
+  (declare (explicit-check))
   (when (and testp notp)
     (error ":TEST and :TEST-NOT were both supplied."))
   (let ((key (and key (%coerce-callable-to-fun key)))
@@ -1210,9 +1228,9 @@
                (%rassoc item alist))))))
 
 (defun rassoc-if (predicate alist &key key)
-  #!+sb-doc
   "Return the first cons in ALIST whose CDR satisfies PREDICATE. If KEY
   is supplied, apply it to the CDR of each cons before testing."
+  (declare (explicit-check))
   (let ((predicate (%coerce-callable-to-fun predicate))
         (key (and key (%coerce-callable-to-fun key))))
     (if key
@@ -1220,9 +1238,9 @@
         (%rassoc-if predicate alist))))
 
 (defun rassoc-if-not (predicate alist &key key)
-  #!+sb-doc
   "Return the first cons in ALIST whose CDR does not satisfy PREDICATE.
   If KEY is supplied, apply it to the CDR of each cons before testing."
+  (declare (explicit-check))
   (let ((predicate (%coerce-callable-to-fun predicate))
         (key (and key (%coerce-callable-to-fun key))))
     (if key
@@ -1270,35 +1288,35 @@
                    temp (cdr temp))))))
 
 (defun mapc (function list &rest more-lists)
-  #!+sb-doc
   "Apply FUNCTION to successive elements of lists. Return the second argument."
+  (declare (explicit-check))
   (map1 function (cons list more-lists) nil t))
 
 (defun mapcar (function list &rest more-lists)
-  #!+sb-doc
   "Apply FUNCTION to successive elements of LIST. Return list of FUNCTION
    return values."
+  (declare (explicit-check))
   (map1 function (cons list more-lists) :list t))
 
 (defun mapcan (function list &rest more-lists)
-  #!+sb-doc
   "Apply FUNCTION to successive elements of LIST. Return NCONC of FUNCTION
    results."
+  (declare (explicit-check))
   (map1 function (cons list more-lists) :nconc t))
 
 (defun mapl (function list &rest more-lists)
-  #!+sb-doc
   "Apply FUNCTION to successive CDRs of list. Return NIL."
+  (declare (explicit-check))
   (map1 function (cons list more-lists) nil nil))
 
 (defun maplist (function list &rest more-lists)
-  #!+sb-doc
   "Apply FUNCTION to successive CDRs of list. Return list of results."
+  (declare (explicit-check))
   (map1 function (cons list more-lists) :list nil))
 
 (defun mapcon (function list &rest more-lists)
-  #!+sb-doc
   "Apply FUNCTION to successive CDRs of lists. Return NCONC of results."
+  (declare (explicit-check))
   (map1 function (cons list more-lists) :nconc nil))
 
 ;;;; Specialized versions
