@@ -113,9 +113,10 @@
            (type fixnum item)
            (optimize speed))
   (labels ((recurse (start end)
-             (declare (type index start end))
+             (declare (type index start end)
+                      (optimize (safety 0)))
              (when (< start end)
-               (let* ((i (+ start (truncate (- end start) 2)))
+               (let* ((i (+ start (truncate (the index (- end start)) 2)))
                       (index (* 2 i))
                       (elt1 (svref vector index))
                       (elt2 (svref vector (1+ index))))
@@ -151,7 +152,6 @@
     (recurse 0 (truncate (length vector) 2))))
 
 (defun proplist-p (character property)
-  #!+sb-doc
   "Returns T if CHARACTER has the specified PROPERTY.
 PROPERTY is a keyword representing one of the properties from PropList.txt,
 with underscores replaced by dashes."
@@ -272,12 +272,10 @@ with underscores replaced by dashes."
        (svref vector index)))
 
 (defun general-category (character)
-  #!+sb-doc
   "Returns the general category of CHARACTER as it appears in UnicodeData.txt"
   (svref-or-null *general-categories* (sb!impl::ucd-general-category character)))
 
 (defun bidi-class (character)
-  #!+sb-doc
   "Returns the bidirectional class of CHARACTER"
   (if (and (eql (general-category character) :Cn)
            (default-ignorable-p character))
@@ -288,12 +286,10 @@ with underscores replaced by dashes."
 
 (declaim (inline combining-class))
 (defun combining-class (character)
-  #!+sb-doc
   "Returns the canonical combining class (CCC) of CHARACTER"
   (aref **character-misc-database** (+ 2 (misc-index character))))
 
 (defun decimal-value (character)
-  #!+sb-doc
   "Returns the decimal digit value associated with CHARACTER or NIL if
 there is no such value.
 
@@ -304,7 +300,6 @@ Because of this, `(decimal-digit c) <=> (digit-char-p c 10)` in
   (sb!impl::ucd-decimal-digit character))
 
 (defun digit-value (character)
-  #!+sb-doc
   "Returns the Unicode digit value of CHARACTER or NIL if it doesn't exist.
 
 Digit values are guaranteed to be integers between 0 and 9 inclusive.
@@ -317,7 +312,6 @@ that have a digit value but no decimal digit value"
     (if (< %digit 10) %digit nil)))
 
 (defun numeric-value (character)
-  #!+sb-doc
   "Returns the numeric value of CHARACTER or NIL if there is no such value.
 Numeric value is the most general of the Unicode numeric properties.
 The only constraint on the numeric value is that it be a rational number."
@@ -326,14 +320,12 @@ The only constraint on the numeric value is that it be a rational number."
       (digit-value character)))
 
 (defun mirrored-p (character)
-  #!+sb-doc
   "Returns T if CHARACTER needs to be mirrored in bidirectional text.
 Otherwise, returns NIL."
   (logbitp 5 (aref **character-misc-database**
                     (+ 5 (misc-index character)))))
 
 (defun bidi-mirroring-glyph (character)
-  #!+sb-doc
   "Returns the mirror image of CHARACTER if it exists.
 Otherwise, returns NIL."
   (when (mirrored-p character)
@@ -341,7 +333,6 @@ Otherwise, returns NIL."
       (when ret (code-char ret)))))
 
 (defun east-asian-width (character)
-  #!+sb-doc
   "Returns the East Asian Width property of CHARACTER as
 one of the keywords :N (Narrow), :A (Ambiguous), :H (Halfwidth),
 :W (Wide), :F (Fullwidth), or :NA (Not applicable)"
@@ -351,14 +342,12 @@ one of the keywords :N (Narrow), :A (Ambiguous), :H (Halfwidth),
                             (+ 5 (misc-index character))))))
 
 (defun script (character)
-  #!+sb-doc
   "Returns the Script property of CHARACTER as a keyword.
 If CHARACTER does not have a known script, returns :UNKNOWN"
   (svref-or-null *scripts*
                  (aref **character-misc-database** (+ 6 (misc-index character)))))
 
 (defun char-block (character)
-  #!+sb-doc
   "Returns the Unicode block in which CHARACTER resides as a keyword.
 If CHARACTER does not have a known block, returns :NO-BLOCK"
   (let* ((code (char-code character))
@@ -367,7 +356,6 @@ If CHARACTER does not have a known block, returns :NO-BLOCK"
         (aref *blocks* block-index) :no-block)))
 
 (defun unicode-1-name (character)
-  #!+sb-doc
   "Returns the name assigned to CHARACTER in Unicode 1.0 if it is distinct
 from the name currently assigned to CHARACTER. Otherwise, returns NIL.
 This property has been officially obsoleted by the Unicode standard, and
@@ -379,7 +367,6 @@ is only included for backwards compatibility."
       (huffman-decode h-code **unicode-character-name-huffman-tree**))))
 
 (defun age (character)
-  #!+sb-doc
   "Returns the version of Unicode in which CHARACTER was assigned as a pair
 of values, both integers, representing the major and minor version respectively.
 If CHARACTER is not assigned in Unicode, returns NIL for both values."
@@ -389,7 +376,6 @@ If CHARACTER is not assigned in Unicode, returns NIL for both values."
     (if (zerop value) (values nil nil) (values major minor))))
 
 (defun hangul-syllable-type (character)
-  #!+sb-doc
   "Returns the Hangul syllable type of CHARACTER.
 The syllable type can be one of :L, :V, :T, :LV, or :LVT.
 If the character is not a Hangul syllable or Jamo, returns NIL"
@@ -408,7 +394,6 @@ If the character is not a Hangul syllable or Jamo, returns NIL"
        (if (= 0 (rem (- cp #xac00) 28)) :LV :LVT)))))
 
 (defun line-break-class (character &key resolve)
-  #!+sb-doc
   "Returns the line breaking class of CHARACTER, as specified in UAX #14.
 If :RESOLVE is NIL, returns the character class found in the property file.
 If :RESOLVE is non-NIL, centain line-breaking classes will be mapped to othec
@@ -443,23 +428,19 @@ appear in an SBCL string. The line-breaking behavior of surrogates is undefined.
     raw-class))
 
 (defun uppercase-p (character)
-  #!+sb-doc
   "Returns T if CHARACTER has the Unicode property Uppercase and NIL otherwise"
   (or (eql (general-category character) :Lu) (proplist-p character :other-uppercase)))
 
 (defun lowercase-p (character)
-  #!+sb-doc
   "Returns T if CHARACTER has the Unicode property Lowercase and NIL otherwise"
   (or (eql (general-category character) :Ll) (proplist-p character :other-lowercase)))
 
 (defun cased-p (character)
-  #!+sb-doc
   "Returns T if CHARACTER has a (Unicode) case, and NIL otherwise"
   (or (uppercase-p character) (lowercase-p character)
       (eql (general-category character) :Lt)))
 
 (defun case-ignorable-p (character)
-  #!+sb-doc
   "Returns T if CHARACTER is Case Ignorable as defined in Unicode 6.3, Chapter
 3"
   (or (member (general-category character)
@@ -468,45 +449,38 @@ appear in an SBCL string. The line-breaking behavior of surrogates is undefined.
               '(:midletter :midnumlet :single-quote))))
 
 (defun alphabetic-p (character)
-  #!+sb-doc
   "Returns T if CHARACTER is Alphabetic according to the Unicode standard
 and NIL otherwise"
   (or (member (general-category character) '(:Lu :Ll :Lt :Lm :Lo :Nl))
       (proplist-p character :other-alphabetic)))
 
 (defun ideographic-p (character)
-  #!+sb-doc
   "Returns T if CHARACTER has the Unicode property Ideographic,
 which loosely corresponds to the set of \"Chinese characters\""
   (proplist-p character :ideographic))
 
 (defun math-p (character)
-  #!+sb-doc
   "Returns T if CHARACTER is a mathematical symbol according to Unicode and
 NIL otherwise"
   (or (eql (general-category character) :sm) (proplist-p character :other-math)))
 
 (defun whitespace-p (character)
-  #!+sb-doc
   "Returns T if CHARACTER is whitespace according to Unicode
 and NIL otherwise"
   (proplist-p character :white-space))
 
 (defun hex-digit-p (character &key ascii)
-  #!+sb-doc
   "Returns T if CHARACTER is a hexadecimal digit and NIL otherwise.
 If :ASCII is non-NIL, fullwidth equivalents of the Latin letters A through F
 are excluded."
   (proplist-p character (if ascii :ascii-hex-digit :hex-digit)))
 
 (defun soft-dotted-p (character)
-  #!+sb-doc
   "Returns T if CHARACTER has a soft dot (such as the dots on i and j) which
 disappears when accents are placed on top of it. and NIL otherwise"
   (proplist-p character :soft-dotted))
 
 (defun default-ignorable-p (character)
-  #!+sb-doc
   "Returns T if CHARACTER is a Default_Ignorable_Code_Point"
   (and
    (or (proplist-p character :other-default-ignorable-code-point)
@@ -595,15 +569,15 @@ disappears when accents are placed on top of it. and NIL otherwise"
                                 (t
                                  (push char chars)
                                  (setf previous-combining-class combining-class))))))
-      (sb!kernel:with-array-data ((string string) (start) (end))
-        (declare (ignore start))
+      (sb!kernel:with-array-data ((string string) (start) (end)
+                                  :check-fill-pointer t)
         (let ((calback (if filter
                            (let ((filter (sb!kernel:%coerce-callable-to-fun filter)))
                              (lambda (char)
                                (when (funcall filter char)
                                  (callback char))))
                            #'callback)))
-          (loop for i below end
+          (loop for i from start below end
                 for char = (schar string i)
                 do
                 (decompose-char char compatibility calback))))
@@ -736,7 +710,6 @@ disappears when accents are placed on top of it. and NIL otherwise"
 
 (defun normalize-string (string &optional (form :nfd)
                                           filter)
-  #!+sb-doc
   "Normalize STRING to the Unicode normalization form form.
 Acceptable values for form are :NFD, :NFC, :NFKD, and :NFKC.
 If FILTER is a function it is called on each decomposed character and
@@ -767,7 +740,6 @@ only characters for which it returns T are collected."
     ((array nil (*)) string)))
 
 (defun normalized-p (string &optional (form :nfd))
-  #!+sb-doc
   "Tests if STRING is normalized to FORM"
   ;; FIXME: can be optimized
   (string= string (normalize-string string form)))
@@ -859,7 +831,6 @@ only characters for which it returns T are collected."
 
 
 (defun uppercase (string &key locale)
-  #!+sb-doc
   "Returns the full uppercase of STRING according to the Unicode standard.
 The result is not guaranteed to have the same length as the input. If :LOCALE
 is NIL, no language-specific case transformations are applied. If :LOCALE is a
@@ -889,7 +860,6 @@ Win32 only)."
          (t nil)))))
 
 (defun lowercase (string &key locale)
-  #!+sb-doc
   "Returns the full lowercase of STRING according to the Unicode standard.
 The result is not guaranteed to have the same length as the input.
 :LOCALE has the same semantics as the :LOCALE argument to UPPERCASE."
@@ -958,7 +928,6 @@ The result is not guaranteed to have the same length as the input.
        (t nil)))))
 
 (defun titlecase (string &key locale)
-  #!+sb-doc
   "Returns the titlecase of STRING. The resulting string can
 be longer than the input.
 :LOCALE has the same semantics as the :LOCALE argument to UPPERCASE."
@@ -989,9 +958,8 @@ be longer than the input.
    (apply #'concatenate 'string (nreverse cased))))
 
 (defun casefold (string)
-  #!+sb-doc
   "Returns the full casefolding of STRING according to the Unicode standard.
-Casefolding remove case information in a way that allaws the results to be used
+Casefolding removes case information in a way that allows the results to be used
 for case-insensitive comparisons.
 The result is not guaranteed to have the same length as the input."
   (string-somethingcase #'char-foldcase string (constantly nil)))
@@ -1006,7 +974,6 @@ The result is not guaranteed to have the same length as the input."
 ;; Word breaking sets this to make their algorithms less tricky
 (defvar *other-break-special-graphemes* nil)
 (defun grapheme-break-class (char)
-  #!+sb-doc
   "Returns the grapheme breaking class of CHARACTER, as specified in UAX #29."
   (let ((cp (when char (char-code char)))
         (gc (when char (general-category char)))
@@ -1038,34 +1005,54 @@ The result is not guaranteed to have the same length as the input."
             (not (binary-search cp not-spacing-mark))) :spacing-mark)
       (t (hangul-syllable-type char)))))
 
+(macrolet ((def (name extendedp)
+             `(defun ,name (function string)
+                (do ((length (length string))
+                     (start 0)
+                     (end 1 (1+ end))
+                     (c1 nil)
+                     (c2 (and (> (length string) 0) (grapheme-break-class (char string 0)))))
+                    ((>= end length)
+                     (if (= end length) (progn (funcall function string start end) nil)))
+                  (flet ((brk () (funcall function string start end) (setf start end)))
+                    (declare (truly-dynamic-extent #'brk))
+                    (shiftf c1 c2 (grapheme-break-class (char string end)))
+                    (cond
+                      ((and (eql c1 :cr) (eql c2 :lf)))
+                      ((or (member c1 '(:control :cr :lf))
+                           (member c2 '(:control :cr :lf)))
+                       (brk))
+                      ((or (and (eql c1 :l) (member c2 '(:l :v :lv :lvt)))
+                           (and (or (eql c1 :v) (eql c1 :lv))
+                                (or (eql c2 :v) (eql c2 :t)))
+                           (and (eql c2 :t) (or (eql c1 :lvt) (eql c1 :t)))))
+                      ((and (eql c1 :regional-indicator) (eql c2 :regional-indicator)))
+                      ((eql c2 :extend))
+                      ,@(when extendedp
+                              `(((or (eql c2 :spacing-mark) (eql c1 :prepend)))))
+                      (t (brk))))))))
+  (def map-legacy-grapheme-boundaries nil)
+  (def map-grapheme-boundaries t))
+
+(macrolet ((def (name mapper)
+             `(defun ,name (function string)
+                (let ((array (make-array 0 :element-type (array-element-type string) :adjustable t :displaced-to string)))
+                  (flet ((fun (string start end)
+                           (declare (type string string))
+                           (funcall function (adjust-array array (- end start) :displaced-to string :displaced-index-offset start))))
+                    (declare (truly-dynamic-extent #'fun))
+                    (,mapper #'fun string))))))
+  (def map-legacy-graphemes map-legacy-grapheme-boundaries)
+  (def map-graphemes map-grapheme-boundaries))
+
 (defun graphemes (string)
-  #!+sb-doc
   "Breaks STRING into graphemes acording to the default
 grapheme breaking rules specified in UAX #29, returning a list of strings."
-  (let* ((chars (coerce string 'list)) clusters (cluster (list (car chars))))
-    (do ((first (car chars) second)
-         (tail (cdr chars) (when tail (cdr tail)))
-         (second (cadr chars) (when tail (cadr tail))))
-        ((not first) (nreverse (mapcar #'(lambda (l) (coerce l 'string)) clusters)))
-      (flet ((brk () (push (nreverse cluster) clusters) (setf cluster (list second)))
-             (nobrk () (push second cluster)))
-        (let ((c1 (grapheme-break-class first))
-              (c2 (grapheme-break-class second)))
-          (cond
-            ((and (eql c1 :cr) (eql c2 :lf)) (nobrk))
-            ((or (member c1 '(:control :cr :lf))
-                 (member c2 '(:control :cr :lf))) (brk))
-             ((or (and (eql c1 :l) (member c2 '(:l :v :lv :lvt)))
-                  (and (or (eql c1 :v) (eql c1 :lv))
-                       (or (eql c2 :v) (eql c2 :t)))
-                  (and (eql c2 :t) (or (eql c1 :lvt) (eql c1 :t))))
-              (nobrk))
-             ((and (eql c1 :regional-indicator) (eql c2 :regional-indicator)) (nobrk))
-             ((or (eql c2 :extend) (eql c2 :spacing-mark) (eql c1 :prepend)) (nobrk))
-             (t (brk))))))))
+  (let (result)
+    (map-graphemes (lambda (a) (push (subseq a 0) result)) string)
+    (nreverse result)))
 
 (defun word-break-class (char)
-  #!+sb-doc
   "Returns the word breaking class of CHARACTER, as specified in UAX #29."
   ;; Words use graphemes as characters to deal with the ignore rule
   (when (listp char) (setf char (car char)))
@@ -1118,7 +1105,6 @@ grapheme breaking rules specified in UAX #29, returning a list of strings."
            (push ,%thing ,list)))))
 
 (defun words (string)
-  #!+sb-doc
   "Breaks STRING into words acording to the default
 word breaking rules specified in UAX #29. Returns a list of strings"
   (let ((chars (mapcar
@@ -1172,7 +1158,6 @@ word breaking rules specified in UAX #29. Returns a list of strings"
             (t (brk))))))))
 
 (defun sentence-break-class (char)
-  #!+sb-doc
   "Returns the sentence breaking class of CHARACTER, as specified in UAX #29."
   (when (listp char) (setf char (car char)))
   (let ((cp (when char (char-code char)))
@@ -1205,7 +1190,6 @@ word breaking rules specified in UAX #29. Returns a list of strings"
       (t nil))))
 
 (defun sentence-prebreak (string)
-  #!+sb-doc
   "Pre-combines some sequences of characters to make the sentence-break
 algorithm simpler..
 Specifically,
@@ -1241,7 +1225,6 @@ Specifically,
     (flush) (nreverse clusters))))
 
 (defun sentences (string)
-  #!+sb-doc
   "Breaks STRING into sentences acording to the default
 sentence breaking rules specified in UAX #29"
   (let ((special-handling '(:close :sp :sep :cr :lf :scontinue :sterm :aterm))
@@ -1459,9 +1442,8 @@ sentence breaking rules specified in UAX #29"
     (values list tail)))
 
 (defun lines (string &key (margin *print-right-margin*))
-  #!+sb-doc
   "Breaks STRING into lines that are no wider than :MARGIN according to the
-line breaking rules outlined in UAX #14. Combining marks will awsays be kept
+line breaking rules outlined in UAX #14. Combining marks will always be kept
 together with their base characters, and spaces (but not other types of
 whitespace) will be removed from the end of lines. If :MARGIN is unspecified,
 it defaults to 80 characters"
@@ -1638,7 +1620,6 @@ it defaults to 80 characters"
   (< (length vector1) (length vector2)))
 
 (defun unicode= (string1 string2 &key (start1 0) end1 (start2 0) end2 (strict t))
-  #!+sb-doc
   "Determines whether STRING1 and STRING2 are canonically equivalent according
 to Unicode. The START and END arguments behave like the arguments to STRING=.
 If :STRICT is NIL, UNICODE= tests compatibility equavalence instead."
@@ -1647,7 +1628,6 @@ If :STRICT is NIL, UNICODE= tests compatibility equavalence instead."
     (string= str1 str2)))
 
 (defun unicode-equal (string1 string2 &key (start1 0) end1 (start2 0) end2 (strict t))
-    #!+sb-doc
   "Determines whether STRING1 and STRING2 are canonically equivalent after
 casefoldin8 (that is, ignoring case differences) according to Unicode. The
 START and END arguments behave like the arguments to STRING=. If :STRICT is
@@ -1659,7 +1639,6 @@ NIL, UNICODE= tests compatibility equavalence instead."
      (normalize-string (casefold str2) (if strict :nfd :nfkd)))))
 
 (defun unicode< (string1 string2 &key (start1 0) end1 (start2 0) end2)
-  #!+sb-doc
   "Determines whether STRING1 sorts before STRING2 using the Unicode Collation
 Algorithm, The function uses an untailored Default Unicode Collation Element Table
 to produce the sort keys. The function uses the Shifted method for dealing
@@ -1672,7 +1651,6 @@ with variable-weight characters, as described in UTS #10"
         (vector< k1 k2))))
 
 (defun unicode<= (string1 string2 &key (start1 0) end1 (start2 0) end2)
-  #!+sb-doc
   "Tests if STRING1 and STRING2 are either UNICODE< or UNICODE="
   (or
    (unicode= string1 string2 :start1 start1 :end1 end1
@@ -1681,13 +1659,11 @@ with variable-weight characters, as described in UTS #10"
              :start2 start2 :end2 end2)))
 
 (defun unicode> (string1 string2 &key (start1 0) end1 (start2 0) end2)
-  #!+sb-doc
   "Tests if STRING2 is UNICODE< STRING1."
    (unicode< string2 string1 :start1 start2 :end1 end2
              :start2 start1 :end2 end1))
 
 (defun unicode>= (string1 string2 &key (start1 0) end1 (start2 0) end2)
-  #!+sb-doc
   "Tests if STRING1 and STRING2 are either UNICODE= or UNICODE>"
   (or
    (unicode= string1 string2 :start1 start1 :end1 end1
@@ -1715,7 +1691,6 @@ with variable-weight characters, as described in UTS #10"
     (apply #'concatenate 'string (nreverse ret))))
 
 (defun confusable-p (string1 string2 &key (start1 0) end1 (start2 0) end2)
-  #!+sb-doc
   "Determines whether STRING1 and STRING2 could be visually confusable
 according to the IDNA confusableSummary.txt table"
     (let* ((form #!+sb-unicode :nfd #!-sb-unicode :nfc)
