@@ -91,7 +91,6 @@
 (defmacro with-restarted-syscall ((&optional (value (gensym))
                                              (errno (gensym)))
                                   syscall-form &rest body)
-  #!+sb-doc
   "Evaluate BODY with VALUE and ERRNO bound to the return values of
 SYSCALL-FORM. Repeat evaluation of SYSCALL-FORM if it is interrupted."
   `(let (,value ,errno)
@@ -112,7 +111,6 @@ SYSCALL-FORM. Repeat evaluation of SYSCALL-FORM if it is interrupted."
 
 #!-win32
 (define-alien-routine ("getenv" posix-getenv) c-string
-  #!+sb-doc
   "Return the \"value\" part of the environment string \"name=value\" which
 corresponds to NAME, or NIL if there is none."
   (name (c-string :not-null t)))
@@ -158,17 +156,18 @@ corresponds to NAME, or NIL if there is none."
 ;;; If the O_CREAT flag is specified, then the file is created with a
 ;;; permission of argument MODE if the file doesn't exist. An integer
 ;;; file descriptor is returned by UNIX-OPEN.
-(defun unix-open (path flags mode)
+(defun unix-open (path flags mode &key #!+win32 overlapped)
   (declare (type unix-pathname path)
            (type fixnum flags)
-           (type unix-file-mode mode))
-  #!+win32 (sb!win32:unixlike-open path flags mode)
+           (type unix-file-mode mode)
+           #!+win32
+           (ignore mode))
+  #!+win32 (sb!win32:unixlike-open path flags :overlapped overlapped)
   #!-win32
   (with-restarted-syscall (value errno)
     (int-syscall ("open" c-string int int)
                  path
-                 (logior #!+win32 o_binary
-                         #!+largefile o_largefile
+                 (logior #!+largefile o_largefile
                          flags)
                  mode)))
 
@@ -274,7 +273,6 @@ corresponds to NAME, or NIL if there is none."
   #!+win32 (sb!win32::windows-isatty fd))
 
 (defun unix-lseek (fd offset whence)
-  #!+sb-doc
   "Unix-lseek accepts a file descriptor and moves the file pointer by
    OFFSET octets.  Whence can be any of the following:
 
@@ -426,7 +424,6 @@ corresponds to NAME, or NIL if there is none."
 (deftype exit-code ()
   `(signed-byte 32))
 (defun os-exit (code &key abort)
-  #!+sb-doc
   "Exit the process with CODE. If ABORT is true, exit is performed using _exit(2),
 avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
   (unless (typep code 'exit-code)
@@ -717,6 +714,7 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
 ;;; they are ready for reading and writing. See the UNIX Programmer's
 ;;; Manual for more information.
 (defun unix-select (nfds rdfds wrfds xpfds to-secs &optional (to-usecs 0))
+  (declare (muffle-conditions t))
   (declare (type integer nfds)
            (type unsigned-byte rdfds wrfds xpfds)
            (type (or (unsigned-byte 31) null) to-secs)
@@ -901,27 +899,20 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
 #!-win32
 (defun fd-type (fd)
   (declare (type unix-fd fd))
-  (let ((fmt (logand
-              s-ifmt
-              (or (with-alien ((buf (struct wrapped_stat)))
+  (let ((mode (or (with-alien ((buf (struct wrapped_stat)))
                     (syscall ("fstat_wrapper" int (* (struct wrapped_stat)))
                              (slot buf 'st-mode)
                              fd (addr buf)))
-                  0))))
-    (cond ((logtest s-ififo fmt)
-           :fifo)
-          ((logtest s-ifchr fmt)
-           :character)
-          ((logtest s-ifdir fmt)
-           :directory)
-          ((logtest s-ifblk fmt)
-           :block)
-          ((logtest s-ifreg fmt)
-           :regular)
-          ((logtest s-ifsock fmt)
-           :socket)
-          (t
-           :unknown))))
+                  0)))
+    (case (logand mode s-ifmt)
+      (#.s-ifchr :character)
+      (#.s-ifdir :directory)
+      (#.s-ifblk :block)
+      (#.s-ifreg :regular)
+      (#.s-ifsock :socket)
+      (#.s-iflnk :link)
+      (#.s-ififo :fifo)
+      (t :unknown))))
 
 ;;;; time.h
 
@@ -940,53 +931,28 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
             (tm-gmtoff long) ;  Seconds east of UTC.
             (tm-zone c-string))) ; Timezone abbreviation.
 
-(define-alien-routine get-timezone sb!alien:void
-  (when time-t :in)
-  (seconds-west sb!alien:int :out)
-  (daylight-savings-p sb!alien:boolean :out))
-
+(define-alien-routine get-timezone int
+  (when time-t)
+  ;; KLUDGE: the runtime `boolean' is defined as `int', but the alien
+  ;; type is N-WORD-BITS wide.
+  (daylight-savings-p (boolean 32) :out))
 #!-win32
 (defun nanosleep (secs nsecs)
-  (declare (optimize (sb!c:alien-funcall-saves-fp-and-pc 0)))
-  (with-alien ((req (struct timespec))
-               (rem (struct timespec)))
-    (setf (slot req 'tv-sec) secs
-          (slot req 'tv-nsec) nsecs)
-    (loop while (and (eql eintr
-                          (nth-value 1
-                                     (int-syscall ("sb_nanosleep" (* (struct timespec))
-                                                                  (* (struct timespec)))
-                                                  (addr req) (addr rem))))
-                     ;; KLUDGE: On Darwin, if an interrupt cases nanosleep to
-                     ;; take longer than the requested time, the call will
-                     ;; return with EINT and (unsigned)-1 seconds in the
-                     ;; remainder timespec, which would cause us to enter
-                     ;; nanosleep again for ~136 years. So, we check that the
-                     ;; remainder time is actually decreasing.
-                     ;;
-                     ;; It would be neat to do this bit of defensive
-                     ;; programming on all platforms, but unfortunately on
-                     ;; Linux, REM can be a little higher than REQ if the
-                     ;; nanosleep() call is interrupted quickly enough,
-                     ;; probably due to the request being rounded up to the
-                     ;; nearest HZ. This would cause the sleep to return way
-                     ;; too early.
-                     #!+darwin
-                     (let ((rem-sec (slot rem 'tv-sec))
-                           (rem-nsec (slot rem 'tv-nsec)))
-                       (when (or (> secs rem-sec)
-                                 (and (= secs rem-sec) (>= nsecs rem-nsec)))
-                         ;; Update for next round.
-                         (setf secs  rem-sec
-                               nsecs rem-nsec)
-                         t)))
-          do (setf (slot req 'tv-sec) (slot rem 'tv-sec)
-                   (slot req 'tv-nsec) (slot rem 'tv-nsec)))))
+  (alien-funcall (extern-alien "sb_nanosleep" (function int time-t int))
+                 secs nsecs)
+  nil)
 
-(defun unix-get-seconds-west (secs)
-  (multiple-value-bind (ignore seconds dst) (get-timezone secs)
-    (declare (ignore ignore) (ignore dst))
-    (values seconds)))
+#!-win32
+(defun nanosleep-double (seconds)
+  (alien-funcall (extern-alien "sb_nanosleep_double" (function (values) double))
+                 seconds)
+  nil)
+
+#!-win32
+(defun nanosleep-float (seconds)
+  (alien-funcall (extern-alien "sb_nanosleep_float" (function (values) float))
+                 seconds)
+  nil)
 
 ;;;; sys/time.h
 
@@ -1011,7 +977,6 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
 
 #!-win32
 (defun unix-getitimer (which)
-  #!+sb-doc
   "UNIX-GETITIMER returns the INTERVAL and VALUE slots of one of
    three system timers (:real :virtual or :profile). On success,
    unix-getitimer returns 5 values,
@@ -1035,7 +1000,6 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
 
 #!-win32
 (defun unix-setitimer (which int-secs int-usec val-secs val-usec)
-  #!+sb-doc
   "UNIX-SETITIMER sets the INTERVAL and VALUE slots of one of
    three system timers (:real :virtual or :profile). A SIGALRM signal
    will be delivered VALUE <seconds+microseconds> from now. INTERVAL,
@@ -1086,7 +1050,6 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
 
   #!-sb-fluid (declaim (inline get-time-of-day))
   (defun get-time-of-day ()
-    #!+sb-doc
     "Return the number of seconds and microseconds since the beginning of
 the UNIX epoch (January 1st 1970.)"
     #!+(or darwin netbsd)
@@ -1116,6 +1079,7 @@ the UNIX epoch (January 1st 1970.)"
                    system-real-time-values))
 
   (defun system-real-time-values ()
+    #!+win32 (declare (notinline get-time-of-day)) ; forward ref
     (multiple-value-bind (sec usec) (get-time-of-day)
       (declare (type unsigned-byte sec) (type (unsigned-byte 31) usec))
       (values sec (truncate usec micro-seconds-per-internal-time-unit))))
@@ -1200,6 +1164,7 @@ the UNIX epoch (January 1st 1970.)"
 ;;; So we're stuck with it for a while -- maybe delete it towards the end
 ;;; of 2009.
 (defun unix-gettimeofday ()
+  #!+win32 (declare (notinline get-time-of-day)) ; forward ref
   (multiple-value-bind (sec usec) (get-time-of-day)
     (values t sec usec nil nil)))
 
