@@ -14,15 +14,11 @@
 (in-package "SB!VM")
 
 ;;; number of bits per word where a word holds one lisp descriptor
-(def!constant n-word-bits 64)
+(defconstant n-word-bits 64)
 
 ;;; the natural width of a machine word (as seen in e.g. register width,
 ;;; address space)
-(def!constant n-machine-word-bits 64)
-
-;;; number of bits per byte where a byte is the smallest addressable
-;;; object
-(def!constant n-byte-bits 8)
+(defconstant n-machine-word-bits 64)
 
 ;;; Floating-point related constants, both format descriptions and FPU
 ;;; control register descriptions.  These don't exactly match up with
@@ -30,43 +26,43 @@
 ;;; defines floating-point values somewhat differently than the IEEE
 ;;; standard does.
 
-(def!constant float-sign-shift 31)
+(defconstant float-sign-shift 31)
 
-(def!constant single-float-bias 126)
+(defconstant single-float-bias 126)
 (defconstant-eqx single-float-exponent-byte (byte 8 23) #'equalp)
 (defconstant-eqx single-float-significand-byte (byte 23 0) #'equalp)
-(def!constant single-float-normal-exponent-min 1)
-(def!constant single-float-normal-exponent-max 254)
-(def!constant single-float-hidden-bit (ash 1 23))
-(def!constant single-float-trapping-nan-bit (ash 1 22))
+(defconstant single-float-normal-exponent-min 1)
+(defconstant single-float-normal-exponent-max 254)
+(defconstant single-float-hidden-bit (ash 1 23))
+(defconstant single-float-trapping-nan-bit (ash 1 22))
 
-(def!constant double-float-bias 1022)
+(defconstant double-float-bias 1022)
 (defconstant-eqx double-float-exponent-byte (byte 11 20) #'equalp)
 (defconstant-eqx double-float-significand-byte (byte 20 0) #'equalp)
-(def!constant double-float-normal-exponent-min 1)
-(def!constant double-float-normal-exponent-max #x7FE)
-(def!constant double-float-hidden-bit (ash 1 20))
-(def!constant double-float-trapping-nan-bit (ash 1 19))
+(defconstant double-float-normal-exponent-min 1)
+(defconstant double-float-normal-exponent-max #x7FE)
+(defconstant double-float-hidden-bit (ash 1 20))
+(defconstant double-float-trapping-nan-bit (ash 1 19))
 
-(def!constant single-float-digits
+(defconstant single-float-digits
   (+ (byte-size single-float-significand-byte) 1))
 
-(def!constant double-float-digits
+(defconstant double-float-digits
   (+ (byte-size double-float-significand-byte) 32 1))
 
 
 (progn
-  (def!constant float-invalid-trap-bit (ash 1 0))
-  (def!constant float-divide-by-zero-trap-bit (ash 1 1))
-  (def!constant float-overflow-trap-bit (ash 1 2))
-  (def!constant float-underflow-trap-bit (ash 1 3))
-  (def!constant float-inexact-trap-bit (ash 1 4))
-  (def!constant float-input-denormal-trap-bit (ash 1 7))
+  (defconstant float-invalid-trap-bit (ash 1 0))
+  (defconstant float-divide-by-zero-trap-bit (ash 1 1))
+  (defconstant float-overflow-trap-bit (ash 1 2))
+  (defconstant float-underflow-trap-bit (ash 1 3))
+  (defconstant float-inexact-trap-bit (ash 1 4))
+  (defconstant float-input-denormal-trap-bit (ash 1 7))
 
-  (def!constant float-round-to-nearest 0)
-  (def!constant float-round-to-positive 1)
-  (def!constant float-round-to-negative 2)
-  (def!constant float-round-to-zero 3)
+  (defconstant float-round-to-nearest 0)
+  (defconstant float-round-to-positive 1)
+  (defconstant float-round-to-negative 2)
+  (defconstant float-round-to-zero 3)
 
   (defconstant-eqx float-rounding-mode (byte 2 22) #'equalp)
 
@@ -74,34 +70,31 @@
   (defconstant-eqx float-traps-byte (byte 8 8) #'equalp)
   (defconstant-eqx float-exceptions-byte (byte 8 0) #'equalp)
 
-  (def!constant float-fast-bit (ash 1 24))) ;; Flush-to-zero mode
+  (defconstant float-fast-bit (ash 1 24))) ;; Flush-to-zero mode
 
 ;;;; Where to put the different spaces.
 
 ;;; On non-gencgc we need large dynamic and static spaces for PURIFY
 #!-gencgc
 (progn
-  (def!constant read-only-space-start #x04000000)
-  (def!constant read-only-space-end   #x07ff8000)
-  (def!constant static-space-start    #x08000000)
-  (def!constant static-space-end      #x097fff00)
+  (defconstant read-only-space-start #x04000000)
+  (defconstant read-only-space-end   #x07ff8000)
+  (defconstant static-space-start    #x08000000)
+  (defconstant static-space-end      #x097fff00)
 
-  (def!constant linkage-table-space-start #x0a000000)
-  (def!constant linkage-table-space-end   #x0b000000))
+  (defconstant linkage-table-space-start #x0a000000)
+  (defconstant linkage-table-space-end   #x0b000000)
+  #!+linux
+  (progn
+    (defconstant dynamic-0-space-start #x4f000000)
+    (defconstant dynamic-0-space-end   #x66fff000)
+    (defconstant dynamic-1-space-start #x67000000)
+    (defconstant dynamic-1-space-end   #x7efff000)))
 
 #!+gencgc
-(!gencgc-space-setup #x04000000 #x4f000000)
+(!gencgc-space-setup #xF0000000 :dynamic-space-start #x1000000000)
 
-(def!constant linkage-table-entry-size 16)
-
-#!+linux
-(progn
-  #!-gencgc
-  (progn
-    (def!constant dynamic-0-space-start #x4f000000)
-    (def!constant dynamic-0-space-end   #x66fff000)
-    (def!constant dynamic-1-space-start #x67000000)
-    (def!constant dynamic-1-space-end   #x7efff000)))
+(defconstant linkage-table-entry-size 16)
 
 ;;;; other miscellaneous constants
 
