@@ -95,13 +95,13 @@
 ;;;; classes
 
 (sb!xc:deftype name-for-class () t) ; FIXME: disagrees w/ LEGAL-CLASS-NAME-P
-(defknown classoid-name (classoid) symbol (flushable))
 (defknown find-classoid (name-for-class &optional t)
   (or classoid null) ())
 (defknown classoid-of (t) classoid (flushable))
 (defknown layout-of (t) layout (flushable))
 (defknown copy-structure (structure-object) structure-object
-  (flushable)) ;; FIXME: can derive the type based on the structure
+  (flushable)
+  :derive-type #'result-type-first-arg)
 
 ;;;; from the "Control Structure" chapter:
 
@@ -179,7 +179,11 @@
 (defknown get-properties (list list) (values t t list) (foldable flushable))
 (defknown symbol-name (symbol) simple-string (movable foldable flushable))
 (defknown make-symbol (string) symbol (flushable))
-(defknown %make-symbol (simple-string) symbol (flushable))
+;; %make-symbol is the internal API, but the primitive object allocator
+;; is %%make-symbol, because when immobile space feature is present,
+;; we dispatch to either the C allocator or the Lisp allocator.
+(defknown %make-symbol (fixnum simple-string) symbol (flushable))
+(defknown sb!vm::%%make-symbol (simple-string) symbol (flushable))
 (defknown copy-symbol (symbol &optional t) symbol (flushable))
 (defknown gensym (&optional (or string unsigned-byte)) symbol ())
 (defknown symbol-package (symbol) (or package null) (flushable))
@@ -255,6 +259,18 @@
 (defknown (1+ 1-) (number) number
   (movable foldable flushable))
 
+(defknown (two-arg-* two-arg-+ two-arg-- two-arg-/)
+  (number number) number
+  ())
+
+(defknown (two-arg-< two-arg-= two-arg->)
+  (number number) boolean
+  ())
+
+(defknown (two-arg-gcd two-arg-lcm two-arg-and two-arg-ior two-arg-xor two-arg-eqv)
+  (integer integer) integer
+  ())
+
 (defknown conjugate (number) number
   (movable foldable flushable))
 
@@ -487,19 +503,19 @@
 (defknown elt (sequence index) t (foldable unsafely-flushable))
 
 (defknown subseq (sequence index &optional sequence-end) consed-sequence
-  (flushable)
-  :derive-type (sequence-result-nth-arg 1))
+  (flushable))
 
 (defknown copy-seq (sequence) consed-sequence (flushable)
-  :derive-type (sequence-result-nth-arg 1))
+  :derive-type (sequence-result-nth-arg 1 :preserve-dimensions t))
 
 (defknown length (sequence) index (foldable flushable dx-safe))
 
 (defknown reverse (sequence) consed-sequence (flushable)
-  :derive-type (sequence-result-nth-arg 1))
+  :derive-type (sequence-result-nth-arg 1 :preserve-dimensions t))
 
 (defknown nreverse (sequence) sequence (important-result)
-  :derive-type #'result-type-first-arg
+  :derive-type (sequence-result-nth-arg 1 :preserve-dimensions t
+                                          :preserve-vector-type t)
   :destroyed-constant-args (nth-constant-nonempty-sequence-args 1))
 
 (defknown make-sequence (type-specifier index
@@ -516,8 +532,15 @@
   (flushable))
 (defknown %concatenate-to-base-string (&rest sequence) simple-base-string
   (flushable))
+(defknown %concatenate-to-list (&rest sequence) list
+    (flushable))
+(defknown %concatenate-to-simple-vector (&rest sequence) simple-vector
+  (flushable))
+(defknown %concatenate-to-vector ((unsigned-byte #.sb!vm:n-widetag-bits) &rest sequence)
+    vector
+  (flushable))
 
-(defknown map (type-specifier callable sequence &rest sequence)
+(defknown map (type-specifier (callable &rest) sequence &rest sequence)
   consed-sequence (call)
 ; :DERIVE-TYPE 'TYPE-SPEC-ARG1 ? Nope... (MAP NIL ...) returns NULL, not NIL.
   )
@@ -527,7 +550,7 @@
 (defknown %map-to-simple-vector-arity-1 (callable sequence) simple-vector
   (flushable call))
 
-(defknown map-into (sequence callable &rest sequence)
+(defknown map-into (sequence (callable &rest) &rest sequence)
   sequence
   (call)
   :derive-type #'result-type-first-arg
@@ -549,8 +572,8 @@
 (defknown (every notany notevery) (callable sequence &rest sequence) boolean
   (foldable unsafely-flushable call))
 
-(defknown reduce (callable sequence &rest t &key (:from-end t) (:start index)
-                  (:end sequence-end) (:initial-value t) (:key callable))
+(defknown reduce ((callable 2) sequence &rest t &key (:from-end t) (:start index)
+                  (:end sequence-end) (:initial-value t) (:key (callable 1)))
   t
   (foldable flushable call))
 
@@ -569,136 +592,132 @@
   :result-arg 0)
 
 (defknown remove
-  (t sequence &rest t &key (:from-end t) (:test callable)
-     (:test-not callable) (:start index) (:end sequence-end)
-     (:count sequence-count) (:key callable))
+  (t sequence &rest t &key (:from-end t) (:test (callable 2))
+     (:test-not (callable 2)) (:start index) (:end sequence-end)
+     (:count sequence-count) (:key (callable 1)))
   consed-sequence
   (flushable call)
   :derive-type (sequence-result-nth-arg 2))
 
 (defknown substitute
-  (t t sequence &rest t &key (:from-end t) (:test callable)
-     (:test-not callable) (:start index) (:end sequence-end)
-     (:count sequence-count) (:key callable))
+  (t t sequence &rest t &key (:from-end t) (:test (callable 2))
+     (:test-not (callable 2)) (:start index) (:end sequence-end)
+     (:count sequence-count) (:key (callable 1)))
   consed-sequence
   (flushable call)
   :derive-type (sequence-result-nth-arg 3))
 
 (defknown (remove-if remove-if-not)
-  (callable sequence &rest t &key (:from-end t) (:start index)
-            (:end sequence-end) (:count sequence-count) (:key callable))
+  ((callable 1) sequence &rest t &key (:from-end t) (:start index)
+            (:end sequence-end) (:count sequence-count) (:key (callable 1)))
   consed-sequence
   (flushable call)
   :derive-type (sequence-result-nth-arg 2))
 
 (defknown (substitute-if substitute-if-not)
-  (t callable sequence &rest t &key (:from-end t) (:start index)
-     (:end sequence-end) (:count sequence-count) (:key callable))
+  (t (callable 1) sequence &rest t &key (:from-end t) (:start index)
+     (:end sequence-end) (:count sequence-count) (:key (callable 1)))
   consed-sequence
   (flushable call)
   :derive-type (sequence-result-nth-arg 3))
 
 (defknown delete
-  (t sequence &rest t &key (:from-end t) (:test callable)
-     (:test-not callable) (:start index) (:end sequence-end)
-     (:count sequence-count) (:key callable))
+  (t sequence &rest t &key (:from-end t) (:test (callable 2))
+     (:test-not (callable 2)) (:start index) (:end sequence-end)
+     (:count sequence-count) (:key (callable 1)))
   sequence
   (flushable call important-result)
   :derive-type (sequence-result-nth-arg 2)
   :destroyed-constant-args (nth-constant-nonempty-sequence-args 2))
 
 (defknown nsubstitute
-  (t t sequence &rest t &key (:from-end t) (:test callable)
-     (:test-not callable) (:start index) (:end sequence-end)
-     (:count sequence-count) (:key callable))
+  (t t sequence &rest t &key (:from-end t) (:test (callable 2))
+     (:test-not (callable 2)) (:start index) (:end sequence-end)
+     (:count sequence-count) (:key (callable 1)))
   sequence
   (flushable call)
   :derive-type (sequence-result-nth-arg 3)
   :destroyed-constant-args (nth-constant-nonempty-sequence-args 3))
 
 (defknown (delete-if delete-if-not)
-  (callable sequence &rest t &key (:from-end t) (:start index)
-            (:end sequence-end) (:count sequence-count) (:key callable))
+  ((callable 1) sequence &rest t &key (:from-end t) (:start index)
+   (:end sequence-end) (:count sequence-count) (:key (callable 1)))
   sequence
   (flushable call important-result)
   :derive-type (sequence-result-nth-arg 2)
   :destroyed-constant-args (nth-constant-nonempty-sequence-args 2))
 
 (defknown (nsubstitute-if nsubstitute-if-not)
-  (t callable sequence &rest t &key (:from-end t) (:start index)
-     (:end sequence-end) (:count sequence-count) (:key callable))
+  (t (callable 1) sequence &rest t &key (:from-end t) (:start index)
+     (:end sequence-end) (:count sequence-count) (:key (callable 1)))
   sequence
   (flushable call)
   :derive-type (sequence-result-nth-arg 3)
   :destroyed-constant-args (nth-constant-nonempty-sequence-args 3))
 
 (defknown remove-duplicates
-  (sequence &rest t &key (:test callable) (:test-not callable) (:start index)
-            (:from-end t) (:end sequence-end) (:key callable))
+  (sequence &rest t &key (:test (callable 2)) (:test-not (callable 2)) (:start index)
+            (:from-end t) (:end sequence-end) (:key (callable 1)))
   consed-sequence
   (unsafely-flushable call)
   :derive-type (sequence-result-nth-arg 1))
 
 (defknown delete-duplicates
-  (sequence &rest t &key (:test callable) (:test-not callable) (:start index)
-            (:from-end t) (:end sequence-end) (:key callable))
+  (sequence &rest t &key (:test (callable 2)) (:test-not (callable 2)) (:start index)
+            (:from-end t) (:end sequence-end) (:key (callable 1)))
   sequence
   (unsafely-flushable call important-result)
   :derive-type (sequence-result-nth-arg 1)
   :destroyed-constant-args (nth-constant-nonempty-sequence-args 1))
 
-(defknown find (t sequence &rest t &key (:test callable)
-                (:test-not callable) (:start index) (:from-end t)
-                (:end sequence-end) (:key callable))
+(defknown find (t sequence &rest t &key (:test (callable 2))
+                (:test-not (callable 2)) (:start index) (:from-end t)
+                (:end sequence-end) (:key (callable 1)))
   t
   (foldable flushable call))
 
 (defknown (find-if find-if-not)
-  (callable sequence &rest t &key (:from-end t) (:start index)
-   (:end sequence-end) (:key callable))
+  ((callable 1) sequence &rest t &key (:from-end t) (:start index)
+   (:end sequence-end) (:key (callable 1)))
   t
   (foldable flushable call))
 
-(defknown position (t sequence &rest t &key (:test callable)
-                    (:test-not callable) (:start index) (:from-end t)
-                    (:end sequence-end) (:key callable))
-  (or index null)
-  (foldable flushable call)
-  :derive-type #'position-derive-type)
+(defknown position (t sequence &rest t &key (:test (callable 2))
+                    (:test-not (callable 2)) (:start index) (:from-end t)
+                    (:end sequence-end) (:key (callable 1)))
+  (or (mod #.(1- sb!xc:array-dimension-limit)) null)
+  (foldable flushable call))
 
 (defknown (position-if position-if-not)
-  (callable sequence &rest t &key (:from-end t) (:start index)
-   (:end sequence-end) (:key callable))
-  (or index null)
-  (foldable flushable call)
-  :derive-type #'position-derive-type)
+  ((callable 1) sequence &rest t &key (:from-end t) (:start index)
+   (:end sequence-end) (:key (callable 1)))
+  (or (mod #.(1- sb!xc:array-dimension-limit)) null)
+  (foldable flushable call))
 
 (defknown count (t sequence &rest t &key
-                   (:test callable) (:test-not callable) (:start index)
-                   (:from-end t) (:end sequence-end) (:key callable))
+                   (:test (callable 2)) (:test-not (callable 2)) (:start index)
+                   (:from-end t) (:end sequence-end) (:key (callable 1)))
   index
-  (foldable flushable call)
-  :derive-type #'count-derive-type)
+  (foldable flushable call))
 
 (defknown (count-if count-if-not)
-  (callable sequence &rest t &key
-            (:from-end t) (:start index) (:end sequence-end) (:key callable))
+  ((callable 1) sequence &rest t &key
+   (:from-end t) (:start index) (:end sequence-end) (:key (callable 1)))
   index
-  (foldable flushable call)
-  :derive-type #'count-derive-type)
+  (foldable flushable call))
 
 (defknown (mismatch search)
-  (sequence sequence &rest t &key (:from-end t) (:test callable)
-   (:test-not callable) (:start1 index) (:end1 sequence-end)
-   (:start2 index) (:end2 sequence-end) (:key callable))
+  (sequence sequence &rest t &key (:from-end t) (:test (callable 2))
+   (:test-not (callable 2)) (:start1 index) (:end1 sequence-end)
+   (:start2 index) (:end2 sequence-end) (:key (callable 1)))
   (or index null)
   (foldable flushable call))
 
 ;;; not FLUSHABLE, since vector sort guaranteed in-place...
-(defknown (stable-sort sort) (sequence callable &rest t &key (:key callable))
+(defknown (stable-sort sort) (sequence (callable 2) &rest t &key (:key (callable 1)))
   sequence
   (call)
-  :derive-type (sequence-result-nth-arg 1)
+  :derive-type #'result-type-first-arg
   :destroyed-constant-args (nth-constant-nonempty-sequence-args 1))
 (defknown sb!impl::stable-sort-list (list function function) list
   (call important-result)
@@ -720,8 +739,8 @@
   (call)
   :destroyed-constant-args (nth-constant-nonempty-sequence-args 1))
 
-(defknown merge (type-specifier sequence sequence callable
-                                &key (:key callable))
+(defknown merge (type-specifier sequence sequence (callable 2)
+                                &key (:key (callable 1)))
   sequence
   (call important-result)
   :derive-type (creation-result-type-specifier-nth-arg 1)
@@ -762,7 +781,7 @@
 
 (defknown cons (t t) cons (movable flushable))
 
-(defknown tree-equal (t t &key (:test callable) (:test-not callable)) boolean
+(defknown tree-equal (t t &key (:test (callable 2)) (:test-not (callable 2))) boolean
   (foldable flushable call))
 (defknown endp (list) boolean (foldable flushable movable))
 (defknown list-length (list) (or index null) (foldable unsafely-flushable))
@@ -779,6 +798,7 @@
 (defknown list* (t &rest t) t (movable flushable))
 (defknown make-list (index &key (:initial-element t)) list
   (movable flushable))
+(defknown %make-list (index t) list (movable flushable))
 
 (defknown sb!impl::|List| (&rest t) list (movable flushable foldable))
 (defknown sb!impl::|List*| (t &rest t) t (movable flushable foldable))
@@ -811,55 +831,55 @@
 (defknown (rplaca rplacd) (cons t) cons ()
   :destroyed-constant-args (nth-constant-args 1))
 
-(defknown subst (t t t &key (:key callable) (:test callable)
-                   (:test-not callable))
+(defknown subst (t t t &key (:key (callable 1)) (:test (callable 2))
+                   (:test-not (callable 2)))
   t (flushable call))
-(defknown nsubst (t t t &key (:key callable) (:test callable)
-                    (:test-not callable))
+(defknown nsubst (t t t &key (:key (callable 1)) (:test (callable 2))
+                    (:test-not (callable 2)))
   t (call)
   :destroyed-constant-args (nth-constant-nonempty-sequence-args 3))
 
 (defknown (subst-if subst-if-not)
-          (t callable t &key (:key callable))
+          (t (callable 1) t &key (:key (callable 1)))
   t (flushable call))
 (defknown (nsubst-if nsubst-if-not)
-          (t callable t &key (:key callable))
+          (t (callable 1) t &key (:key (callable 1)))
   t (call)
   :destroyed-constant-args (nth-constant-nonempty-sequence-args 3))
 
-(defknown sublis (list t &key (:key callable) (:test callable)
-                       (:test-not callable))
+(defknown sublis (list t &key (:key (callable 1)) (:test (callable 2))
+                       (:test-not (callable 2)))
   t (flushable call))
-(defknown nsublis (list t &key (:key callable) (:test callable)
-                        (:test-not callable))
+(defknown nsublis (list t &key (:key (callable 1)) (:test (callable 2))
+                        (:test-not (callable 2)))
   t (flushable call)
   :destroyed-constant-args (nth-constant-nonempty-sequence-args 2))
 
-(defknown member (t list &key (:key callable) (:test callable)
-                    (:test-not callable))
+(defknown member (t list &key (:key (callable 1)) (:test (callable 2))
+                    (:test-not (callable 2)))
   list (foldable flushable call))
-(defknown (member-if member-if-not) (callable list &key (:key callable))
+(defknown (member-if member-if-not) ((callable 1) list &key (:key (callable 1)))
   list (foldable flushable call))
 
 (defknown tailp (t list) boolean (foldable flushable))
 
-(defknown adjoin (t list &key (:key callable) (:test callable)
-                    (:test-not callable))
+(defknown adjoin (t list &key (:key (callable 1)) (:test (callable 2))
+                    (:test-not (callable 2)))
   cons (flushable call))
 
 (defknown (union intersection set-difference set-exclusive-or)
-  (list list &key (:key callable) (:test callable) (:test-not callable))
+  (list list &key (:key (callable 1)) (:test (callable 2)) (:test-not (callable 2)))
   list
   (foldable flushable call))
 
 (defknown (nunion nintersection nset-difference nset-exclusive-or)
-  (list list &key (:key callable) (:test callable) (:test-not callable))
+  (list list &key (:key (callable 1)) (:test (callable 2)) (:test-not (callable 2)))
   list
   (foldable flushable call important-result)
   :destroyed-constant-args (nth-constant-nonempty-sequence-args 1 2))
 
 (defknown subsetp
-  (list list &key (:key callable) (:test callable) (:test-not callable))
+  (list list &key (:key (callable 1)) (:test (callable 2)) (:test-not (callable 2)))
   boolean
   (foldable flushable call))
 
@@ -867,10 +887,10 @@
 (defknown pairlis (t t &optional t) list (flushable))
 
 (defknown (rassoc assoc)
-          (t list &key (:key callable) (:test callable) (:test-not callable))
+          (t list &key (:key (callable 1)) (:test (callable 2)) (:test-not (callable 2)))
   list (foldable flushable call))
 (defknown (assoc-if-not assoc-if rassoc-if rassoc-if-not)
-          (callable list &key (:key callable)) list (foldable flushable call))
+          ((callable 1) list &key (:key (callable 1))) list (foldable flushable call))
 
 (defknown (memq assq) (t list) list (foldable flushable))
 (defknown delq (t list) list (flushable)
@@ -897,7 +917,7 @@
   :derive-type #'result-type-last-arg)
 (defknown remhash (t hash-table) boolean ()
   :destroyed-constant-args (nth-constant-args 2))
-(defknown maphash (callable hash-table) null (flushable call))
+(defknown maphash ((callable 2) hash-table) null (flushable call))
 (defknown clrhash (hash-table) hash-table ()
   :destroyed-constant-args (nth-constant-args 2))
 (defknown hash-table-count (hash-table) index (flushable))
@@ -926,7 +946,7 @@
 
 (defknown %make-array ((or index list)
                        (unsigned-byte #.sb!vm:n-widetag-bits)
-                       (unsigned-byte 16)
+                       (mod #.sb!vm:n-word-bits)
                        &key
                        (:element-type type-specifier)
                        (:initial-element t)
@@ -937,15 +957,35 @@
                        (:displaced-index-offset index))
     array (flushable))
 
-(defknown sb!impl::fill-data-vector (vector list sequence) t ())
+(defknown fill-data-vector (vector list sequence) t ())
+
+;; INITIAL-CONTENTS is the first argument to FILL-ARRAY because
+;; of the possibility of source-transforming it for various recognized
+;; contents after the source-transform for MAKE-ARRAY has run.
+;; The original MAKE-ARRAY call might resemble either of:
+;;   (MAKE-ARRAY DIMS :ELEMENT-TYPE (F) :INITIAL-CONTENTS (G))
+;;   (MAKE-ARRAY DIMS :INITIAL-CONTENTS (F) :ELEMENT-TYPE (G))
+;; and in general we must bind temporaries to preserve left-to-right
+;; evaluation of F and G if they have side effects.
+;; Now ideally :ELEMENT-TYPE will be a constant, so it doesn't matter
+;; if it moves. But if INITIAL-CONTENTS were the second argument to FILL-ARRAY,
+;; then the multi-dimensional array creation form would look like
+;;  (FILL-ARRAY (allocate-array) initial-contents)
+;; which would mean that if the user expressed the MAKE- call the
+;; second way shown above, we would have to bind INITIAL-CONTENTS,
+;; to ensure its evaluation before the allocation,
+;; causing difficulty if doing any futher macro-like processing.
+(defknown fill-array (sequence simple-array) (simple-array) (flushable)
+  :result-arg 1)
 
 (defknown vector (&rest t) simple-vector (flushable))
 
-(defknown aref (array &rest index) t (foldable))
+(defknown aref (array &rest index) t (foldable)
+  :call-type-deriver #'array-call-type-deriver)
 (defknown row-major-aref (array index) t (foldable))
 
 (defknown array-element-type (array) (or list symbol)
-  (foldable flushable recursive))
+  (foldable flushable))
 (defknown array-rank (array) array-rank (foldable flushable))
 ;; FIXME: there's a fencepost bug, but for all practical purposes our
 ;; ARRAY-RANK-LIMIT is infinite, thus masking the bug. e.g. if the
@@ -953,9 +993,11 @@
 ;; be in the range 0 through 6, not 0 through 7.
 (defknown array-dimension (array array-rank) index (foldable flushable))
 (defknown array-dimensions (array) list (foldable flushable))
-(defknown array-in-bounds-p (array &rest integer) boolean (foldable flushable))
+(defknown array-in-bounds-p (array &rest integer) boolean (foldable flushable)
+  :call-type-deriver #'array-call-type-deriver)
 (defknown array-row-major-index (array &rest index) array-total-size
-  (foldable flushable))
+  (foldable flushable)
+  :call-type-deriver #'array-call-type-deriver)
 (defknown array-total-size (array) array-total-size (foldable flushable))
 (defknown adjustable-array-p (array) boolean (movable foldable flushable))
 
@@ -1024,6 +1066,18 @@
   (or index null)
   (foldable flushable))
 
+(defknown (two-arg-string= two-arg-string-equal)
+  (string-designator string-designator)
+  boolean
+  (foldable flushable))
+
+(defknown (two-arg-string< two-arg-string> two-arg-string<= two-arg-string>=
+           two-arg-string/= two-arg-string-lessp two-arg-string-greaterp
+           two-arg-string-not-lessp two-arg-string-not-greaterp two-arg-string-not-equal)
+  (string-designator string-designator)
+  (or index null)
+  (foldable flushable))
+
 (defknown make-string (index &key (:element-type type-specifier)
                        (:initial-element character))
   simple-string (flushable))
@@ -1038,7 +1092,8 @@
 (defknown (nstring-upcase nstring-downcase nstring-capitalize)
   (string &key (:start index) (:end sequence-end))
   string ()
-  :destroyed-constant-args (nth-constant-nonempty-sequence-args 1))
+  :destroyed-constant-args (nth-constant-nonempty-sequence-args 1)
+  :derive-type #'result-type-first-arg)
 
 (defknown string (string-designator) string (flushable))
 
@@ -1101,17 +1156,22 @@
   (character character &optional readtable (or readtable null)) (eql t)
   ())
 
-(defknown set-macro-character (character callable &optional t (or readtable null))
+(defknown set-macro-character (character (callable 2 no-function-conversion)
+                               &optional t (or readtable null))
   (eql t)
-  ())
+  (call))
 (defknown get-macro-character (character &optional (or readtable null))
   (values callable boolean) (flushable))
 
 (defknown make-dispatch-macro-character (character &optional t readtable)
   (eql t) ())
+;;; FIXME: (CALLABLE 3 ...) causes a style warning in the :UNICODE-DISPATCH-MACROS
+;;; test in reader.impure.lisp
+;;;   The function NIL is called by SET-DISPATCH-MACRO-CHARACTER with three arguments, but wants exactly two.
 (defknown set-dispatch-macro-character
-  (character character callable &optional (or readtable null)) (eql t)
-  ())
+  (character character (callable 3 no-function-conversion)
+   &optional (or readtable null)) (eql t)
+  (call))
 (defknown get-dispatch-macro-character
   (character character &optional (or readtable null)) (or callable null)
   ())
@@ -1146,10 +1206,10 @@
   null
   ())
 (defknown set-pprint-dispatch
-  (type-specifier (or null callable)
+  (type-specifier (callable 2 no-function-conversion)
    &optional real sb!pretty:pprint-dispatch-table)
   null
-  ())
+  (call))
 
 ;;; may return any type due to eof-value...
 ;;; and because READ generally returns anything.
@@ -1232,7 +1292,7 @@
   (deffrob write-to-string () simple-string
            (unsafely-flushable)))
 
-(defknown (prin1-to-string princ-to-string) (t) simple-string (flushable))
+(defknown (prin1-to-string princ-to-string) (t) simple-string (unsafely-flushable))
 (defknown sb!impl::stringify-object (t) simple-string)
 
 (defknown write-char (character &optional stream-designator) character ()
@@ -1257,10 +1317,19 @@
 (defknown format ((or (member nil t) stream string)
                   (or string function) &rest t)
   (or string null)
-  ())
+    ())
+(defknown sb!format::format-error* (string list &rest t &key &allow-other-keys)
+    nil)
+(defknown sb!format::format-error (string &rest t) nil)
+(defknown sb!format::format-error-at* ((or null string) (or null index) string list
+                                       &rest t &key &allow-other-keys)
+    nil)
+(defknown sb!format::format-error-at ((or null string) (or null index) string
+                                      &rest t)
+    nil)
 (defknown sb!format::args-exhausted (string integer) nil)
 
-(defknown (y-or-n-p yes-or-no-p) (&optional string &rest t) boolean
+(defknown (y-or-n-p yes-or-no-p) (&optional (or string null function) &rest t) boolean
   ())
 
 ;;;; from the "File System Interface" chapter:
@@ -1395,12 +1464,13 @@
 
 ;;;; from the "Conditions" chapter:
 
-(defknown error (t &rest t) nil)
-(defknown cerror (format-control t &rest t) null)
+(defknown (signal warn) (condition-designator-head &rest t) null)
+(defknown error (condition-designator-head &rest t) nil)
+(defknown cerror (format-control condition-designator-head &rest t) null)
 (defknown invalid-method-error (t format-control &rest t) *) ; FIXME: first arg is METHOD
 (defknown method-combination-error (format-control &rest t) *)
-(defknown signal (t &rest t) null)
-(defknown warn (t &rest t) null)
+(defknown assert-error (t &rest t) null)
+(defknown check-type-error (t t type-specifier &optional (or null string)) t)
 (defknown invoke-debugger (condition) nil)
 (defknown break (&optional format-control &rest t) null)
 (defknown make-condition (type-specifier &rest t) condition ())
@@ -1552,9 +1622,14 @@
 (defknown %typep-wrapper (t t (or type-specifier ctype)) t
   (movable flushable always-translatable))
 
+;;; An identity wrapper to avoid complaints about constant modification
+(defknown ltv-wrapper (t) t
+  (movable flushable always-translatable)
+  :derive-type #'result-type-first-arg)
+
 (defknown %cleanup-point () t)
 (defknown %special-bind (t t) t)
-(defknown %special-unbind (t) t)
+(defknown %special-unbind (index) t)
 (defknown %listify-rest-args (t index) list (flushable))
 (defknown %more-arg-context (t t) (values t index) (flushable))
 (defknown %more-arg (t index) t)
@@ -1567,7 +1642,8 @@
 #!-precise-arg-count-error
 (defknown %verify-arg-count (index index) (values))
 
-(defknown %arg-count-error (t t) nil)
+(defknown %arg-count-error (t) nil)
+(defknown %local-arg-count-error (t t) nil)
 (defknown %unknown-values () *)
 (defknown %catch (t t) t)
 (defknown %unwind-protect (t t) t)
@@ -1581,23 +1657,23 @@
 (defknown %nip-values (t t &rest t) (values))
 (defknown %dummy-dx-alloc (t t) t)
 (defknown %allocate-closures (t) *)
-(defknown %type-check-error (t t) nil)
-(defknown %type-check-error/c (t t) nil)
+(defknown %type-check-error (t t t) nil)
+(defknown %type-check-error/c (t t t) nil)
 
 ;; FIXME: This function does not return, but due to the implementation
 ;; of FILTER-LVAR we cannot write it here.
-(defknown %compile-time-type-error (t t t t t) *)
-(defknown case-failure (t t t) nil)
+(defknown %compile-time-type-error (t t t t t t) *)
+(defknown (etypecase-failure ecase-failure) (t t) nil)
 
 (defknown %odd-key-args-error () nil)
-(defknown %unknown-key-arg-error (t) nil)
+(defknown %unknown-key-arg-error (t t) nil)
 (defknown (%ldb %mask-field) (bit-index bit-index integer) unsigned-byte
   (movable foldable flushable))
 (defknown (%dpb %deposit-field) (integer bit-index bit-index integer) integer
   (movable foldable flushable))
 (defknown %negate (number) number (movable foldable flushable))
-(defknown %check-bound (array index fixnum) index
-  (movable foldable flushable dx-safe))
+(defknown (%check-bound check-bound) (array index t) index
+  (dx-safe))
 (defknown data-vector-ref (simple-array index) t
   (foldable unsafely-flushable always-translatable))
 (defknown data-vector-ref-with-offset (simple-array fixnum fixnum) t
@@ -1625,11 +1701,11 @@
 (defknown array-bounding-indices-bad-error (t t t) nil)
 (defknown sequence-bounding-indices-bad-error (t t t) nil)
 (defknown %find-position
-  (t sequence t index sequence-end function function)
+  (t sequence t index sequence-end (function 1) (function 2))
   (values t (or index null))
   (flushable call))
 (defknown (%find-position-if %find-position-if-not)
-  (function sequence t index sequence-end function)
+  ((function 1) sequence t index sequence-end (function 1))
   (values t (or index null))
   (call))
 (defknown effective-find-position-test (callable callable)
@@ -1705,7 +1781,7 @@
 
 (defknown (setf aref) (t array &rest index) t ()
   :destroyed-constant-args (nth-constant-args 2)
-  :derive-type #'result-type-first-arg)
+  :call-type-deriver #'array-call-type-deriver)
 (defknown %set-row-major-aref (array index t) t ()
   :destroyed-constant-args (nth-constant-args 1))
 (defknown (%rplaca %rplacd) (cons t) t ()
@@ -1772,7 +1848,7 @@
 (defknown style-warn (t &rest t) null ())
 
 (defknown coerce-to-condition ((or condition symbol string function)
-                               list type-specifier symbol)
+                               type-specifier symbol &rest t)
     condition
     ())
 
@@ -1813,6 +1889,25 @@
 (defknown spin-loop-hint () (values)
     (always-translatable))
 
+;;;; Stream methods
+
+;;; Avoid a ton of FBOUNDP checks in the string stream constructors etc,
+;;; by wiring in the needed functions instead of dereferencing their fdefns.
+(defknown (ill-in ill-bin ill-out ill-bout
+           sb!impl::string-inch sb!impl::string-in-misc
+           sb!impl::string-ouch sb!impl::string-sout sb!impl::string-out-misc
+           sb!impl::finite-base-string-ouch sb!impl::finite-base-string-out-misc
+           sb!impl::fill-pointer-ouch sb!impl::fill-pointer-sout
+           sb!impl::fill-pointer-misc
+           sb!impl::case-frob-upcase-out sb!impl::case-frob-upcase-sout
+           sb!impl::case-frob-downcase-out sb!impl::case-frob-downcase-sout
+           sb!impl::case-frob-capitalize-out sb!impl::case-frob-capitalize-sout
+           sb!impl::case-frob-capitalize-first-out sb!impl::case-frob-capitalize-first-sout
+           sb!impl::case-frob-capitalize-aux-out sb!impl::case-frob-capitalize-aux-sout
+           sb!impl::case-frob-misc
+           sb!pretty::pretty-out sb!pretty::pretty-misc) * *)
+(defknown sb!pretty::pretty-sout * * (recursive))
+
 ;;;; PCL
 
 (defknown sb!pcl::pcl-instance-p (t) boolean
