@@ -64,6 +64,7 @@
            #!+64-bit
            signed-byte-64-p
            weak-pointer-p code-component-p lra-p
+           sb!vm::unbound-marker-p
            simple-fun-p
            closurep
            funcallable-instance-p)
@@ -114,6 +115,8 @@
   ()
   :result-arg 0)
 
+;;; Return the length of VECTOR.
+;;; Ordinary code should prefer to use (LENGTH (THE VECTOR FOO)) instead.
 (defknown vector-length (vector) index (flushable dx-safe))
 
 (defknown vector-sap ((simple-unboxed-array (*))) system-area-pointer
@@ -121,80 +124,102 @@
 
 (defknown lowtag-of (t) (unsigned-byte #.sb!vm:n-lowtag-bits)
   (flushable movable))
-(defknown widetag-of (t) (unsigned-byte #.sb!vm:n-widetag-bits)
+;;; WIDETAG-OF needs extra code to handle LIST and FUNCTION lowtags.
+;;; When dealing with known other-pointers (dispatching on array
+;;; element type for example), %OTHER-POINTER-WIDETAG is faster.
+(defknown (widetag-of %other-pointer-widetag) (t)
+  (unsigned-byte #.sb!vm:n-widetag-bits)
   (flushable movable))
 
-;; FIXME: Sure looks like 24 should be (- n-word-bytes n-widetag-bits).
-;;        If not, add a comment why not.
-(defknown (get-header-data get-closure-length) (t) (unsigned-byte 24)
+;;; Return the data from the header of object, which for GET-HEADER-DATA
+;;; must be an other-pointer, and for GET-CLOSURE-LENGTH a fun-pointer.
+(defknown (get-header-data get-closure-length) (t)
+    (unsigned-byte #.(- sb!vm:n-word-bits sb!vm:n-widetag-bits))
   (flushable))
-(defknown set-header-data (t (unsigned-byte 24)) t
-  ())
+;;; This unconventional setter returns its first arg, not the newval.
+(defknown set-header-data
+    (t (unsigned-byte #.(- sb!vm:n-word-bits sb!vm:n-widetag-bits))) t)
+#!+64-bit
+(progn
+(defknown sb!vm::get-header-data-high (t) (unsigned-byte 32) (flushable))
+(defknown sb!vm::cas-header-data-high
+    (t (unsigned-byte 32) (unsigned-byte 32)) (unsigned-byte 32)))
 
 (defknown %array-dimension (t index) index
   (flushable))
 (defknown %set-array-dimension (t index index) index
   ())
-(defknown %array-rank (t) index
+(defknown %array-rank (t) array-rank
   (flushable))
 
+(defknown sb!kernel::check-array-shape (simple-array list)
+  (simple-array)
+  (flushable)
+  :result-arg 0)
+
 (defknown %make-instance (index) instance
   (flushable))
 (defknown %make-structure-instance (defstruct-description list &rest t) instance
   (flushable always-translatable))
 (defknown %instance-layout (instance) layout
   (foldable flushable))
+(defknown %funcallable-instance-layout (funcallable-instance) layout
+  (foldable flushable))
 (defknown %set-instance-layout (instance layout) layout
   ())
+(defknown %set-funcallable-instance-layout (funcallable-instance layout) layout
+  ())
 (defknown %instance-length (instance) index
   (foldable flushable))
+(defknown %instance-cas (instance index t t) t ())
 (defknown %instance-ref (instance index) t
   (flushable always-translatable))
 (defknown %instance-set (instance index t) t
-  (always-translatable))
+  (always-translatable)
+  :derive-type #'result-type-last-arg)
 (defknown %layout-invalid-error (t layout) nil)
 
-(defknown %raw-instance-ref/word (instance index) sb!vm:word
-  (flushable always-translatable))
-(defknown %raw-instance-set/word (instance index sb!vm:word) sb!vm:word
-  (always-translatable))
-(defknown %raw-instance-ref/single (instance index) single-float
-  (flushable always-translatable))
-(defknown %raw-instance-set/single (instance index single-float) single-float
-  (always-translatable))
-(defknown %raw-instance-ref/double (instance index) double-float
-  (flushable always-translatable))
-(defknown %raw-instance-set/double (instance index double-float) double-float
-  (always-translatable))
-(defknown %raw-instance-ref/complex-single (instance index)
-  (complex single-float)
-  (flushable always-translatable))
-(defknown %raw-instance-set/complex-single
-    (instance index (complex single-float))
-  (complex single-float)
-  (always-translatable))
-(defknown %raw-instance-ref/complex-double (instance index)
-  (complex double-float)
-  (flushable always-translatable))
-(defknown %raw-instance-set/complex-double
-    (instance index (complex double-float))
-  (complex double-float)
-  (always-translatable))
+#!+(or x86 x86-64)
+(defknown %raw-instance-cas/word (instance index sb!vm:word sb!vm:word)
+  sb!vm:word ())
+#.`(progn
+     ,@(map 'list
+            (lambda (rsd)
+              (let* ((reader (sb!kernel::raw-slot-data-accessor-name rsd))
+                     (name (copy-seq (string reader)))
+                     (writer (intern (replace name "-SET/"
+                                              :start1 (search "-REF/" name))))
+                     (type (sb!kernel::raw-slot-data-raw-type rsd)))
+                `(progn
+                   (defknown ,reader (instance index) ,type
+                     (flushable always-translatable))
+                   (defknown ,writer (instance index ,type) ,type
+                     (always-translatable) :derive-type #'result-type-last-arg)
+                   ;; Interpreter stubs, harmless but unnecessary on host
+                   #-sb-xc-host
+                   (progn (defun ,reader (instance index)
+                            (,reader instance index))
+                          (defun ,writer (instance index new-value)
+                            (,writer instance index new-value))))))
+            sb!kernel::*raw-slot-data*))
 
-#!+(or x86 x86-64 ppc)
+#!+compare-and-swap-vops
 (defknown %raw-instance-atomic-incf/word (instance index sb!vm:word) sb!vm:word
     (always-translatable))
-#!+(or x86 x86-64 ppc)
+#!+compare-and-swap-vops
 (defknown %array-atomic-incf/word (t index sb!vm:word) sb!vm:word
   (always-translatable))
 
 ;;; These two are mostly used for bit-bashing operations.
-(defknown %vector-raw-bits (t fixnum) sb!vm:word
+(defknown %vector-raw-bits (t index) sb!vm:word
   (flushable))
-(defknown (%set-vector-raw-bits) (t fixnum sb!vm:word) sb!vm:word
+(defknown (%set-vector-raw-bits) (t index sb!vm:word) sb!vm:word
   ())
 
 
+;;; Allocate an unboxed, non-fancy vector with type code TYPE, length LENGTH,
+;;; and WORDS words long. Note: it is your responsibility to ensure that the
+;;; relation between LENGTH and WORDS is correct.
 (defknown allocate-vector ((unsigned-byte 8) index
                            ;; The number of words is later converted
                            ;; to bytes, make sure it fits.
@@ -211,9 +236,11 @@
     (simple-array * (*))
     (flushable movable))
 
-(defknown make-array-header ((unsigned-byte 8) (unsigned-byte 24)) array
+;;; Allocate an array header with type code TYPE and rank RANK.
+(defknown make-array-header ((unsigned-byte 8) (mod #.sb!xc:array-rank-limit)) array
   (flushable movable))
 
+(defknown make-array-header* (&rest t) array (flushable movable))
 
 (defknown make-weak-pointer (t) weak-pointer
   (flushable))
@@ -304,22 +331,22 @@
 
 ;;;; bignum operations
 
-(defknown %allocate-bignum (bignum-length) bignum-type
+(defknown %allocate-bignum (bignum-length) bignum
   (flushable))
 
-(defknown %bignum-length (bignum-type) bignum-length
+(defknown %bignum-length (bignum) bignum-length
   (foldable flushable movable))
 
-(defknown %bignum-set-length (bignum-type bignum-length) bignum-type
+(defknown %bignum-set-length (bignum bignum-length) bignum
   ())
 
-(defknown %bignum-ref (bignum-type bignum-index) bignum-element-type
+(defknown %bignum-ref (bignum bignum-index) bignum-element-type
   (flushable))
 #!+(or x86 x86-64)
-(defknown %bignum-ref-with-offset (bignum-type fixnum (signed-byte 24))
+(defknown %bignum-ref-with-offset (bignum fixnum (signed-byte 24))
   bignum-element-type (flushable always-translatable))
 
-(defknown %bignum-set (bignum-type bignum-index bignum-element-type)
+(defknown %bignum-set (bignum bignum-index bignum-element-type)
   bignum-element-type
   ())
 
@@ -404,16 +431,15 @@
 
 ;;;; code/function/fdefn object manipulation routines
 
+;;; Return a SAP pointing to the instructions part of CODE-OBJ.
 (defknown code-instructions (t) system-area-pointer (flushable movable))
+;;; Extract the INDEXth element from the header of CODE-OBJ. Can be
+;;; set with SETF.
 (defknown code-header-ref (t index) t (flushable))
 (defknown code-header-set (t index t) t ())
 
-(defknown fun-subtype (function) (unsigned-byte #.sb!vm:n-widetag-bits)
+(defknown fun-subtype (function) (member . #.sb!vm::+fun-header-widetags+)
   (flushable))
-(defknown ((setf fun-subtype))
-          ((unsigned-byte #.sb!vm:n-widetag-bits) function)
-  (unsigned-byte #.sb!vm:n-widetag-bits)
-  ())
 
 (defknown make-fdefn (t) fdefn (flushable movable))
 (defknown fdefn-p (t) boolean (movable foldable flushable))
@@ -421,26 +447,35 @@
 (defknown fdefn-fun (fdefn) (or function null) (flushable))
 (defknown (setf fdefn-fun) (function fdefn) t ())
 (defknown fdefn-makunbound (fdefn) t ())
+;;; FDEFN -> FUNCTION, trapping if not FBOUNDP
 (defknown safe-fdefn-fun (fdefn) function ())
 
-(defknown %simple-fun-self (function) function
-  (flushable))
-(defknown (setf %simple-fun-self) (function function) function
-  ())
+(defknown %simple-fun-type (function) t (flushable))
 
-(defknown %closure-fun (function) function
-  (flushable))
+#!+(or x86 x86-64) (defknown sb!vm::%closure-callee (function) fixnum (flushable))
+(defknown %closure-fun (function) function (flushable))
 
 (defknown %closure-index-ref (function index) t
   (flushable))
 
+;; T argument is for the 'fun' slot.
+(defknown sb!vm::%copy-closure (index t) function (flushable))
+
+(defknown %fun-fun (function) function (flushable recursive))
+
 (defknown %make-funcallable-instance (index) function
   ())
 
 (defknown %funcallable-instance-info (function index) t (flushable))
 (defknown %set-funcallable-instance-info (function index t) t ())
+
+#!+sb-fasteval
+(defknown sb!interpreter:fun-proto-fn (sb!interpreter:interpreted-function)
+  sb!interpreter::interpreted-fun-prototype (flushable))
 
 
 (defknown %data-vector-and-index (array index)
                                  (values (simple-array * (*)) index)
                                  (foldable flushable))
+
+(defknown restart-point (t) t ())
