@@ -84,8 +84,15 @@
 (declaim (list *current-path*))
 (defvar *current-path*)
 
+(defun call-with-current-source-form (thunk &rest forms)
+  (let ((*current-path* (when (and (some #'identity forms)
+                                   (boundp '*source-paths*))
+                          (or (some #'get-source-path forms)
+                              (when (boundp '*current-path*)
+                                *current-path*)))))
+    (funcall thunk)))
+
 (defvar *derive-function-types* nil
-  #!+sb-doc
   "Should the compiler assume that function types will never change,
   so that it can use type information inferred from current definitions
   to optimize code which uses those definitions? Setting this true
@@ -105,7 +112,7 @@
   (let ((answer
          (typecase env
            (null nil)
-           #!+sb-fasteval
+           #!+(and sb-fasteval (host-feature sb-xc))
            (sb!interpreter:basic-env
             (sb!interpreter::fun-lexically-notinline-p name env))
            (t
@@ -118,12 +125,6 @@
     (eq (or answer (info :function :inlinep name))
         :notinline)))
 
-;; This will get redefined in PCL boot.
-(declaim (notinline maybe-update-info-for-gf))
-(defun maybe-update-info-for-gf (name)
-  (declare (ignore name))
-  nil)
-
 (defun maybe-defined-here (name where)
   (if (and (eq :defined where)
            (member name *fun-names-in-this-file* :test #'equal))
@@ -138,6 +139,8 @@
     (setf (info :function :where-from name) :assumed))
   (let ((where (info :function :where-from name)))
     (when (and (eq where :assumed)
+               ;; Slot accessors are defined just-in-time, if not already.
+               (not (typep name '(cons (eql sb!pcl::slot-accessor))))
                ;; In the ordinary target Lisp, it's silly to report
                ;; undefinedness when the function is defined in the
                ;; running Lisp. But at cross-compile time, the current
@@ -150,7 +153,7 @@
                ;; complain about undefined functions.
                (not latep))
       (note-undefined-reference name :function))
-    (let ((ftype (info :function :type name))
+    (let ((ftype (proclaimed-ftype name))
           (notinline (fun-lexically-notinline-p name)))
       (make-global-var
        :kind :global-function
@@ -162,7 +165,7 @@
                  ftype
                  (specifier-type 'function))
        :defined-type (if (and (not latep) (not notinline))
-                         (or (maybe-update-info-for-gf name) ftype)
+                         ftype
                          (specifier-type 'function))
        :where-from (if notinline
                        where
@@ -226,33 +229,30 @@
         (when old-free-fun
           (clear-invalid-functionals old-free-fun)
           old-free-fun))
-      (ecase (info :function :kind name)
-        ;; FIXME: The :MACRO and :SPECIAL-FORM cases could be merged.
-        (:macro
-         (compiler-error "The macro name ~S was found ~A." name context))
-        (:special-form
-         (compiler-error "The special form name ~S was found ~A."
-                         name
-                         context))
-        ((:function nil)
-         (check-fun-name name)
-         (let ((expansion (fun-name-inline-expansion name))
-               (inlinep (info :function :inlinep name)))
-           (setf (gethash name *free-funs*)
-                 (if (or expansion inlinep)
-                     (let ((where (info :function :where-from name)))
-                       (make-defined-fun
-                        :%source-name name
-                        :inline-expansion expansion
-                        :inlinep inlinep
-                        :where-from (if (eq inlinep :notinline)
-                                        where
-                                        (maybe-defined-here name where))
-                        :type (if (and (eq inlinep :notinline)
-                                       (neq where :declared))
-                                  (specifier-type 'function)
-                                  (info :function :type name))))
-                     (find-global-fun name nil))))))))
+      (let ((kind (info :function :kind name)))
+        (ecase kind
+          ((:macro :special-form)
+           (compiler-error "The ~(~S~) name ~S was found ~A."
+                           kind name context))
+          ((:function nil)
+           (check-fun-name name)
+           (let ((expansion (fun-name-inline-expansion name))
+                 (inlinep (info :function :inlinep name)))
+             (setf (gethash name *free-funs*)
+                   (if (or expansion inlinep)
+                       (let ((where (info :function :where-from name)))
+                         (make-defined-fun
+                          :%source-name name
+                          :inline-expansion expansion
+                          :inlinep inlinep
+                          :where-from (if (eq inlinep :notinline)
+                                          where
+                                          (maybe-defined-here name where))
+                          :type (if (and (eq inlinep :notinline)
+                                         (neq where :declared))
+                                    (specifier-type 'function)
+                                    (proclaimed-ftype name))))
+                       (find-global-fun name nil)))))))))
 
 ;;; Return the LEAF structure for the lexically apparent function
 ;;; definition of NAME.
@@ -379,20 +379,14 @@
                     ;; user-defined MAKE-LOAD-FORM methods?
                     (when (emit-make-load-form value)
                       #+sb-xc-host
-                      (aver (zerop (layout-raw-slot-metadata
-                                    (%instance-layout value))))
+                      (aver (eql (layout-bitmap (%instance-layout value))
+                                 sb!kernel::+layout-all-tagged+))
                       (do-instance-tagged-slot (i value)
                         (grovel (%instance-ref value i)))))
-                   ;; The cross-compiler can dump certain instances that are not
-                   ;; subtypes of STRUCTURE!OBJECT, as long as it has processed
-                   ;; the defstruct.
-                   #+sb-xc-host
-                   ((satisfies sb!kernel::xc-dumpable-structure-instance-p)
-                    (do-instance-tagged-slot (i value)
-                      (grovel (%instance-ref value i))))
                    (t
                     (compiler-error
-                     "Objects of type ~S can't be dumped into fasl files."
+                      "Objects of type ~/sb!impl:print-type-specifier/ ~
+                       can't be dumped into fasl files."
                      (type-of value)))))))
       ;; Dump all non-trivial named constants using the name.
       (if (and namep (not (typep constant '(or symbol character
@@ -689,7 +683,7 @@
     (:always-bound t)
     ;; Compiling to fasl considers a symbol always-bound if its
     ;; :always-bound info value is now T or will eventually be T.
-    (:eventually (fasl-output-p *compile-object*))))
+    (:eventually (producing-fasl-file))))
 
 ;;; Convert a reference to a symbolic constant or variable. If the
 ;;; symbol is entered in the LEXENV-VARS we use that definition,
@@ -866,10 +860,6 @@
                           (find-free-fun fun "shouldn't happen! (no-cmacro)")
                           form))))
 
-(defun muffle-warning-or-die ()
-  (muffle-warning)
-  (bug "no MUFFLE-WARNING restart"))
-
 ;;; Expand FORM using the macro whose MACRO-FUNCTION is FUN, trapping
 ;;; errors which occur during the macroexpansion.
 (defun careful-expand-macro (fun form &optional cmacro)
@@ -918,7 +908,8 @@
                                    ignoring anything  horrible here..)~:@>~:>"
                                (wherestring)
                                c)
-                              (muffle-warning-or-die)))
+                              (muffle-warning)
+                              (bug "no MUFFLE-WARNING restart")))
                    (error
                      (lambda (c)
                        (cond
@@ -960,6 +951,18 @@
 
 ;;;; code coverage
 
+;;; Used as the CDR of the code coverage instrumentation records
+;;; (instead of NIL) to ensure that any well-behaving user code will
+;;; not have constants EQUAL to that record. This avoids problems with
+;;; the records getting coalesced with non-record conses, which then
+;;; get mutated when the instrumentation runs. Note that it's
+;;; important for multiple records for the same location to be
+;;; coalesced. -- JES, 2008-01-02
+;;; Use of #. mandates :COMPILE-TOPLEVEL for several Lisps
+;;; even though for us it's immediately accessible to EVAL.
+(eval-when (:compile-toplevel :load-toplevel :execute)
+  (defconstant +code-coverage-unmarked+ '%code-coverage-unmarked%))
+
 ;;; Check the policy for whether we should generate code coverage
 ;;; instrumentation. If not, just return the original START
 ;;; ctran. Otherwise insert code coverage instrumentation after
@@ -990,17 +993,16 @@
                   (*allow-instrumenting* nil))
               (push (ctran-block start)
                     (gethash path *code-coverage-blocks*))
-              (let ((*allow-instrumenting* nil))
-                (ir1-convert start next nil
-                             `(locally
-                                  (declare (optimize speed
-                                                     (safety 0)
-                                                     (debug 0)
-                                                     (check-constant-modification 0)))
-                                ;; We're being naughty here, and
-                                ;; modifying constant data. That's ok,
-                                ;; we know what we're doing.
-                                (%rplacd ',store t))))
+              (ir1-convert start next nil
+                           `(locally
+                                (declare (optimize speed
+                                                   (safety 0)
+                                                   (debug 0)
+                                                   (check-constant-modification 0)))
+                              ;; We're being naughty here, and
+                              ;; modifying constant data. That's ok,
+                              ;; we know what we're doing.
+                              (%rplacd ',store t)))
               next)))
       start))
 
@@ -1252,7 +1254,7 @@
       (dolist (var-name (rest decl))
         (unless (symbolp var-name)
           (compiler-error "Variable name is not a symbol: ~S." var-name))
-        (when (boundp var-name)
+        (unless (eq (info :variable :kind var-name) :unknown)
           (program-assert-symbol-home-package-unlocked
            context var-name "declaring the type of ~A"))
         (let* ((bound-var (find-in-bindings vars var-name))
@@ -1275,12 +1277,12 @@
                                (warn
                                 'type-warning
                                 :format-control
-                                "The type declarations ~S and ~S for ~S conflict."
+                                 "The type declarations ~
+                                  ~/sb!impl:print-type/ and ~
+                                  ~/sb!impl:print-type/ for ~
+                                  ~S conflict."
                                 :format-arguments
-                                (list
-                                 (type-specifier old-type)
-                                 (type-specifier type)
-                                 var-name))))
+                                (list old-type type var-name))))
                             (bound-var
                              (setf (leaf-type bound-var) int
                                    (leaf-where-from bound-var) :declared))
@@ -1325,7 +1327,10 @@
         (when (fboundp name)
           (program-assert-symbol-home-package-unlocked
            context name "declaring the ftype of ~A"))
-        (let ((found (find name fvars :key #'leaf-source-name :test #'equal)))
+        (let ((found (find name fvars :key (lambda (x)
+                                             (unless (consp x) ;; macrolet
+                                               (leaf-source-name x)))
+                                      :test #'equal)))
           (cond
            (found
             (setf (leaf-type found) type)
@@ -1357,7 +1362,9 @@
       ;; for GLOBAL variables.
       (let ((kind (info :variable :kind name)))
         (unless (member kind '(:special :unknown))
-          (error "Can't declare ~(~A~) variable locally special: ~S" kind name)))
+          (compiler-error
+           "Can't declare ~(~A~) variable locally special: ~S"
+           kind name)))
       (program-assert-symbol-home-package-unlocked
        context name "declaring ~A special")
       (let ((var (find-in-bindings vars name)))
@@ -1418,7 +1425,11 @@
   (let ((sense (cdr (assoc (first spec) *inlinep-translations* :test #'eq)))
         (new-fenv ()))
     (dolist (name (rest spec))
-      (let ((fvar (find name fvars :key #'leaf-source-name :test #'equal)))
+      (let ((fvar (find name fvars
+                        :key (lambda (x)
+                               (unless (consp x) ;; macrolet
+                                 (leaf-source-name x)))
+                        :test #'equal)))
         (if fvar
             (setf (functional-inlinep fvar) sense)
             (let ((found (find-lexically-apparent-fun
@@ -1445,7 +1456,10 @@
     (atom
      (find-in-bindings vars name))
     ((cons (eql function) (cons * null))
-     (find (cadr name) fvars :key #'leaf-source-name :test #'equal))
+     (find (cadr name) fvars :key (lambda (x)
+                                    (if (consp x) ;; MACROLET
+                                        (car x)
+                                        (leaf-source-name x))) :test #'equal))
     (t
      (compiler-error "Malformed function or variable name ~S." name))))
 
@@ -1486,9 +1500,12 @@
                       (t
                        "an unknown function"))
                 (second name)))))
-        ((and (consp var) (eq (car var) 'macro))
+        ((and (consp var)
+              (or (eq (car var) 'macro)
+                  (and (consp (cdr var))
+                       (eq (cadr var) 'macro))))
          ;; Just ignore the IGNORE decl: we don't currently signal style-warnings
-         ;; for unused symbol-macros, so there's no need to do anything.
+         ;; for unused macrolet or symbol-macros, so there's no need to do anything.
          )
         ((functional-p var)
          (setf (leaf-ever-used var) t))
@@ -1544,7 +1561,9 @@
                   (neq :indefinite extent))
              (let* ((fname (cadr name))
                     (bound-fun (find fname fvars
-                                     :key #'leaf-source-name
+                                     :key (lambda (x)
+                                            (unless (consp x) ;; macrolet
+                                              (leaf-source-name x)))
                                      :test #'equal))
                     (fun (or bound-fun (lexenv-find fname funs))))
                (etypecase fun
@@ -1570,8 +1589,10 @@
 
 ;;; FIXME: This is non-ANSI, so the default should be T, or it should
 ;;; go away, I think.
+;;; Or just rename the declaration to SB-C:RESULT-TYPE so that it's not
+;;; a symbol in the CL package, and then eliminate this switch.
+;;; It's permissible to have implementation-specific declarations.
 (defvar *suppress-values-declaration* nil
-  #!+sb-doc
   "If true, processing of the VALUES declaration is inhibited.")
 
 ;;; Process a single declaration spec, augmenting the specified LEXENV
@@ -1581,10 +1602,19 @@
   (declare (type list raw-spec vars fvars))
   (declare (type lexenv res))
   (let ((spec (canonized-decl-spec raw-spec))
-        (optimize-qualities)
-        (result-type *wild-type*))
+        (optimize-qualities))
+    ;; FIXME: we can end up with a chain of spurious parent lexenvs,
+    ;; when logically the processing of decls should yield at most
+    ;; two new lexenvs: one for the bindings and one for post-binding.
+    ;; It's possible that there's a simple fix of re-linking the resulting
+    ;; lexenv directly to *lexenv* as its parent.
     (values
      (case (first spec)
+       (type
+        (process-type-decl (cdr spec) res vars context))
+       ((ignore ignorable)
+        (process-ignore-decl spec vars fvars res)
+        res)
        (special (process-special-decl spec res vars binding-form-p context))
        (ftype
         (unless (cdr spec)
@@ -1592,9 +1622,6 @@
         (process-ftype-decl (second spec) res (cddr spec) fvars context))
        ((inline notinline maybe-inline)
         (process-inline-decl spec res fvars))
-       ((ignore ignorable)
-        (process-ignore-decl spec vars fvars res)
-        res)
        (optimize
         (multiple-value-bind (new-policy specified-qualities)
             (process-optimize-decl spec (lexenv-policy res))
@@ -1610,17 +1637,6 @@
          :default res
          :handled-conditions (process-unmuffle-conditions-decl
                               spec (lexenv-handled-conditions res))))
-       (type
-        (process-type-decl (cdr spec) res vars context))
-       (values
-        (unless *suppress-values-declaration*
-          (let ((types (cdr spec)))
-            (setq result-type
-                  (compiler-values-specifier-type
-                   (if (singleton-p types)
-                       (car types)
-                       `(values ,@types)))))
-          res))
        ((dynamic-extent truly-dynamic-extent indefinite-extent)
         (process-extent-decl (cdr spec) vars fvars (first spec))
         res)
@@ -1629,6 +1645,8 @@
          :default res
          :disabled-package-locks (process-package-lock-decl
                                   spec (lexenv-disabled-package-locks res))))
+       ;; We may want to detect LAMBDA-LIST and VALUES decls here,
+       ;; and report them as "Misplaced" rather than "Unrecognized".
        (t
         (unless (info :declaration :recognized (first spec))
           (compiler-warn "unrecognized declaration ~S" raw-spec))
@@ -1636,7 +1654,6 @@
           (if fn
               (funcall fn res spec vars fvars)
               res))))
-     result-type
      optimize-qualities)))
 
 ;;; Use a list of DECLARE forms to annotate the lists of LAMBDA-VAR
@@ -1655,38 +1672,65 @@
                       (allow-lambda-list nil))
   (declare (list decls vars fvars))
   (let ((result-type *wild-type*)
+        (allow-values-decl allow-lambda-list)
+        (explicit-check)
+        (allow-explicit-check allow-lambda-list)
         (lambda-list (if allow-lambda-list :unspecified nil))
         (optimize-qualities)
         (*post-binding-variable-lexenv* nil))
-    (dolist (decl decls)
-      (dolist (spec (rest decl))
-        (flet
-            ((process-it ()
-               (unless (consp spec)
-                 (compiler-error "malformed declaration specifier ~S in ~S"
-                                 spec decl))
-               (if (and (typep spec '(cons (eql lambda-list) (cons t null)))
-                        (eq allow-lambda-list t))
-                   (setq lambda-list (cadr spec) allow-lambda-list nil)
-                   (multiple-value-bind (new-env new-result-type new-qualities)
-                       (process-1-decl spec lexenv vars fvars
-                                       binding-form-p context)
-                     (setq lexenv new-env
-                           optimize-qualities
-                           (nconc new-qualities optimize-qualities))
-                     (unless (eq new-result-type *wild-type*)
-                       (setq result-type
-                             (values-type-intersection result-type
-                                                       new-result-type)))))))
+    (flet ((process-it (spec decl)
+             (cond ((atom spec)
+                    (compiler-error "malformed declaration specifier ~S in ~S"
+                                    spec decl))
+                   ((and (eq allow-lambda-list t)
+                         (typep spec '(cons (eql lambda-list) (cons t null))))
+                    (setq lambda-list (cadr spec) allow-lambda-list nil))
+                   ((and allow-values-decl
+                         (typep spec '(cons (eql values)))
+                         (not *suppress-values-declaration*))
+                    ;; Why do we allow more than one VALUES decl? I don't know.
+                    (setq result-type
+                          (values-type-intersection
+                           result-type
+                           (compiler-values-specifier-type
+                            (let ((types (cdr spec)))
+                              (if (singleton-p types)
+                                  (car types)
+                                  `(values ,@types)))))))
+                   ((and allow-explicit-check
+                         (typep spec '(cons (eql explicit-check))))
+                    ;; EXPLICIT-CHECK by itself specifies that all argument and
+                    ;; result types are checked by the function body.
+                    ;; Alternatively, a subset of arguments, and/or :RESULT,
+                    ;; can be specified to indicate that only a subset are
+                    ;; checked; in that case, the compiler asserts all others.
+                    (awhen (remove-if (lambda (x)
+                                        (or (member x vars
+                                                    :test #'eq
+                                                    :key #'lambda-var-%source-name)
+                                            (eq x :result)))
+                                      (cdr spec))
+                      (compiler-error "explicit-check list ~S must refer to lambda vars"
+                                      it))
+                    (setq explicit-check (or (cdr spec) t)
+                          allow-explicit-check nil)) ; at most one of this decl
+                   (t
+                    (multiple-value-bind (new-env new-qualities)
+                        (process-1-decl spec lexenv vars fvars
+                                        binding-form-p context)
+                      (setq lexenv new-env
+                            optimize-qualities
+                            (nconc new-qualities optimize-qualities)))))))
+      (dolist (decl decls)
+        (dolist (spec (rest decl))
           (if (eq context :compile)
-              (let ((*current-path* (or (get-source-path spec)
-                                        (get-source-path decl)
-                                        *current-path*)))
-                (process-it))
-            ;; Kludge: EVAL calls this function to deal with LOCALLY.
-              (process-it)))))
+              (with-current-source-form (spec decl) ; TODO this is a slight change to the previous code. make sure the behavior is identical
+                (process-it spec decl))
+              ;; Kludge: EVAL calls this function to deal with LOCALLY.
+              (process-it spec decl)))))
     (warn-repeated-optimize-qualities (lexenv-policy lexenv) optimize-qualities)
-    (values lexenv result-type *post-binding-variable-lexenv* lambda-list)))
+    (values lexenv result-type *post-binding-variable-lexenv*
+            lambda-list explicit-check)))
 
 (defun %processing-decls (decls vars fvars ctran lvar binding-form-p fun)
   (multiple-value-bind (*lexenv* result-type post-binding-lexenv)
