@@ -29,29 +29,6 @@
       (combination-step-info node)
       nil))
 
-;;; If there is any CHECK-xxx template for TYPE, then return it,
-;;; otherwise return NIL.
-(defun type-check-template (type)
-  (declare (type ctype type))
-  (multiple-value-bind (check-ptype exact) (primitive-type type)
-    (if exact
-        (primitive-type-check check-ptype)
-        (let ((name (hairy-type-check-template-name type)))
-          (if name
-              (template-or-lose name)
-              nil)))))
-
-;;; Emit code in BLOCK to check that VALUE is of the specified TYPE,
-;;; yielding the checked result in RESULT. VALUE and result may be of
-;;; any primitive type. There must be CHECK-xxx VOP for TYPE. Any
-;;; other type checks should have been converted to an explicit type
-;;; test.
-(defun emit-type-check (node block value result type)
-  (declare (type tn value result) (type node node) (type ir2-block block)
-           (type ctype type))
-  (emit-move-template node block (type-check-template type) value result)
-  (values))
-
 ;;; Allocate an indirect value cell.
 (defevent make-value-cell-event "Allocate heap value cell for lexical var.")
 (defun emit-make-value-cell (node block value res)
@@ -61,7 +38,7 @@
 ;;;; leaf reference
 
 ;;; Return the TN that holds the value of THING in the environment ENV.
-(declaim (ftype (function ((or nlx-info lambda-var clambda) physenv) tn)
+(declaim (ftype (sfunction ((or nlx-info lambda-var clambda) physenv) tn)
                 find-in-physenv))
 (defun find-in-physenv (thing physenv)
   (or (cdr (assoc thing (ir2-physenv-closure (physenv-info physenv))))
@@ -119,21 +96,61 @@
 (defun emit-constant (value)
   (constant-tn (find-constant value) t))
 
-(defun %return-is-boxed (node)
-  (declare (type creturn node))
-  (let* ((fun (return-lambda node))
-         (returns (tail-set-info (lambda-tail-set fun))))
-    (or (xep-p fun)
-        (eq (return-info-kind returns) :unknown))))
+(defun boxed-combination-ref-p (combination lvar)
+  (let ((args (combination-args combination)))
+    (flet ((struct-slot-raw-p (dd index)
+             (let ((slot (and dd
+                              (find index (dd-slots dd) :key #'dsd-index))))
+               (and slot
+                    (= (sb!kernel::dsd-%raw-type slot) -1))))
+           (constant (lvar)
+             (and (constant-lvar-p lvar)
+                  (lvar-value lvar))))
+      (case (combination-fun-source-name combination nil)
+        (data-vector-set-with-offset
+         (and (eq lvar (car (last args)))
+              (csubtypep (lvar-type (car args))
+                         (specifier-type '(array t)))))
+        (initialize-vector
+         (csubtypep (lvar-type (car args))
+                    (specifier-type '(array t))))
+        (%make-structure-instance
+         (let* ((dd (constant (car args)))
+                (slot-specs (constant (cadr args)))
+                (slot (position lvar (cddr args)))
+                (slot (and slot
+                           (nth slot slot-specs))))
+           (struct-slot-raw-p dd
+                              (cddr slot))))
+        (%instance-set
+         (let* ((instance (lvar-type (car args)))
+                (layout (and (structure-classoid-p instance)
+                             (classoid-layout instance))))
+           (and (eq lvar (car (last args)))
+                (struct-slot-raw-p (and layout
+                                        (layout-info layout))
+                                   (constant (cadr args))))))
+        ((%special-bind %set-sap-ref-lispobj
+          %rplaca %rplacd cons list list*
+          values)
+         t)
+        (t
+         (call-full-like-p combination))))))
 
 (defun boxed-ref-p (ref)
-  (let ((dest (lvar-dest (ref-lvar ref))))
-    (cond ((and (basic-combination-p dest)
-                (call-full-like-p dest))
-           t)
-          ((and (return-p dest) (%return-is-boxed dest)))
-          (t
-           nil))))
+  (let* ((lvar (ref-lvar ref))
+         (dest (lvar-dest lvar)))
+    (cond ((basic-combination-p dest)
+           (if (combination-p dest)
+               (boxed-combination-ref-p dest lvar)
+               (call-full-like-p dest)))
+          ((set-p dest)
+           (global-var-p (set-var dest)))
+          ((return-p dest)
+           (let* ((fun (return-lambda dest))
+                  (returns (tail-set-info (lambda-tail-set fun))))
+             (or (xep-p fun)
+                 (eq (return-info-kind returns) :unknown)))))))
 
 ;;; Convert a REF node. The reference must not be delayed.
 (defun ir2-convert-ref (node block)
@@ -562,27 +579,6 @@
           dest))
   (values))
 
-;;; Move each SRC TN into the corresponding DEST TN, checking types
-;;; and defaulting any unsupplied source values to NIL
-(defun move-results-checked (node block src dest types)
-  (declare (type node node) (type ir2-block block) (list src dest types))
-  (let ((nsrc (length src))
-        (ndest (length dest))
-        (ntypes (length types)))
-    (mapc (lambda (from to type)
-            (if type
-                (emit-type-check node block from to type)
-                (emit-move node block from to)))
-          (if (> ndest nsrc)
-              (append src (make-list (- ndest nsrc)
-                                     :initial-element (emit-constant nil)))
-              src)
-          dest
-          (if (> ndest ntypes)
-              (append types (make-list (- ndest ntypes)))
-              types)))
-  (values))
-
 ;;; If necessary, emit coercion code needed to deliver the RESULTS to
 ;;; the specified lvar. NODE and BLOCK provide context for emitting
 ;;; code. Although usually obtained from STANDARD-RESULT-TNs or
@@ -621,27 +617,25 @@
              (2lvar (lvar-info lvar))
              (value (cast-value node))
              (2value (lvar-info value)))
-    (cond ((eq (ir2-lvar-kind 2lvar) :unused))
-          ((eq (ir2-lvar-kind 2lvar) :unknown)
-           (aver (eq (ir2-lvar-kind 2value) :unknown))
-           (aver (not (cast-type-check node)))
-           (move-results-coerced node block
-                                 (ir2-lvar-locs 2value)
-                                 (ir2-lvar-locs 2lvar)))
-          ((eq (ir2-lvar-kind 2lvar) :fixed)
-           (aver (eq (ir2-lvar-kind 2value) :fixed))
-           (if (cast-type-check node)
-               (move-results-checked node block
-                                     (ir2-lvar-locs 2value)
-                                     (ir2-lvar-locs 2lvar)
-                                     (multiple-value-bind (check types)
-                                         (cast-check-types node nil)
-                                       (aver (eq check :simple))
-                                       types))
-               (move-results-coerced node block
-                                     (ir2-lvar-locs 2value)
-                                     (ir2-lvar-locs 2lvar))))
-          (t (bug "CAST cannot be :DELAYED.")))))
+    (ecase (ir2-lvar-kind 2lvar)
+      (:unused)
+      ((:unknown :fixed)
+       (aver (not (cast-type-check node)))
+       (move-results-coerced node block
+                             (ir2-lvar-locs 2value)
+                             (ir2-lvar-locs 2lvar))))))
+
+(defoptimizer (%check-bound ir2-hook) ((array bound index) node block)
+  (declare (ignore block))
+  (when (constant-lvar-p bound)
+    (let* ((bound-type (specifier-type `(integer 0 (,(lvar-value bound)))))
+           (index-type (lvar-type index)))
+      (when (eq (type-intersection bound-type index-type)
+                *empty-type*)
+        (let ((*compiler-error-context* node))
+          (compiler-warn "Derived type ~s is not a suitable index for ~s."
+                         (type-specifier index-type)
+                         (type-specifier (lvar-type array))))))))
 
 ;;;; template conversion
 
@@ -729,9 +723,9 @@
                               (args-type-optional type))
                       (list type))))
          (primitive-t *backend-t-primitive-type*))
-    (loop for rtype in rtypes
-          for type = (or (pop types) primitive-t)
-          collect type)))
+    (mapcar (lambda (rtype)
+              (declare (ignore rtype))
+              (or (pop types) primitive-t)) rtypes)))
 
 ;;; Return a list of TNs usable in a CALL to TEMPLATE delivering values to
 ;;; LVAR. As an efficiency hack, we pick off the common case where the LVAR is
@@ -1046,7 +1040,7 @@
     (if (eq (ir2-lvar-kind 2lvar) :delayed)
         (let ((name (lvar-fun-name lvar t)))
           (aver name)
-          (values (make-load-time-constant-tn :fdefinition name) t))
+          (values (make-load-time-constant-tn :fdefinition name) name))
         (let* ((locs (ir2-lvar-locs 2lvar))
                (loc (first locs))
                (function-ptype (primitive-type-or-lose 'function)))
@@ -1086,17 +1080,17 @@
     (multiple-value-bind (fun-tn named)
         (fun-lvar-tn node block (basic-combination-fun node))
       (if named
+          ;; Named call to an immobile fdefn from an immobile component
+          ;; uses the FUN-TN only to preserve liveness of the fdefn.
+          ;; The name becomes an info arg.
           (vop* tail-call-named node block
-                (fun-tn old-fp return-pc pass-refs)
-                (nil)
-                nargs
-                (emit-step-p node))
+                (#!-immobile-code fun-tn old-fp return-pc pass-refs)  ; args
+                (nil)                                                 ; results
+                nargs #!+immobile-code named (emit-step-p node))      ; info
           (vop* tail-call node block
                 (fun-tn old-fp return-pc pass-refs)
                 (nil)
-                nargs
-                (emit-step-p node)))))
-
+                nargs (emit-step-p node)))))
   (values))
 
 ;;; like IR2-CONVERT-LOCAL-CALL-ARGS, only different
@@ -1134,8 +1128,11 @@
       (multiple-value-bind (fun-tn named)
           (fun-lvar-tn node block (basic-combination-fun node))
         (if named
-            (vop* call-named node block (fp fun-tn args) (loc-refs)
-                  arg-locs nargs nvals (emit-step-p node))
+            (vop* call-named node block
+                  (fp #!-immobile-code fun-tn args)           ; args
+                  (loc-refs)                                  ; results
+                  arg-locs nargs #!+immobile-code named nvals ; info
+                  (emit-step-p node))
             (vop* call node block (fp fun-tn args) (loc-refs)
                   arg-locs nargs nvals (emit-step-p node)))
         (move-lvar-result node block locs lvar))))
@@ -1152,8 +1149,11 @@
       (multiple-value-bind (fun-tn named)
           (fun-lvar-tn node block (basic-combination-fun node))
         (if named
-            (vop* multiple-call-named node block (fp fun-tn args) (loc-refs)
-                  arg-locs nargs (emit-step-p node))
+            (vop* multiple-call-named node block
+                  (fp #!-immobile-code fun-tn args)     ; args
+                  (loc-refs)                            ; results
+                  arg-locs nargs #!+immobile-code named ; info
+                  (emit-step-p node))
             (vop* multiple-call node block (fp fun-tn args) (loc-refs)
                   arg-locs nargs (emit-step-p node))))))
   (values))
@@ -1182,6 +1182,19 @@
          (fname (lvar-fun-name lvar t)))
     (declare (type (or symbol cons) fname))
 
+    ;; Warn about cross-compiling certain full-calls,
+    ;; as it is indicative of dependency order problems.
+    #+sb-xc-host
+    (let ((compname (component-name (node-component node))))
+      ;; Don't care too much about macro performance.
+      (unless (and (stringp compname) (string/= compname "DEFMACRO"))
+        ;; Catch FOO and (SETF FOO) both.
+        (let ((stem (if (atom fname) fname (second fname))))
+          (when (member stem
+                        sb-cold::*full-calls-to-warn-about*
+                        :test #'string=)
+            (warn "Full call to ~S" fname)))))
+
     (let* ((inlineable-p (not (let ((*lexenv* (node-lexenv node)))
                                 (fun-lexically-notinline-p fname))))
            (inlineable-bit (if inlineable-p 1 0))
@@ -1274,7 +1287,8 @@
         (vop verify-arg-count node block
              arg-count-location
              min
-             max)))))
+             max)
+        min))))
 
 ;;; Do all the stuff that needs to be done on XEP entry:
 ;;; -- Create frame.
@@ -1291,18 +1305,21 @@
       (vop xep-allocate-frame node block start-label)
       ;; Arg verification needs to be done before the stack pointer is adjusted
       ;; so that the extra arguments are still present when the error is signalled
-      (unless (eq (functional-kind fun) :toplevel)
-        (setf arg-count-tn (make-arg-count-location))
-        #!+precise-arg-count-error
-        (xep-verify-arg-count node block fun arg-count-tn))
-      (cond ((and (optional-dispatch-p ef) (optional-dispatch-more-entry ef))
-             ;; COPY-MORE-ARG should handle SP adjustemnt, but it
-             ;; isn't done on all targets.
-             #!-precise-arg-count-error
-             (vop xep-setup-sp node block)
-             (vop copy-more-arg node block (optional-dispatch-max-args ef)))
-            (t
-             (vop xep-setup-sp node block)))
+      (let ((verified (unless (eq (functional-kind fun) :toplevel)
+                        (setf arg-count-tn (make-arg-count-location))
+                        #!+precise-arg-count-error
+                        (xep-verify-arg-count node block fun arg-count-tn))))
+        #!-x86-64
+        (declare (ignore verified))
+       (cond ((and (optional-dispatch-p ef) (optional-dispatch-more-entry ef))
+              ;; COPY-MORE-ARG should handle SP adjustemnt, but it
+              ;; isn't done on all targets.
+              #!-precise-arg-count-error
+              (vop xep-setup-sp node block)
+              (vop copy-more-arg node block (optional-dispatch-max-args ef)
+                   #!+x86-64 verified))
+             (t
+              (vop xep-setup-sp node block))))
       (when (ir2-physenv-closure env)
         (let ((closure (make-normal-tn *backend-t-primitive-type*)))
           (when (policy fun (> store-closure-debug-pointer 1))
@@ -1490,8 +1507,8 @@
 ;;; lvar.
 (defun ir2-convert-mv-bind (node block)
   (declare (type mv-combination node) (type ir2-block block))
-  (let* ((lvar (first (basic-combination-args node)))
-         (fun (ref-leaf (lvar-uses (basic-combination-fun node))))
+  (let* ((fun (ref-leaf (lvar-uses (basic-combination-fun node))))
+         (args (basic-combination-args node))
          (vars (lambda-vars fun)))
     (aver (eq (functional-kind fun) :mv-let))
     (mapc (lambda (src var)
@@ -1501,10 +1518,19 @@
                          (lambda-var-explicit-value-cell var))
                     (emit-make-value-cell node block src dest)
                     (emit-move node block src dest)))))
-          (lvar-tns node block lvar
-                            (mapcar (lambda (x)
-                                      (primitive-type (leaf-type x)))
-                                    vars))
+          (if (singleton-p args)
+              (lvar-tns node block (first args)
+                        (mapcar (lambda (x)
+                                  (primitive-type (leaf-type x)))
+                                vars))
+              (let ((vars vars))
+                (loop for lvar in args
+                      for values = (nth-value 1 (values-types
+                                                 (lvar-derived-type lvar)))
+                      while vars
+                      nconc
+                      (lvar-tns node block lvar (loop repeat values
+                                                      collect (primitive-type (leaf-type (pop vars))))))))
           vars))
   (values))
 
@@ -1620,6 +1646,7 @@ not stack-allocated LVAR ~S." source-lvar)))))
   (let ((tns (mapcar (lambda (x)
                        (lvar-tn node block x))
                      values)))
+
     (move-lvar-result node block tns (node-lvar node))))
 
 ;;; In the normal case where unknown values are desired, we use the
@@ -1671,18 +1698,30 @@ not stack-allocated LVAR ~S." source-lvar)))))
 ;;; implementation.
 (defoptimizer (%special-bind ir2-convert) ((var value) node block)
   (let ((name (leaf-source-name (lvar-value var))))
-    #!-(and sb-thread x86-64)
-    (vop bind node block (lvar-tn node block value) (emit-constant name))
-    #!+(and sb-thread x86-64)
-    (progn
-      ;; GC must understand that the symbol is implicitly live even though
-      ;; binding makes no references to the object.
-      (emit-constant name)
-      (vop sb!vm::bind/let node block (lvar-tn node block value) name))))
+    ;; Emit either BIND or DYNBIND, preferring BIND if both exist.
+    ;; If only one exists, it's DYNBIND.
+    ;; Even if the backend supports load-time TLS index assignment,
+    ;; there might be only one vop (as with arm64).
+    (macrolet ((doit (bind dynbind)
+                 (if (gethash 'bind *backend-parsed-vops*) bind dynbind)))
+      (doit
+       (progn
+         ;; Inform later SYMBOL-VALUE calls that they can
+         ;; assume a nonzero tls-index.
+         ;; FIXME: setting INFO is inefficient when not actually
+         ;; changing anything
+         (unless (info :variable :wired-tls name)
+           (setf (info :variable :wired-tls name) :always-has-tls))
+         ;; We force the symbol into the code constants in case BIND
+         ;; does not actually reference it, as with x86.
+         (emit-constant name)
+         (vop bind node block (lvar-tn node block value) name))
+       (vop dynbind node block (lvar-tn node block value)
+            (emit-constant name))))))
 
-(defoptimizer (%special-unbind ir2-convert) ((var) node block)
-  (declare (ignore var))
-  (vop unbind node block))
+(defoptimizer (%special-unbind ir2-convert) ((n) node block)
+  (declare (ignorable n))
+  (vop unbind node block #!+(and sb-thread unbind-n-vop) (lvar-value n)))
 
 ;;; ### It's not clear that this really belongs in this file, or
 ;;; should really be done this way, but this is the least violation of
@@ -1703,7 +1742,7 @@ not stack-allocated LVAR ~S." source-lvar)))))
                               ;; CLHS says "bound and then made to have no value" -- user
                               ;; should not be able to tell the difference between that and this.
                               (about-to-modify-symbol-value var 'progv)
-                              (%primitive bind unbound-marker var))))
+                              (%primitive dynbind unbound-marker var))))
                         (,bind (vars vals)
                           (declare (optimize (speed 2) (debug 0)
                                              (insert-debug-catch 0)))
@@ -1713,7 +1752,7 @@ not stack-allocated LVAR ~S." source-lvar)))))
                                  (let ((val (car vals))
                                        (var (car vars)))
                                    (about-to-modify-symbol-value var 'progv val t)
-                                   (%primitive bind val var))
+                                   (%primitive dynbind val var))
                                  (,bind (cdr vars) (cdr vals))))))
                  (,bind ,vars ,vals))
                nil
@@ -1791,7 +1830,13 @@ not stack-allocated LVAR ~S." source-lvar)))))
   (let* ((2info (nlx-info-info info))
          (kind (cleanup-kind (nlx-info-cleanup info)))
          (block-tn (physenv-live-tn
-                    (make-normal-tn (primitive-type-or-lose 'catch-block))
+                    (make-normal-tn
+                     (primitive-type-or-lose
+                      (ecase kind
+                        (:catch
+                         'catch-block)
+                        ((:unwind-protect :block :tagbody)
+                         'unwind-block))))
                     (node-physenv node)))
          (res (make-stack-pointer-tn))
          (target-label (ir2-nlx-info-target 2info)))
@@ -1909,23 +1954,32 @@ not stack-allocated LVAR ~S." source-lvar)))))
 
 (macrolet ((def (name)
              `(defoptimizer (,name ir2-convert) ((&rest args) node block)
-                (let* ((refs (reference-tn-list
-                              (loop for arg in args
-                                    for tn = (make-normal-tn *backend-t-primitive-type*)
-                                    do
-                                    (emit-move node block (lvar-tn node block arg) tn)
-                                    collect tn)
-                              nil))
-                       (lvar (node-lvar node))
-                       (res (lvar-result-tns
-                             lvar
-                             (list (primitive-type (specifier-type 'list))))))
-                  (when (and lvar (lvar-dynamic-extent lvar))
-                    (vop current-stack-pointer node block
-                         (ir2-lvar-stack-pointer (lvar-info lvar))))
-                  (vop* ,name node block (refs) ((first res) nil)
-                        (length args))
-                  (move-lvar-result node block res lvar)))))
+                (cond #!+gencgc
+                      ((>= (length args)
+                           (/ sb!vm:large-object-size
+                              (* sb!vm:n-word-bytes 2)))
+                       ;; The VOPs will try to allocate all space at once
+                       ;; And it'll end up in large objects, and no conses
+                       ;; are welcome there.
+                       (ir2-convert-full-call node block))
+                      (t
+                       (let* ((refs (reference-tn-list
+                                     (loop for arg in args
+                                           for tn = (make-normal-tn *backend-t-primitive-type*)
+                                           do
+                                           (emit-move node block (lvar-tn node block arg) tn)
+                                           collect tn)
+                                     nil))
+                              (lvar (node-lvar node))
+                              (res (lvar-result-tns
+                                    lvar
+                                    (list (primitive-type (specifier-type 'list))))))
+                         (when (and lvar (lvar-dynamic-extent lvar))
+                           (vop current-stack-pointer node block
+                                (ir2-lvar-stack-pointer (lvar-info lvar))))
+                         (vop* ,name node block (refs) ((first res) nil)
+                               (length args))
+                         (move-lvar-result node block res lvar)))))))
   (def list)
   (def list*))
 
@@ -1979,6 +2033,29 @@ not stack-allocated LVAR ~S." source-lvar)))))
          (results (lvar-result-tns lvar (list (primitive-type-or-lose t)))))
     (emit-move node block (lvar-tn node block value) (first results))
     (move-lvar-result node block results lvar)))
+
+;;; An identity to avoid complaints about constant modification
+(defoptimizer (ltv-wrapper ir2-convert) ((x) node block)
+  (let* ((lvar (node-lvar node))
+         (results (lvar-result-tns lvar (list (primitive-type-or-lose t)))))
+    (emit-move node block (lvar-tn node block x) (first results))
+    (move-lvar-result node block results lvar)))
+
+#-sb-xc-host ;; package-lock-violation-p is not present yet
+(defoptimizer (set ir2-hook) ((symbol value) node block)
+  (declare (ignore value block))
+  (when (constant-lvar-p symbol)
+    (let* ((symbol (lvar-value symbol))
+           (kind (info :variable :kind symbol)))
+      (when (and (eq kind :unknown)
+                 (sb!impl::package-lock-violation-p (symbol-package symbol) symbol))
+        (let ((*compiler-error-context* node))
+          (compiler-warn "violating package lock on ~/sb-impl:print-symbol-with-prefix/"
+                         symbol))))))
+
+(defoptimizer (restart-point ir2-convert) ((location) node block)
+  (setf (restart-location-label (lvar-value location))
+        (block-label (ir2-block-block block))))
 
 ;;; Convert the code in a component into VOPs.
 (defun ir2-convert (component)
@@ -2070,7 +2147,7 @@ not stack-allocated LVAR ~S." source-lvar)))))
                                    (leaf-has-source-name-p (ref-leaf use))
                                    (leaf-source-name (ref-leaf use))))
                         (ftype (and (info :function :info name) ; only use the FTYPE if
-                                    (info :function :type name)))) ; NAME was DEFKNOWN
+                                    (proclaimed-ftype name)))) ; NAME was DEFKNOWN
                    (unless (or (node-tail-p last)
                                (policy last (zerop safety))
                                (and (fun-type-p ftype)
@@ -2110,7 +2187,10 @@ not stack-allocated LVAR ~S." source-lvar)))))
               (ir2-convert-full-call node 2block))
              (:known
               (let* ((info (basic-combination-fun-info node))
-                     (fun (fun-info-ir2-convert info)))
+                     (fun (fun-info-ir2-convert info))
+                     (hook (fun-info-ir2-hook info)))
+                (when hook
+                  (funcall hook node 2block))
                 (cond (fun
                        (funcall fun node 2block))
                       ((eq (basic-combination-info node) :full)
