@@ -28,7 +28,9 @@
                                      &key &allow-other-keys &aux))
                                   ((and val (symbolp val)) (list val))
                                   (t val))
-              sum (ash 1 (position symbol lambda-list-parser-states))))
+              for weight = (or (position symbol lambda-list-parser-states)
+                               (error "Not a parser state: ~S" symbol))
+              sum (ash 1 weight)))
       ;; Otherwise the input is required to be a list of symbols.
       (with-unique-names (k)
         `(loop for ,k in ,list
@@ -68,7 +70,7 @@
 ;;; wrong, we use COMPILER-ERROR, aborting compilation to the last
 ;;; recovery point.
 (declaim (ftype (sfunction
-                 (list &key (:context t) (:accept integer) (:silent boolean)
+                 (list &key (:context t) (:accept integer) (:silent t)
                             (:condition-class symbol))
                  (values (unsigned-byte 13) list list list list list list list))
                 parse-lambda-list))
@@ -680,7 +682,8 @@
            (max (+ min (length opt)))
            (bounds (list min max)))
       (cond ((ll-kwds-keyp llks)
-             `(,(if (eq (cddr macro-context) 'define-compiler-macro)
+             `(,(if (typep macro-context
+                           '(cons t (cons t (eql define-compiler-macro))))
                     'cmacro-check-ds-list/&key
                     'check-ds-list/&key)
                 ,input ,@bounds ,pattern
@@ -849,7 +852,7 @@
 (defun ds-bind-error (input min max pattern)
   (multiple-value-bind (name kind lambda-list) (get-ds-bind-context pattern)
     #-sb-xc-host
-    (declare (optimize sb!c::allow-non-returning-tail-call))
+    (declare (optimize allow-non-returning-tail-call))
     (case kind
      (:special-form
       ;; IR1 translators should call COMPILER-ERROR instead of
@@ -908,7 +911,11 @@
              (setq tail (cdr next))))))
     (multiple-value-bind (kind name) (get-ds-bind-context pattern)
       #-sb-xc-host
-      (declare (optimize sb!c::allow-non-returning-tail-call))
+      (declare (optimize allow-non-returning-tail-call))
+      ;; KLUDGE: Compiling (COERCE x 'list) transforms to COERCE-TO-LIST,
+      ;; but COERCE-TO-LIST is an inline function not yet defined, and
+      ;; its subsequent definition would signal an inlining failure warning.
+      (declare (notinline coerce))
       (error 'sb!kernel::defmacro-lambda-list-broken-key-list-error
              :kind kind :name name
              :problem problem
@@ -1032,16 +1039,6 @@
            ;; Typecheck the next cell so that calling code doesn't get an atom.
            (return (the cons (cdr plist)))))))
 
-;;; This is a variant of destructuring-bind that provides the name
-;;; of the containing construct in generated error messages.
-(def!macro named-ds-bind (context lambda-list data &body body &environment env)
-  (declare (ignorable env))
-  `(binding* ,(expand-ds-bind lambda-list data t nil context
-                              (and (eq (car context) :macro)
-                                   (eq (cddr context) 'deftype)
-                                   ''*))
-     ,@body))
-
 ;;; Make a lambda expression that receives an s-expression, destructures it
 ;;; according to LAMBDA-LIST, and executes BODY.
 ;;; NAME and KIND provide error-reporting context.
@@ -1090,7 +1087,24 @@
                       (when whole `((,(car whole) ,ll-whole)))))
              ;; Drop &WHOLE and &ENVIRONMENT
              (new-ll (make-lambda-list llks nil req opt rest keys aux))
-             (parse (parse-ds-lambda-list new-ll)))
+             (parse (parse-ds-lambda-list new-ll))
+             ((declared-lambda-list decls)
+              (let ((ll
+                      (loop for (nil . declarations) in decls
+                            thereis
+                            (loop for x in declarations
+                                  when (and (consp x)
+                                            (eql (car x) 'lambda-list))
+                                  return x))))
+                (values
+                 (or ll
+                     ;; Normalize the lambda list by unparsing.
+                     `(lambda-list ,(unparse-ds-lambda-list parse)))
+                 (if ll
+                     (loop for (declare . declarations) in decls
+                           collect (list* declare
+                                          (remove 'lambda-list declarations :key #'car)))
+                     decls)))))
     ;; Signal a style warning for duplicate names, but disregard &AUX variables
     ;; because most folks agree that (LET* ((X (F)) (X (G X))) ..) makes sense
     ;; - some would even say that it is idiomatic - and &AUX bindings are just
@@ -1103,28 +1117,30 @@
               (style-warn-once lambda-list "variable ~S occurs more than once"
                                (car tail))))
           (append whole env (ds-lambda-list-variables parse nil)))
+    ;; Maybe kill docstring, but only under the cross-compiler.
+    #!+(and (not sb-doc) (host-feature sb-xc-host)) (setq docstring nil)
     (values `(,@(if lambda-name `(named-lambda ,lambda-name) '(lambda))
                   (,ll-whole ,@ll-env ,@(and ll-aux (cons '&aux ll-aux)))
-               ,@(when (and docstring (eq doc-string-allowed :internal))
-                   (prog1 (list docstring) (setq docstring nil)))
-               ;; MACROLET doesn't produce an object capable of reflection,
-               ;; so don't bother inserting a different lambda-list.
-               ,@(unless (eq kind 'macrolet)
-                   ;; Normalize the lambda list by unparsing.
-                   `((declare (lambda-list ,(unparse-ds-lambda-list parse)))))
-               ,@(if outer-decls (list outer-decls))
-               ,@(and (not env) (eq envp t) `((declare (ignore ,@ll-env))))
-               ,@(sb!c:macro-policy-decls)
-               (,@(if kind
-                      `(named-ds-bind ,(if (eq kind :special-form)
-                                           `(:special-form . ,name)
-                                           `(:macro ,name . ,kind)))
-                      '(destructuring-bind))
-                   ,new-ll (,accessor ,ll-whole)
-                 ,@decls
-                 ,@(if wrap-block
-                       `((block ,(fun-name-block-name name) ,@forms))
-                       forms)))
+              ,@(when (and docstring (eq doc-string-allowed :internal))
+                  (prog1 (list docstring) (setq docstring nil)))
+              ;; MACROLET doesn't produce an object capable of reflection,
+              ;; so don't bother inserting a different lambda-list.
+              ,@(unless (eq kind 'macrolet)
+
+                  `((declare ,declared-lambda-list)))
+              ,@(if outer-decls (list outer-decls))
+              ,@(and (not env) (eq envp t) `((declare (ignore ,@ll-env))))
+              ,@(sb!c:macro-policy-decls)
+              (,@(if kind
+                     `(named-ds-bind ,(if (eq kind :special-form)
+                                          `(:special-form . ,name)
+                                          `(:macro ,name . ,kind)))
+                     '(destructuring-bind))
+                  ,new-ll (,accessor ,ll-whole)
+               ,@decls
+               ,@(if wrap-block
+                     `((block ,(fun-name-block-name name) ,@forms))
+                     forms)))
             docstring)))
 
 ;;; Functions should probably not retain &AUX variables as part
