@@ -268,14 +268,15 @@
     (once-only ((n-object object))
       (ecase (numeric-type-complexp type)
         (:real
-         (if (and #!-(or x86 x86-64 arm arm64) ;; Not implemented elsewhere yet
-                  nil
-                  (eql (numeric-type-class type) 'integer)
-                  (eql (numeric-type-low type) 0)
-                  (fixnump (numeric-type-high type)))
-             `(fixnum-mod-p ,n-object ,(numeric-type-high type))
-             `(and (typep ,n-object ',base)
-                   ,(transform-numeric-bound-test n-object type base))))
+         (cond #!+(or x86 x86-64 arm arm64) ;; Not implemented elsewhere yet
+               ((and
+                 (eql (numeric-type-class type) 'integer)
+                 (eql (numeric-type-low type) 0)
+                 (fixnump (numeric-type-high type)))
+                `(fixnum-mod-p ,n-object ,(numeric-type-high type)))
+               (t
+                `(and (typep ,n-object ',base)
+                      ,(transform-numeric-bound-test n-object type base)))))
         (:complex
          `(and (complexp ,n-object)
                ,(once-only ((n-real `(realpart (truly-the complex ,n-object)))
@@ -304,7 +305,12 @@
            (when (policy *lexenv* (> speed inhibit-warnings))
              (compiler-notify "can't open-code test of unknown type ~S"
                               (type-specifier type)))
-           `(%typep ,object ',spec))
+           `(let ((object ,object)
+                  (cache (load-time-value (cons #'sb!kernel::cached-typep ',spec)
+                                          t)))
+              (truly-the (values t &optional)
+                         (funcall (truly-the function (car (truly-the cons cache)))
+                                  cache object))))
           (t
            (ecase (first spec)
              (satisfies
@@ -536,35 +542,49 @@
   ;; By design the simple arrays of of rank 1 occupy a contiguous
   ;; range of widetags, and unlike the arbitrary-widetags code for unions,
   ;; this nonstandard predicate can be generically defined for all backends.
-  (if (and (not (array-type-complexp type))
-           (eq (array-type-element-type type) *wild-type*)
-           (equal (array-type-dimensions type) '(*)))
-      (return-from source-transform-array-typep
-        `(simple-rank-1-array-*-p ,obj)))
-  (multiple-value-bind (pred stype) (find-supertype-predicate type)
-    (if (and (array-type-p stype)
-             ;; (If the element type hasn't been defined yet, it's
-             ;; not safe to assume here that it will eventually
-             ;; have (UPGRADED-ARRAY-ELEMENT-TYPE type)=T, so punt.)
-             (not (unknown-type-p (array-type-element-type type)))
-             (or (eq (array-type-complexp stype) (array-type-complexp type))
-                 (and (eql (array-type-complexp stype) :maybe)
-                      (eql (array-type-complexp type) t))))
-        (once-only ((n-obj obj))
-          (multiple-value-bind (tests headerp)
-              (test-array-dimensions n-obj type stype)
-            `(and (,pred ,n-obj)
-                  ,@(when (and (eql (array-type-complexp stype) :maybe)
-                               (eql (array-type-complexp type) t))
-                      ;; KLUDGE: this is a bit lame; if we get here,
-                      ;; we already know that N-OBJ is an array, but
-                      ;; (NOT SIMPLE-ARRAY) doesn't know that.  On the
-                      ;; other hand, this should get compiled down to
-                      ;; two widetag tests, so it's only a bit lame.
-                      `((typep ,n-obj '(not simple-array))))
-                  ,@tests
-                  ,@(test-array-element-type n-obj type stype headerp))))
-        `(%typep ,obj ',(type-specifier type)))))
+  (let ((dims (array-type-dimensions type))
+        (et (array-type-element-type type)))
+   (if (and (not (array-type-complexp type))
+            (eq et *wild-type*)
+            (equal dims '(*)))
+       (return-from source-transform-array-typep
+         `(simple-rank-1-array-*-p ,obj)))
+    (multiple-value-bind (pred stype) (find-supertype-predicate type)
+      (if (and (array-type-p stype)
+               ;; (If the element type hasn't been defined yet, it's
+               ;; not safe to assume here that it will eventually
+               ;; have (UPGRADED-ARRAY-ELEMENT-TYPE type)=T, so punt.)
+               (not (unknown-type-p (array-type-element-type type)))
+               (or (eq (array-type-complexp stype) (array-type-complexp type))
+                   (and (eql (array-type-complexp stype) :maybe)
+                        (eql (array-type-complexp type) t))))
+          (let ((complex-tag (and
+                              (eql (array-type-complexp type) t)
+                              (singleton-p dims)
+                              (and (neq et *wild-type*)
+                                   (sb!vm:saetp-complex-typecode
+                                    (find-saetp-by-ctype (array-type-element-type type)))))))
+            (once-only ((n-obj obj))
+              (if complex-tag
+                  `(and (eq (%other-pointer-widetag ,n-obj) ,complex-tag)
+                        ,@(unless (eq (car dims) '*)
+                            `((= (%array-dimension ,n-obj 0) ,(car dims)))))
+                  (multiple-value-bind (tests headerp)
+                      (test-array-dimensions n-obj type stype)
+                    `(and ,@(unless (and headerp (eql pred 'arrayp))
+                              ;; ARRAY-HEADER-P from TESTS will test for that
+                              `((,pred ,n-obj)))
+                          ,@(when (and (eql (array-type-complexp stype) :maybe)
+                                       (eql (array-type-complexp type) t))
+                              ;; KLUDGE: this is a bit lame; if we get here,
+                              ;; we already know that N-OBJ is an array, but
+                              ;; (NOT SIMPLE-ARRAY) doesn't know that.  On the
+                              ;; other hand, this should get compiled down to
+                              ;; two widetag tests, so it's only a bit lame.
+                              `((typep ,n-obj '(not simple-array))))
+                          ,@tests
+                          ,@(test-array-element-type n-obj type stype headerp))))))
+          `(%typep ,obj ',(type-specifier type))))))
 
 ;;; Transform a type test against some instance type. The type test is
 ;;; flushed if the result is known at compile time. If not properly
@@ -708,10 +728,8 @@
 
            (t
             (/noshow "default case -- ,PRED and CLASS-CELL-TYPEP")
-            `(block typep
-               (classoid-cell-typep ,get-layout-or-return-false
-                                       ',(find-classoid-cell name :create t)
-                                       object)))))))))
+            `(classoid-cell-typep ',(find-classoid-cell name :create t)
+                                  object))))))))
 
 ;;; If the specifier argument is a quoted constant, then we consider
 ;;; converting into a simple predicate or other stuff. If the type is
@@ -806,28 +824,65 @@
         t))))
 
 ;;; Drops dimension information from vector types.
+;;; Returns four values
+;;; * vector ctype
+;;; * upgraded-element ctype or requsted element
+;;; * T if the upgraded-element is upgraded, i.e. it
+;;;   does not contain any unknown types.
+;;; * T if there were any dimensions
 (defun simplify-vector-type (type)
-  (aver (csubtypep type (specifier-type '(array * (*)))))
-  (let* ((array-type
-          (if (csubtypep type (specifier-type 'simple-array))
-              'simple-array
-              'array))
-         (complexp
-          (not
-           (or (eq 'simple-array array-type)
-               (neq *empty-type*
-                    (type-intersection type (specifier-type 'simple-array)))))))
-    (dolist (etype
-              #+sb-xc-host '(t bit character)
-              #-sb-xc-host sb!kernel::*specialized-array-element-types*
-              #+sb-xc-host (values nil nil nil)
-              #-sb-xc-host (values `(,array-type * (*)) t complexp))
-      (when etype
-        (let ((simplified (specifier-type `(,array-type ,etype (*)))))
-          (when (csubtypep type simplified)
-            (return (values (type-specifier simplified)
-                            etype
-                            complexp))))))))
+  (labels ((process-compound-type (types)
+             (let (array-types
+                   element-types
+                   (upgraded t)
+                   dimensions-removed)
+               (dolist (type types)
+                 (unless (or (hairy-type-p type)
+                             (sb!kernel::negation-type-p type))
+                   (multiple-value-bind (type et upgraded dimensions) (simplify type)
+                     (push type array-types)
+                     (push et element-types)
+                     (when dimensions
+                       (setf dimensions-removed t))
+                     (unless upgraded
+                       (setf upgraded nil)))))
+               (values (apply #'type-union array-types)
+                       (if (member *wild-type* element-types)
+                           *wild-type*
+                           (apply #'type-union element-types))
+                       upgraded
+                       dimensions-removed)))
+           (simplify (type)
+             (cond ((and (array-type-p type)
+                         (singleton-p (array-type-dimensions type)))
+                    (let* ((upgraded t)
+                           (et (array-type-specialized-element-type type))
+                           (et (cond ((neq et *wild-type*)
+                                      et)
+                                     ((eq (array-type-element-type type) *wild-type*)
+                                      et)
+                                     (t
+                                      (setf upgraded nil)
+                                      (array-type-element-type type)))))
+                      (values (specifier-type
+                               (list (if (array-type-complexp type)
+                                         'array
+                                         'simple-array)
+                                     (type-specifier et)
+                                     '(*)))
+                              et
+                              upgraded
+                              (not (eq (car (array-type-dimensions type)) '*)))))
+                   ((union-type-p type)
+                    (process-compound-type (union-type-types type)))
+                   ((intersection-type-p type)
+                    (process-compound-type (intersection-type-types type)))
+                   ((member-type-p type)
+                    (process-compound-type
+                     (mapcar #'ctype-of (member-type-members type))))
+                   (t
+                    (error "~a is not a subtype of VECTOR." type)))))
+    (simplify type)))
 
 (deftransform coerce ((x type) (* *) * :node node)
   (unless (constant-lvar-p type)
@@ -865,8 +920,8 @@
                    (the ,result-type
                         (complex (coerce (realpart x) ',part-type)
                                  (coerce (imagpart x) ',part-type))))))))
-           ;; Special case STRING and SIMPLE-STRING as they are union types
-           ;; in SBCL.
+          ;; Special case STRING and SIMPLE-STRING as they are union types
+          ;; in SBCL.
           ((member tval '(string simple-string))
            `(the ,tval
                  (if (typep x ',tval)
@@ -874,50 +929,51 @@
                      (replace (make-array (length x) :element-type 'character) x))))
           ((eq tval 'character)
            `(character x))
-           ;; Special case VECTOR
-           ((eq tval 'vector)
-            `(the ,tval
-               (if (vectorp x)
-                   x
-                   (replace (make-array (length x)) x))))
-           ;; Handle specialized element types for 1D arrays.
-           ((csubtypep tspec (specifier-type '(array * (*))))
-            ;; Can we avoid checking for dimension issues like (COERCE FOO
-            ;; '(SIMPLE-VECTOR 5)) returning a vector of length 6?
-            ;;
-            ;; CLHS actually allows this for all code with SAFETY < 3,
-            ;; but we're a conservative bunch.
-            (if (or (policy node (zerop safety)) ; no need in unsafe code
-                    (and (array-type-p tspec)    ; no need when no dimensions
-                         (equal (array-type-dimensions tspec) '(*))))
-                ;; We can!
-                (multiple-value-bind (vtype etype complexp) (simplify-vector-type tspec)
-                  (unless vtype
-                    (give-up-ir1-transform))
-                  `(the ,vtype
-                     (if (typep x ',vtype)
-                         x
-                         (replace
-                          (make-array (length x) :element-type ',etype
-                                      ,@(when complexp
-                                              (list :fill-pointer t
-                                                    :adjustable t)))
-                          x))))
-                ;; No, duh. Dimension checking required.
-                (give-up-ir1-transform
-                 "~@<~S specifies dimensions other than (*) in safe code.~:@>"
-                 tval)))
-           ((type= tspec (specifier-type 'list))
-            `(coerce-to-list x))
-           ((csubtypep tspec (specifier-type 'function))
-            (if (csubtypep (lvar-type x) (specifier-type 'symbol))
-                `(coerce-symbol-to-fun x)
-                ;; if X can later be derived as FUNCTION then we don't want
-                ;; to call COERCE-TO-FUN, because there's no smartness
-                ;; that can undo that and see that it's really (IDENTITY X).
-                (progn (delay-ir1-transform node :constraint)
-                       `(coerce-to-fun x))))
-           (t
-            (give-up-ir1-transform
-             "~@<open coding coercion to ~S not implemented.~:@>"
-             tval))))))
+          ;; Special case VECTOR
+          ((eq tval 'vector)
+           `(the ,tval
+                 (if (vectorp x)
+                     x
+                     (replace (make-array (length x)) x))))
+          ;; Handle specialized element types for 1D arrays.
+          ((csubtypep tspec (specifier-type '(array * (*))))
+           ;; Can we avoid checking for dimension issues like (COERCE FOO
+           ;; '(SIMPLE-VECTOR 5)) returning a vector of length 6?
+           ;;
+           ;; CLHS actually allows this for all code with SAFETY < 3,
+           ;; but we're a conservative bunch.
+           (if (or (policy node (zerop safety)) ; no need in unsafe code
+                   (and (array-type-p tspec) ; no need when no dimensions
+                        (equal (array-type-dimensions tspec) '(*))))
+               ;; We can!
+               (multiple-value-bind (vtype etype upgraded) (simplify-vector-type tspec)
+                 (unless upgraded
+                   (give-up-ir1-transform))
+                 (let ((vtype (type-specifier vtype)))
+                   `(the ,vtype
+                         (if (typep x ',vtype)
+                             x
+                             (replace
+                              (make-array (length x)
+                                          ,@(and (not (eq etype *universal-type*))
+                                                 (not (eq etype *wild-type*))
+                                                 `(:element-type ',(type-specifier etype))))
+                              x)))))
+               ;; No, duh. Dimension checking required.
+               (give-up-ir1-transform
+                "~@<~S specifies dimensions other than (*) in safe code.~:@>"
+                tval)))
+          ((type= tspec (specifier-type 'list))
+           `(coerce-to-list x))
+          ((csubtypep tspec (specifier-type 'function))
+           (if (csubtypep (lvar-type x) (specifier-type 'symbol))
+               `(coerce-symbol-to-fun x)
+               ;; if X can later be derived as FUNCTION then we don't want
+               ;; to call COERCE-TO-FUN, because there's no smartness
+               ;; that can undo that and see that it's really (IDENTITY X).
+               (progn (delay-ir1-transform node :constraint)
+                      `(coerce-to-fun x))))
+          (t
+           (give-up-ir1-transform
+            "~@<open coding coercion to ~S not implemented.~:@>"
+            tval))))))
