@@ -31,36 +31,14 @@
 
 (in-package "SB-PCL")
 
-(defun allocate-standard-instance (wrapper
-                                   &optional (slots-init nil slots-init-p))
-  (let ((instance (%make-standard-instance nil (get-instance-hash-code)))
-        (no-of-slots (wrapper-no-of-instance-slots wrapper)))
-    (setf (std-instance-wrapper instance) wrapper)
-    (setf (std-instance-slots instance)
-          (cond (slots-init-p
-                 ;; Inline the slots vector allocation and initialization.
-                 (let ((slots (make-array no-of-slots :initial-element 0)))
-                   (do ((rem-slots slots-init (rest rem-slots))
-                        (i 0 (1+ i)))
-                       ((>= i no-of-slots)) ;endp rem-slots))
-                     (declare (list rem-slots)
-                              (type index i))
-                     (setf (aref slots i) (first rem-slots)))
-                   slots))
-                (t
-                 (make-array no-of-slots
-                             :initial-element +slot-unbound+))))
+(defun allocate-standard-instance (wrapper)
+  (let ((instance (%make-standard-instance
+                   (make-array (layout-length wrapper)
+                               :initial-element +slot-unbound+)
+                   #-compact-instance-header 0)))
+    (setf (%instance-layout instance) wrapper)
     instance))
 
-(defmacro allocate-standard-funcallable-instance-slots
-    (wrapper &optional slots-init-p slots-init)
-  `(let ((no-of-slots (wrapper-no-of-instance-slots ,wrapper)))
-    ,(if slots-init-p
-         `(if ,slots-init-p
-           (make-array no-of-slots :initial-contents ,slots-init)
-           (make-array no-of-slots :initial-element +slot-unbound+))
-         `(make-array no-of-slots :initial-element +slot-unbound+))))
-
 (define-condition unset-funcallable-instance-function
     (reference-condition simple-error)
   ()
@@ -68,10 +46,24 @@
    :references (list '(:amop :generic-function allocate-instance)
                      '(:amop :function set-funcallable-instance-function))))
 
-(defun allocate-standard-funcallable-instance
-    (wrapper &optional (slots-init nil slots-init-p))
-  (let ((fin (%make-standard-funcallable-instance
-              nil (get-instance-hash-code))))
+(defun allocate-standard-funcallable-instance (wrapper)
+  (let* ((slots (make-array (layout-length wrapper) :initial-element +slot-unbound+))
+         (fin (cond #+(and compact-instance-header immobile-code)
+                    ((not (eql (layout-bitmap wrapper) -1))
+                     (let ((f (truly-the funcallable-instance
+                               (sb-sys:%primitive sb-vm::alloc-generic-function slots))))
+                       ;; Set layout prior to writing raw slots
+                       (setf (%funcallable-instance-layout f) wrapper)
+                       (sb-vm::%set-fin-trampoline f)))
+                    (t
+                     (let ((f (%make-standard-funcallable-instance
+                               slots
+                               #-compact-instance-header (sb-impl::new-instance-hash-code))))
+                       (setf (%funcallable-instance-layout f) wrapper)
+                       f)))))
+    #+compact-instance-header
+    (set-header-data slots
+                     (ash (logand (sb-impl::new-instance-hash-code) #xFFFFFFFF) 24))
     (set-funcallable-instance-function
      fin
      #'(lambda (&rest args)
@@ -80,10 +72,6 @@
                 :format-control "~@<The function of funcallable instance ~
                                  ~S has not been set.~@:>"
                 :format-arguments (list fin))))
-    (setf (fsc-instance-wrapper fin) wrapper
-          (fsc-instance-slots fin)
-          (allocate-standard-funcallable-instance-slots
-           wrapper slots-init-p slots-init))
     fin))
 
 (defun classify-slotds (slotds)
@@ -119,7 +107,7 @@
                    `(setf ,wr ,(if (eq class 'standard-generic-function)
                                    '*sgf-wrapper*
                                    `(!boot-make-wrapper
-                                     (early-class-size ',class)
+                                     (!early-class-size ',class)
                                      ',class))
                           ,class (allocate-standard-instance
                                   ,(if (eq class 'standard-generic-function)
@@ -154,7 +142,7 @@
     ;; First, make a class metaobject for each of the early classes. For
     ;; each metaobject we also set its wrapper. Except for the class T,
     ;; the wrapper is always that of STANDARD-CLASS.
-    (dolist (definition *early-class-definitions*)
+    (dolist (definition *!early-class-definitions*)
       (let* ((name (ecd-class-name definition))
              (meta (ecd-metaclass definition))
              (wrapper (ecase meta
@@ -169,7 +157,7 @@
              (class (or (find-class name nil)
                         (allocate-standard-instance wrapper))))
         (setf (find-class name) class)))
-    (dolist (definition *early-class-definitions*)
+    (dolist (definition *!early-class-definitions*)
       (let ((name (ecd-class-name definition))
             (meta (ecd-metaclass definition))
             (source (ecd-source-location definition))
@@ -179,7 +167,7 @@
         (let ((direct-default-initargs
                (getf other-initargs :direct-default-initargs)))
           (multiple-value-bind (slots cpl default-initargs direct-subclasses)
-              (early-collect-inheritance name)
+              (!early-collect-inheritance name)
             (let* ((class (find-class name))
                    (wrapper (cond ((eq class slot-class)
                                    slot-class-wrapper)
@@ -206,7 +194,6 @@
                                   (t
                                    (!boot-make-wrapper (length slots) name))))
                    (proto nil))
-              (when (eq name t) (setq *the-wrapper-of-t* wrapper))
               (set (make-class-symbol name) class)
               (dolist (slot slots)
                 (unless (eq (getf slot :allocation :instance) :instance)
@@ -271,21 +258,32 @@
                     (symbol-value (make-class-symbol name)))
                   *standard-method-class-names*))
 
-    (let* ((smc-class (find-class 'standard-method-combination))
-           (smc-wrapper (!bootstrap-get-slot 'standard-class
-                                             smc-class
-                                             'wrapper))
-           (smc (allocate-standard-instance smc-wrapper)))
-      (flet ((set-slot (name value)
-               (!bootstrap-set-slot 'standard-method-combination
-                                    smc
-                                    name
-                                    value)))
-        (set-slot 'source nil)
-        (set-slot 'type-name 'standard)
-        (set-slot '%documentation "The standard method combination.")
-        (set-slot 'options ()))
-      (setq *standard-method-combination* smc))))
+    (flet ((make-method-combination (class-name)
+             (let* ((class (find-class class-name))
+                    (wrapper (!bootstrap-get-slot
+                              'standard-class class 'wrapper))
+                    (instance (allocate-standard-instance wrapper)))
+               (flet ((set-slot (name value)
+                        (!bootstrap-set-slot class-name instance name value)))
+                 (values instance #'set-slot)))))
+      ;; Create the STANDARD method combination object.
+      (multiple-value-bind (method-combination set-slot)
+          (make-method-combination 'standard-method-combination)
+        (funcall set-slot 'source nil)
+        (funcall set-slot 'type-name 'standard)
+        (funcall set-slot 'options '())
+        (funcall set-slot '%documentation "The standard method combination.")
+        (setq *standard-method-combination* method-combination))
+      ;; Create the OR method combination object.
+      (multiple-value-bind (method-combination set-slot)
+          (make-method-combination 'short-method-combination)
+        (funcall set-slot 'source 'nil)
+        (funcall set-slot 'type-name 'or)
+        (funcall set-slot 'operator 'or)
+        (funcall set-slot 'identity-with-one-argument t)
+        (funcall set-slot '%documentation nil)
+        (funcall set-slot 'options '(:most-specific-first))
+        (setq *or-method-combination* method-combination)))))
 
 ;;; Initialize a class metaobject.
 (defun !bootstrap-initialize-class
@@ -425,7 +423,7 @@
 
 (defun !bootstrap-accessor-definitions (early-p)
   (let ((*early-p* early-p))
-    (dolist (definition *early-class-definitions*)
+    (dolist (definition *!early-class-definitions*)
       (let ((name (ecd-class-name definition))
             (meta (ecd-metaclass definition)))
         (unless (or (eq meta 'built-in-class) (eq meta 'system-class))
@@ -540,7 +538,6 @@
                (lclass (find-classoid name))
                (wrapper (classoid-layout lclass)))
           (set (get-built-in-class-symbol name) class)
-          (set (get-built-in-wrapper-symbol name) wrapper)
           (setf (classoid-pcl-class lclass) class)
 
           (!bootstrap-initialize-class 'built-in-class class
@@ -550,6 +547,7 @@
                                        wrapper prototype))))))
 
 (defun class-of (x)
+  (declare (explicit-check))
   (wrapper-class* (layout-of x)))
 
 (defun eval-form (form)
@@ -721,19 +719,25 @@
 (defmethod no-primary-method (generic-function &rest args)
   (error 'no-primary-method :generic-function generic-function :args args))
 
+;; FIXME shouldn't this specialize on STANDARD-METHOD-COMBINATION?
 (defmethod invalid-qualifiers ((gf generic-function)
                                combin
                                method)
-  (let ((qualifiers (method-qualifiers method)))
-    (let ((why (cond
-                 ((cdr qualifiers) "has too many qualifiers")
-                 (t (aver (not (member (car qualifiers)
-                                       '(:around :before :after))))
-                    "has an invalid qualifier"))))
-      (invalid-method-error
-       method
-       "The method ~S on ~S ~A.~%~
-        Standard method combination requires all methods to have one~%~
-        of the single qualifiers :AROUND, :BEFORE and :AFTER or to~%~
-        have no qualifier at all."
-       method gf why))))
+  (let* ((qualifiers (method-qualifiers method))
+         (qualifier (first qualifiers))
+         (type-name (method-combination-type-name combin))
+         (why (cond
+                ((cdr qualifiers)
+                 "has too many qualifiers")
+                (t
+                 (aver (not (standard-method-combination-qualifier-p
+                             qualifier)))
+                 "has an invalid qualifier"))))
+    (invalid-method-error
+     method
+     "~@<The method ~S on ~S ~A.~
+      ~@:_~@:_~
+      ~@(~A~) method combination requires all methods to have one of ~
+      the single qualifiers ~{~S~#[~; and ~:;, ~]~} or to have no ~
+      qualifier at all.~@:>"
+     method gf why type-name +standard-method-combination-qualifiers+)))
