@@ -53,7 +53,8 @@
                  (declare (ignore initargs))
                  (error 'metaobject-initialization-violation
                   ;; FIXME: I'm pretty sure this wants to be "~~@<~A~~@:>"
-                  :format-control ,(format nil "~@<~A~@:>" control)
+                  :format-control ,(coerce (format nil "~@<~A~@:>" control)
+                                           'base-string)
                   :format-arguments (list ',name)
                   :references (list '(:amop :initialization method))))))
   (def reinitialize-instance ((method method) &rest initargs)
@@ -252,35 +253,34 @@
 (defun real-get-method (generic-function qualifiers specializers
                         &optional (errorp t)
                         always-check-specializers)
-  (let ((lspec (length specializers))
+  (let ((specializer-count (length specializers))
         (methods (generic-function-methods generic-function)))
     (when (or methods always-check-specializers)
-      (let ((nreq (length (arg-info-metatypes (gf-arg-info
-                                               generic-function)))))
+      (let ((required-parameter-count
+             (length (arg-info-metatypes (gf-arg-info generic-function)))))
         ;; Since we internally bypass FIND-METHOD by using GET-METHOD
-        ;; instead we need to to this here or users may get hit by a
+        ;; instead we need to do this here or users may get hit by a
         ;; failed AVER instead of a sensible error message.
-        (when (/= lspec nreq)
+        (unless (= specializer-count required-parameter-count)
           (error
            'find-method-length-mismatch
-           :format-control
-           "~@<The generic function ~S takes ~D required argument~:P; ~
-            was asked to find a method with specializers ~S~@:>"
-           :format-arguments (list generic-function nreq specializers)))))
-    (let ((hit
-           (dolist (method methods)
-             (let ((mspecializers (method-specializers method)))
-               (aver (= lspec (length mspecializers)))
-               (when (and (equal qualifiers (safe-method-qualifiers method))
-                          (every #'same-specializer-p specializers
-                                 (method-specializers method)))
-                 (return method))))))
-      (cond (hit hit)
+           :format-control   "~@<The generic function ~S takes ~D ~
+                              required argument~:P; was asked to ~
+                              find a method with specializers ~:S~@:>"
+           :format-arguments (list generic-function required-parameter-count
+                                   (unparse-specializers generic-function specializers))))))
+    (flet ((congruentp (other-method)
+             (let ((other-specializers (method-specializers other-method)))
+               (aver (= specializer-count (length other-specializers)))
+               (and (equal qualifiers (safe-method-qualifiers other-method))
+                    (every #'same-specializer-p specializers other-specializers)))))
+      (declare (dynamic-extent #'congruentp))
+      (cond ((find-if #'congruentp methods))
             ((null errorp) nil)
             (t
-             (error "~@<There is no method on ~S with ~
-                    ~:[no qualifiers~;~:*qualifiers ~S~] ~
-                    and specializers ~S.~@:>"
+             (error "~@<There is no method on ~S with ~:[no ~
+                     qualifiers~;~:*qualifiers ~:S~] and specializers ~
+                     ~:S.~@:>"
                     generic-function qualifiers specializers))))))
 
 (defmethod find-method ((generic-function standard-generic-function)
@@ -396,13 +396,18 @@
 (defmethod initialize-instance :after ((gf standard-generic-function)
                                        &key (lambda-list nil lambda-list-p)
                                        argument-precedence-order)
-  (with-slots (arg-info) gf
+  ;; FIXME: Because ARG-INFO is a STRUCTURE-OBJECT, it does not get
+  ;; a permutation vector, and therefore the code that SLOT-VALUE transforms
+  ;; to winds up punting to #'(SLOT-ACCESSOR :GLOBAL ARG-INFO READER).
+  ;; Using SLOT-VALUE the "slow" way sidesteps some bootstrap issues.
+  (declare (notinline slot-value))
+  (progn ; WAS: with-slots (arg-info) gf
     (if lambda-list-p
         (set-arg-info gf
                       :lambda-list lambda-list
                       :argument-precedence-order argument-precedence-order)
         (set-arg-info gf))
-    (when (arg-info-valid-p arg-info)
+    (when (arg-info-valid-p (slot-value gf 'arg-info))
       (update-dfun gf))))
 
 (defmethod reinitialize-instance :around
@@ -474,16 +479,51 @@
                     situations.~@:>"
    :format-arguments (list 'print-object)))
 
+(defun defer-ftype-computation (gf)
+  ;; Is there any reason not to do this as soon as possible?
+  ;; While doing it with every ADD/REMOVE-METHOD call could result in
+  ;; wasted work, it seems like unnecessary complexity.
+  ;; I think it's just to get through bootstrap, probably,
+  ;; but if it's a semantics thing, it deserves some explanation.
+  (let ((name (generic-function-name gf)))
+    (when (legal-fun-name-p name) ; tautological ?
+      (unless (eq (info :function :where-from name) :declared)
+        (when (and (fboundp name) (eq (fdefinition name) gf))
+          (setf (info :function :type name) :generic-function))))))
+
+(defun compute-gf-ftype (name)
+  (let ((gf (and (fboundp name) (fdefinition name))))
+    (if (generic-function-p gf)
+        (let* ((ll (generic-function-lambda-list gf))
+               ;; If the GF has &REST without &KEY then we don't augment
+               ;; the FTYPE with keywords, so as not to complain about keywords
+               ;; which seem not to be accepted.
+               (type (sb-c::ftype-from-lambda-list
+                      (if (and (member '&rest ll) (not (member '&key ll)))
+                          ll
+                          (generic-function-pretty-arglist gf)))))
+          ;; It would be nice if globaldb were transactional,
+          ;; so that either both updates or neither occur.
+          (setf (info :function :type name) type
+                (info :function :where-from name) :defined-method)
+          type)
+        ;; The defaulting expression for (:FUNCTION :TYPE) does not store
+        ;; the default. For :GENERIC-FUNCTION that is not FBOUNDP we also
+        ;; don't, however this branch should never be reached because the
+        ;; info only stores :GENERIC-FUNCTION when methods are loaded.
+        ;; Maybe AVER that it does not happen?
+        (sb-c::ftype-from-fdefn name))))
+
 (defun real-add-method (generic-function method &optional skip-dfun-update-p)
   (flet ((similar-lambda-lists-p (old-method new-lambda-list)
            (binding* (((a-llks a-nreq a-nopt)
                        (analyze-lambda-list (method-lambda-list old-method)))
                       ((b-llks b-nreq b-nopt)
                        (analyze-lambda-list new-lambda-list)))
-               (and (= a-nreq b-nreq)
-                    (= a-nopt b-nopt)
-                    (eq (ll-keyp-or-restp a-llks)
-                        (ll-keyp-or-restp b-llks))))))
+             (and (= a-nreq b-nreq)
+                  (= a-nopt b-nopt)
+                  (eq (ll-keyp-or-restp a-llks)
+                      (ll-keyp-or-restp b-llks))))))
     (multiple-value-bind (lock qualifiers specializers new-lambda-list
                           method-gf name)
         (values-for-add-method generic-function method)
@@ -546,31 +586,31 @@
               ;; invocation time; I dunno what the rationale was, and it
               ;; sucks.  Nevertheless, it's probably a programmer error, so
               ;; let's warn anyway. -- CSR, 2003-08-20
-              (let ((mc (generic-function-method-combination generic-functioN)))
-                (cond
-                  ((eq mc *standard-method-combination*)
-                   (when (and qualifiers
-                              (or (cdr qualifiers)
-                                  (not (memq (car qualifiers)
-                                             '(:around :before :after)))))
-                     (warn "~@<Invalid qualifiers for standard method ~
-                            combination in method ~S:~2I~_~S.~@:>"
-                           method qualifiers)))
-                  ((short-method-combination-p mc)
-                   (let ((mc-name (method-combination-type-name mc)))
-                     (when (or (null qualifiers)
-                               (cdr qualifiers)
-                               (and (neq (car qualifiers) :around)
-                                    (neq (car qualifiers) mc-name)))
-                       (warn "~@<Invalid qualifiers for ~S method combination ~
-                              in method ~S:~2I~_~S.~@:>"
-                             mc-name method qualifiers))))))
+              (let* ((mc (generic-function-method-combination generic-function))
+                     (type-name (method-combination-type-name mc)))
+                (flet ((invalid ()
+                         (warn "~@<Invalid qualifiers for ~S method ~
+                                combination in method ~S:~2I~_~S.~@:>"
+                               type-name method qualifiers)))
+                  (cond
+                    ((and (eq mc *standard-method-combination*)
+                          qualifiers
+                          (or (cdr qualifiers)
+                              (not (standard-method-combination-qualifier-p
+                                    (car qualifiers)))))
+                     (invalid))
+                    ((and (short-method-combination-p mc)
+                          (or (null qualifiers)
+                              (cdr qualifiers)
+                              (not (short-method-combination-qualifier-p
+                                    type-name (car qualifiers)))))
+                     (invalid)))))
               (unless skip-dfun-update-p
                 (update-ctors 'add-method
                               :generic-function generic-function
                               :method method)
                 (update-dfun generic-function))
-              (setf (gf-info-needs-update generic-function) t)
+              (defer-ftype-computation generic-function)
               (map-dependents generic-function
                               (lambda (dep)
                                 (update-dependent generic-function
@@ -587,80 +627,24 @@
       ;; it would be bad to unwind and leave the gf in an inconsistent
       ;; state.
       (sb-thread::with-recursive-system-lock (lock)
-        (let* ((specializers (method-specializers method)) ; flushable?
+        (let* ((specializers (method-specializers method))
                (methods (generic-function-methods generic-function))
                (new-methods (remove method methods)))
-          (declare (ignore specializers))
           (setf (method-generic-function method) nil
                 (generic-function-methods generic-function) new-methods)
-          (dolist (specializer (method-specializers method))
+          (dolist (specializer specializers)
             (remove-direct-method specializer method))
           (set-arg-info generic-function)
           (update-ctors 'remove-method
                         :generic-function generic-function
                         :method method)
           (update-dfun generic-function)
-          (setf (gf-info-needs-update generic-function) t)
+          (defer-ftype-computation generic-function)
           (map-dependents generic-function
                           (lambda (dep)
                             (update-dependent generic-function
                                               dep 'remove-method method)))))))
   generic-function)
-
-
-;; Tell INFO about the generic function's methods' keys so that the
-;; compiler doesn't complain that the keys defined for some method are
-;; unrecognized.
-(fmakunbound 'sb-c::maybe-update-info-for-gf)
-(defun sb-c::maybe-update-info-for-gf (name)
-    (let ((gf (if (fboundp name) (fdefinition name))))
-      (when (and gf (generic-function-p gf) (not (early-gf-p gf))
-                 (not (eq :declared (info :function :where-from name)))
-                 (gf-info-needs-update gf))
-        (let* ((methods (generic-function-methods gf))
-               (gf-lambda-list (generic-function-lambda-list gf))
-               (tfun (constantly t))
-               keysp)
-          (multiple-value-bind (llks gf.required gf.optional gf.rest gf.keys)
-              (parse-lambda-list gf-lambda-list)
-            ;; 7.6.4 point 5 probably entails that if any method says
-            ;; &allow-other-keys then the gf should be construed to
-            ;; accept any key.
-            (let* ((allowp (or (ll-kwds-allowp llks)
-                               (find '&allow-other-keys methods
-                                     :test #'find
-                                     :key #'method-lambda-list)))
-                   (ftype
-                    (specifier-type
-                     ;; SERIOUSLY? Do we not already have like at least
-                     ;; two other variations on this code?
-                     `(function
-                       (,@(mapcar tfun gf.required)
-                          ,@(if gf.optional
-                                `(&optional ,@(mapcar tfun gf.optional)))
-                          ,@(if gf.rest
-                                `(&rest t))
-                          ,@(when (ll-kwds-keyp llks)
-                              (let ((all-keys
-                                     (mapcar
-                                      (lambda (x)
-                                        (list x t))
-                                      (remove-duplicates
-                                       (nconc
-                                        (mapcan #'function-keywords methods)
-                                        (mapcar #'parse-key-arg-spec gf.keys))))))
-                                (when all-keys
-                                  (setq keysp t)
-                                  `(&key ,@all-keys))))
-                          ,@(when (and (not keysp) allowp)
-                              `(&key))
-                          ,@(when allowp
-                              `(&allow-other-keys)))
-                       *))))
-              (setf (info :function :type name) ftype
-                    (info :function :where-from name) :defined-method
-                    (gf-info-needs-update gf) nil)
-              ftype))))))
 
 (defun compute-applicable-methods-function (generic-function arguments)
   (values (compute-applicable-methods-using-types
@@ -1581,8 +1565,9 @@
 ;;; function of a standard-generic-function
 (let (initial-print-object-cache)
   (defun standard-compute-discriminating-function (gf)
+    (declare (notinline slot-value))
     (let ((dfun-state (slot-value gf 'dfun-state)))
-          (when (special-case-for-compute-discriminating-function-p gf)
+      (when (special-case-for-compute-discriminating-function-p gf)
             ;; if we have a special case for
             ;; COMPUTE-DISCRIMINATING-FUNCTION, then (at least for the
             ;; special cases implemented as of 2006-05-09) any information
@@ -1590,16 +1575,16 @@
             (aver (null dfun-state)))
           (typecase dfun-state
             (null
-             (when (eq gf #'compute-applicable-methods)
+             (when (eq gf (load-time-value #'compute-applicable-methods t))
                (update-all-c-a-m-gf-info gf))
              (cond
-               ((eq gf #'slot-value-using-class)
+               ((eq gf (load-time-value #'slot-value-using-class t))
                 (update-slot-value-gf-info gf 'reader)
                 #'slot-value-using-class-dfun)
-               ((eq gf #'(setf slot-value-using-class))
+               ((eq gf (load-time-value #'(setf slot-value-using-class) t))
                 (update-slot-value-gf-info gf 'writer)
                 #'setf-slot-value-using-class-dfun)
-               ((eq gf #'slot-boundp-using-class)
+               ((eq gf (load-time-value #'slot-boundp-using-class t))
                 (update-slot-value-gf-info gf 'boundp)
                 #'slot-boundp-using-class-dfun)
                ;; KLUDGE: PRINT-OBJECT is not a special-case in the sense
@@ -1607,12 +1592,8 @@
                ;; However, it is important that the machinery for printing
                ;; conditions for stack and heap exhaustion, and the
                ;; restarts offered by the debugger, work without consuming
-               ;; many extra resources.  This way (testing by name of GF
-               ;; rather than by identity) was the only way I found to get
-               ;; this to bootstrap, given that the PRINT-OBJECT generic
-               ;; function is only set up later, in
-               ;; SRC;PCL;PRINT-OBJECT.LISP.  -- CSR, 2008-06-09
-               ((eq (slot-value gf 'name) 'print-object)
+               ;; many extra resources.  -- CSR, 2008-06-09
+               ((eq gf (locally (declare (optimize (safety 0))) #'print-object))
                 (let ((nkeys (nth-value 3 (get-generic-fun-info gf))))
                   (cond ((/= nkeys 1)
                          ;; KLUDGE: someone has defined a method
@@ -1624,11 +1605,11 @@
                              (make-caching-dfun gf (copy-cache initial-print-object-cache))
                            (set-dfun gf dfun cache info)))
                         ;; the relevant PRINT-OBJECT methods get defined
-                        ;; late, by delayed DEF!METHOD.  We mustn't cache
+                        ;; late, by delayed DEFMETHOD.  We mustn't cache
                         ;; the effective method for our classes earlier
                         ;; than the relevant PRINT-OBJECT methods are
                         ;; defined...
-                        ((boundp 'sb-impl::*delayed-def!method-args*)
+                        ((boundp '*!delayed-defmethod-args*)
                          (make-initial-dfun gf))
                         (t (multiple-value-bind (dfun cache info)
                                (make-final-dfun-internal
@@ -1697,67 +1678,37 @@
                                (method-lambda-list method)))
     (declare (ignore nreq nopt))
     (values keywords (ll-kwds-allowp llks))))
-
-(defmethod function-keyword-parameters ((method standard-method))
-  (multiple-value-bind (llks nreq nopt keywords keyword-parameters)
-      (analyze-lambda-list (if (consp method)
-                               (early-method-lambda-list method)
-                               (method-lambda-list method)))
-    (declare (ignore nreq nopt keywords))
-    (values keyword-parameters (ll-kwds-allowp llks))))
-
-;; FIXME: this is just: "parse, set keys to nil, unparse" (I think).
-(defun method-ll->generic-function-ll (ll)
-  (multiple-value-bind (llks nreq nopt keywords keyword-parameters)
-      (analyze-lambda-list ll)
-    (declare (ignore llks nreq nopt keywords))
-    (remove-if (lambda (s)
-                 (or (memq s keyword-parameters)
-                     (eq s '&allow-other-keys)))
-               ll)))
 
 ;;; This is based on the rules of method lambda list congruency
 ;;; defined in the spec. The lambda list it constructs is the pretty
 ;;; union of the lambda lists of the generic function and of all its
-;;; methods.  It doesn't take method applicability into account at all
-;;; yet.
-
-;;; (Notice that we ignore &AUX variables as they're not part of the
-;;; "public interface" of a function.)
-
-(defmethod generic-function-pretty-arglist
-           ((generic-function standard-generic-function))
-  (let ((gf-lambda-list (generic-function-lambda-list generic-function))
-        (methods (generic-function-methods generic-function)))
-    (if (null methods)
-        gf-lambda-list
-        (multiple-value-bind (gf.llks gf.required gf.optional gf.rest gf.keys)
-            (parse-lambda-list gf-lambda-list :silent t)
-          ;; Possibly extend the keyword parameters of the gf by
-          ;; additional key parameters of its methods:
-          (let ((methods.keys nil) (methods.allowp nil))
-            (dolist (m methods)
-              (multiple-value-bind (m.keyparams m.allow-other-keys)
-                  (function-keyword-parameters m)
-                (setq methods.keys (union methods.keys m.keyparams :key #'maybe-car))
-                (setq methods.allowp (or methods.allowp m.allow-other-keys))))
-            (let ((arglist '()))
-              (when (or (ll-kwds-allowp gf.llks) methods.allowp)
-                (push '&allow-other-keys arglist))
-              (when (or gf.keys methods.keys)
-                ;; We make sure that the keys of the gf appear before
-                ;; those of its methods, since they're probably more
-                ;; generally appliable.
-                (setq arglist (nconc (list '&key) gf.keys
-                                     (nset-difference methods.keys gf.keys)
-                                     arglist)))
-              (when gf.rest
-                (setq arglist (nconc (cons '&rest gf.rest) arglist)))
-              (when gf.optional
-                (setq arglist (nconc (list '&optional) gf.optional arglist)))
-              (nconc gf.required arglist)))))))
-
-(defun maybe-car (thing)
-  (if (listp thing)
-      (car thing)
-      thing))
+;;; methods.  It doesn't take method applicability into account; we
+;;; also ignore non-public parts of the interface (e.g. &AUX, default
+;;; and supplied-p parameters)
+;;; The compiler uses this for type-checking that callers pass acceptable
+;;; keywords, so don't make this do anything fancy like looking at effective
+;;; methods without also fixing the compiler.
+(defmethod generic-function-pretty-arglist ((gf standard-generic-function))
+  (let ((gf-lambda-list (generic-function-lambda-list gf))
+        (methods (generic-function-methods gf)))
+    (flet ((canonize (k)
+             (multiple-value-bind (kw var)
+                 (parse-key-arg-spec k)
+               (if (and (eql (symbol-package kw) *keyword-package*)
+                        (string= kw var))
+                   var
+                   (list (list kw var))))))
+      (multiple-value-bind (llks required optional rest keys)
+          (parse-lambda-list gf-lambda-list :silent t)
+        (collect ((keys (mapcar #'canonize keys)))
+        ;; Possibly extend the keyword parameters of the gf by
+        ;; additional key parameters of its methods:
+          (dolist (m methods
+                     (make-lambda-list llks nil required optional rest (keys)))
+            (binding* (((m.llks nil nil nil m.keys)
+                        (parse-lambda-list (method-lambda-list m) :silent t)))
+              (setq llks (logior llks m.llks))
+              (dolist (k m.keys)
+                (unless (member (parse-key-arg-spec k) (keys)
+                                :key #'parse-key-arg-spec :test #'eq)
+                  (keys (canonize k)))))))))))
