@@ -32,6 +32,7 @@
 #include "genesis/static-symbols.h"
 #include "genesis/primitive-objects.h"
 #include "thread.h"
+#include "gc-internal.h"
 
 #ifdef LISP_FEATURE_OS_PROVIDES_DLADDR
 # include <dlfcn.h>
@@ -54,43 +55,52 @@ sbcl_putwc(wchar_t c, FILE *file)
 struct compiled_debug_fun *
 debug_function_from_pc (struct code* code, void *pc)
 {
-  uword_t code_header_len = sizeof(lispobj) * HeaderValue(code->header);
-  uword_t offset
-    = (uword_t) pc - (uword_t) code - code_header_len;
-  struct compiled_debug_fun *df;
-  struct compiled_debug_info *di;
-  struct vector *v;
-  int i, len;
+    uword_t code_header_len = sizeof(lispobj) * code_header_words(code->header);
+    uword_t offset
+        = (uword_t) pc - (uword_t) code - code_header_len;
+    struct compiled_debug_fun *df;
+    struct compiled_debug_info *di;
+    struct vector *v;
+    int i, len;
 
-  if (lowtag_of(code->debug_info) != INSTANCE_POINTER_LOWTAG)
-    return 0;
+    if (lowtag_of(code->debug_info) != INSTANCE_POINTER_LOWTAG)
+        return NULL;
 
-  di = (struct compiled_debug_info *) native_pointer(code->debug_info);
-  v = (struct vector *) native_pointer(di->fun_map);
-  len = fixnum_value(v->length);
-  df = (struct compiled_debug_fun *) native_pointer(v->data[0]);
+    di = (struct compiled_debug_info *) native_pointer(code->debug_info);
 
-  if (len == 1)
-    return df;
+    if (lowtag_of(di->fun_map) != INSTANCE_POINTER_LOWTAG)
+        return NULL;
 
-  for (i = 1;; i += 2) {
-    unsigned next_pc;
+    v = (struct vector *) native_pointer(di->fun_map);
 
-    if (i == len)
-      return ((struct compiled_debug_fun *) native_pointer(v->data[i - 1]));
+    len = fixnum_value(v->length);
 
-    if (offset >= (uword_t)fixnum_value(df->elsewhere_pc)) {
-      struct compiled_debug_fun *p
-        = ((struct compiled_debug_fun *) native_pointer(v->data[i + 1]));
-      next_pc = fixnum_value(p->elsewhere_pc);
-    } else
-      next_pc = fixnum_value(v->data[i]);
+    if (lowtag_of(v->data[0]) != INSTANCE_POINTER_LOWTAG)
+        return NULL;
 
-    if (offset < next_pc)
-      return ((struct compiled_debug_fun *) native_pointer(v->data[i - 1]));
-  }
+    df = (struct compiled_debug_fun *) native_pointer(v->data[0]);
 
-  return NULL;
+    if (len == 1)
+        return df;
+
+    for (i = 1;; i += 2) {
+        unsigned next_pc;
+
+        if (i == len)
+            return ((struct compiled_debug_fun *) native_pointer(v->data[i - 1]));
+
+        if (offset >= (uword_t)fixnum_value(df->elsewhere_pc)) {
+            struct compiled_debug_fun *p
+                = ((struct compiled_debug_fun *) native_pointer(v->data[i + 1]));
+            next_pc = fixnum_value(p->elsewhere_pc);
+        } else
+            next_pc = fixnum_value(v->data[i]);
+
+        if (offset < next_pc)
+            return ((struct compiled_debug_fun *) native_pointer(v->data[i - 1]));
+    }
+
+    return NULL;
 }
 
 static void
@@ -140,66 +150,70 @@ static int string_equal (lispobj *object, char *string)
 static void
 print_entry_name (lispobj name)
 {
-  if (lowtag_of (name) == LIST_POINTER_LOWTAG) {
-    putchar('(');
-    while (name != NIL) {
-      struct cons *cons = (struct cons *) native_pointer(name);
-      print_entry_name(cons->car);
-      name = cons->cdr;
-      if (name != NIL)
-        putchar(' ');
-    }
-    putchar(')');
-  } else if (lowtag_of(name) == OTHER_POINTER_LOWTAG) {
-    lispobj *object = (lispobj *) native_pointer(name);
-    if (widetag_of(*object) == SYMBOL_HEADER_WIDETAG) {
-      struct symbol *symbol = (struct symbol *) object;
-      if (symbol->package != NIL) {
-        struct package *pkg
-          = (struct package *) native_pointer(symbol->package);
-        lispobj pkg_name = pkg->_name;
-        if (string_equal(native_pointer(pkg_name), "COMMON-LISP"))
-            ;
-        else if (string_equal(native_pointer(pkg_name), "COMMON-LISP-USER")) {
-            fputs("CL-USER::", stdout);
+    if (lowtag_of (name) == LIST_POINTER_LOWTAG) {
+        putchar('(');
+        while (name != NIL) {
+            if (lowtag_of (name) != LIST_POINTER_LOWTAG) {
+                printf("%p: unexpected lowtag while printing a cons\n",
+                       (void*)name);
+                return;
+            }
+            struct cons *cons = (struct cons *) native_pointer(name);
+            print_entry_name(cons->car);
+            name = cons->cdr;
+            if (name != NIL)
+                putchar(' ');
         }
-        else if (string_equal(native_pointer(pkg_name), "KEYWORD")) {
-            putchar(':');
-        } else {
-            print_string(native_pointer(pkg_name));
-            fputs("::", stdout);
-        }
-      }
-      print_string(native_pointer(symbol->name));
-    } else if (widetag_of(*object) == SIMPLE_BASE_STRING_WIDETAG
+        putchar(')');
+    } else if (lowtag_of(name) == OTHER_POINTER_LOWTAG) {
+        lispobj *object = native_pointer(name);
+        if (widetag_of(*object) == SYMBOL_WIDETAG) {
+            struct symbol *symbol = (struct symbol *) object;
+            if (symbol->package != NIL) {
+                struct package *pkg
+                    = (struct package *) native_pointer(symbol->package);
+                lispobj pkg_name = pkg->_name;
+                if (string_equal(native_pointer(pkg_name), "COMMON-LISP"))
+                    ;
+                else if (string_equal(native_pointer(pkg_name), "COMMON-LISP-USER")) {
+                    fputs("CL-USER::", stdout);
+                }
+                else if (string_equal(native_pointer(pkg_name), "KEYWORD")) {
+                    putchar(':');
+                } else {
+                    print_string(native_pointer(pkg_name));
+                    fputs("::", stdout);
+                }
+            }
+            print_string(native_pointer(symbol->name));
+        } else if (widetag_of(*object) == SIMPLE_BASE_STRING_WIDETAG
 #ifdef SIMPLE_CHARACTER_STRING_WIDETAG
-               || widetag_of(*object) == SIMPLE_CHARACTER_STRING_WIDETAG
+                   || widetag_of(*object) == SIMPLE_CHARACTER_STRING_WIDETAG
 #endif
-               ) {
-      putchar('"');
-      print_string(object);
-      putchar('"');
+            ) {
+            putchar('"');
+            print_string(object);
+            putchar('"');
+        } else {
+            printf("<??? type %d>", (int) widetag_of(*object));
+        }
     } else {
-      printf("<??? type %d>", (int) widetag_of(*object));
+        printf("<??? lowtag %d>", (int) lowtag_of(name));
     }
-  } else {
-    printf("<??? lowtag %d>", (int) lowtag_of(name));
-  }
 }
 
 static void
 print_entry_points (struct code *code)
 {
-    lispobj function = code->entry_points;
-
-    while (function != NIL) {
-        struct simple_fun *header = (struct simple_fun *) native_pointer(function);
-        print_entry_name(header->name);
-
-        function = header->next;
-        if (function != NIL)
-            printf (", ");
-    }
+    int n_funs = code_n_funs(code);
+    for_each_simple_fun(index, fun, code, 0, {
+        if (widetag_of(fun->header) != SIMPLE_FUN_WIDETAG) {
+            printf("%p: bogus function entry", fun);
+            return;
+        }
+        print_entry_name(fun->name);
+        if ((index + 1) < n_funs) printf (", ");
+    });
 }
 
 
@@ -245,15 +259,15 @@ code_pointer(lispobj object)
     lispobj *headerp, header;
     int type, len;
 
-    headerp = (lispobj *) native_pointer(object);
+    headerp = native_pointer(object);
     header = *headerp;
     type = widetag_of(header);
 
     switch (type) {
         case CODE_HEADER_WIDETAG:
             break;
-        case RETURN_PC_HEADER_WIDETAG:
-        case SIMPLE_FUN_HEADER_WIDETAG:
+        case RETURN_PC_WIDETAG:
+        case SIMPLE_FUN_WIDETAG:
             len = HEADER_LENGTH(header);
             if (len == 0)
                 headerp = NULL;
@@ -335,7 +349,7 @@ previous_info(struct call_info *info)
     lispobj lra;
 
     if (!cs_valid_pointer_p(info->frame)) {
-        printf("Bogus callee value (0x%08lx).\n", (uword_t)info->frame);
+        printf("Bogus callee value (0x%lx).\n", (long)info->frame);
         return 0;
     }
 
@@ -361,7 +375,7 @@ previous_info(struct call_info *info)
             }
         }
     } else if (fixnump(lra)) {
-        info->code = native_pointer(this_frame->code);
+        info->code = (struct code*)native_pointer(this_frame->code);
         info->pc = (uword_t)(info->code + lra);
         info->lra = NIL;
     } else {
@@ -400,7 +414,7 @@ lisp_backtrace(int nframes)
             else
                 print_entry_points((struct code *)info.code);
 
-            printf(" %p", (uword_t) info.code | OTHER_POINTER_LOWTAG);
+            printf(" %p", (void*)((uword_t) info.code | OTHER_POINTER_LOWTAG));
         }
         else
             printf("CODE = ???");
@@ -408,12 +422,12 @@ lisp_backtrace(int nframes)
                info.frame);
 
         if (info.lra != NIL)
-            printf(" LRA = %p", info.lra);
+            printf(" LRA = %p", (void*)info.lra);
         else
             printf(" <no LRA>");
 
         if (info.pc)
-            printf(" pc = %p", info.pc);
+            printf(" pc = %p", (void*)(long)info.pc);
         putchar('\n');
 
     } while (i++ < nframes && previous_info(&info));
@@ -424,8 +438,8 @@ lisp_backtrace(int nframes)
 static int
 altstack_pointer_p (void *p) {
 #ifndef LISP_FEATURE_WIN32
-    void* stack_start = ((void *)arch_os_get_current_thread()) + dynamic_values_bytes;
-    void* stack_end = stack_start + 32*SIGSTKSZ;
+    void* stack_start = ((char*)arch_os_get_current_thread()) + dynamic_values_bytes;
+    void* stack_end = (char*)stack_start + 32*SIGSTKSZ;
 
     return (p > stack_start && p <= stack_end);
 #else
@@ -437,15 +451,26 @@ altstack_pointer_p (void *p) {
 static int
 stack_pointer_p (void *p)
 {
-  /* we are using sizeof(long) here, because that is the right value on both
-   * x86 and x86-64.  (But note that false positives would not cause much harm
-   * given the heuristical nature of x86_call_context.) */
-  uword_t stack_alignment = sizeof(void*);
+    /* we are using sizeof(long) here, because that is the right value on both
+     * x86 and x86-64.  (But note that false positives would not cause much harm
+     * given the heuristical nature of x86_call_context.) */
+    uword_t stack_alignment = sizeof(void*);
+    void *stack_start;
+    struct thread *thread = arch_os_get_current_thread();
 
-  return (altstack_pointer_p(p)
-          || (p < (void *) arch_os_get_current_thread()->control_stack_end
-              && (p > (void *) &p || altstack_pointer_p(&p))
-              && (((uword_t) p) & (stack_alignment-1)) == 0));
+    if (altstack_pointer_p(p))
+        return 1;
+
+    if (altstack_pointer_p(&p)) {
+        stack_start = (void *) thread->control_stack_start;
+    } else {
+        /* Use the current frame address, since there should be no
+         * relevant frames below. */
+        stack_start = &p;
+    }
+    return p >= stack_start
+        && p < (void *) thread->control_stack_end
+        && (((uword_t) p) & (stack_alignment-1)) == 0;
 }
 
 static int
