@@ -175,8 +175,8 @@ lose:
  * def to something like ifdef nommapoverlap */
 /* currently hpux only */
 #ifdef LISP_FEATURE_HPUX
-os_vm_address_t copy_core_bytes(int fd, os_vm_offset_t offset,
-                                os_vm_address_t addr, int len)
+void copy_core_bytes(int fd, os_vm_offset_t offset,
+                     os_vm_address_t addr, int len)
 {
   unsigned char buf[4096];
   int c,x;
@@ -202,14 +202,13 @@ os_vm_address_t copy_core_bytes(int fd, os_vm_offset_t offset,
     memcpy(addr+x, buf, 4096);
   }
   os_flush_icache(addr, len);
-  return addr;
 }
 #endif
 
 #ifdef LISP_FEATURE_SB_CORE_COMPRESSION
 # define ZLIB_BUFFER_SIZE (1u<<16)
-os_vm_address_t inflate_core_bytes(int fd, os_vm_offset_t offset,
-                                   os_vm_address_t addr, int len)
+void inflate_core_bytes(int fd, os_vm_offset_t offset,
+                        os_vm_address_t addr, int len)
 {
     z_stream stream;
     unsigned char buf[ZLIB_BUFFER_SIZE];
@@ -270,16 +269,20 @@ os_vm_address_t inflate_core_bytes(int fd, os_vm_offset_t offset,
     }
 
     inflateEnd(&stream);
-    return addr;
 }
 # undef ZLIB_BUFFER_SIZE
 #endif
 
 int merge_core_pages = -1;
 
+#ifdef LISP_FEATURE_LINUX
+os_vm_address_t anon_dynamic_space_start;
+#endif
+
 static void
 process_directory(int fd, lispobj *ptr, int count, os_vm_offset_t file_offset)
 {
+    extern void immobile_space_coreparse(uword_t,uword_t);
     struct ndir_entry *entry;
     int compressed;
 
@@ -300,28 +303,21 @@ process_directory(int fd, lispobj *ptr, int count, os_vm_offset_t file_offset)
         lispobj *free_pointer = (lispobj *) addr + entry->nwords;
         uword_t len = os_vm_page_size * entry->page_count;
         if (len != 0) {
-            os_vm_address_t real_addr;
             FSHOW((stderr, "/mapping %ld(0x%lx) bytes at 0x%lx\n",
                    len, len, (uword_t)addr));
             if (compressed) {
 #ifdef LISP_FEATURE_SB_CORE_COMPRESSION
-                real_addr = inflate_core_bytes(fd, offset + file_offset, addr, len);
+                inflate_core_bytes(fd, offset + file_offset, addr, len);
 #else
                 lose("This runtime was not built with zlib-compressed core support... aborting\n");
 #endif
             } else {
 #ifdef LISP_FEATURE_HPUX
-                real_addr = copy_core_bytes(fd, offset + file_offset, addr, len);
+                copy_core_bytes(fd, offset + file_offset, addr, len);
 #else
-                real_addr = os_map(fd, offset + file_offset, addr, len);
+                os_map(fd, offset + file_offset, addr, len);
 #endif
             }
-            if (real_addr != addr) {
-                lose("file mapped in wrong place! "
-                     "(0x%08x != 0x%08lx)\n",
-                     real_addr,
-                     addr);
-            }
         }
 
 #ifdef MADV_MERGEABLE
@@ -337,9 +333,9 @@ process_directory(int fd, lispobj *ptr, int count, os_vm_offset_t file_offset)
         case DYNAMIC_CORE_SPACE_ID:
             if (len > dynamic_space_size) {
                 fprintf(stderr,
-                        "dynamic space too small for core: %ldKiB required, %ldKiB available.\n",
-                        len >> 10,
-                        (uword_t)dynamic_space_size >> 10);
+                        "dynamic space too small for core: %luKiB required, %luKiB available.\n",
+                        (unsigned long)len >> 10,
+                        (unsigned long)dynamic_space_size >> 10);
                 exit(1);
             }
 #ifdef LISP_FEATURE_GENCGC
@@ -368,6 +364,13 @@ process_directory(int fd, lispobj *ptr, int count, os_vm_offset_t file_offset)
              * space 0. (We checked above that for GENCGC,
              * addr==DYNAMIC_SPACE_START.) */
             current_dynamic_space = (lispobj *)addr;
+#ifdef LISP_FEATURE_LINUX
+            anon_dynamic_space_start = addr + len;
+            // This assertion is here because of the test in zero_pages_with_mmap()
+            // which trusts that if addr > anon_dynamic_space_start
+            // then addr did not come from any file mapping.
+            gc_assert((lispobj)anon_dynamic_space_start > STATIC_SPACE_END);
+#endif
             break;
         case STATIC_CORE_SPACE_ID:
             if (addr != (os_vm_address_t)STATIC_SPACE_START) {
@@ -383,6 +386,27 @@ process_directory(int fd, lispobj *ptr, int count, os_vm_offset_t file_offset)
                 lose("core/runtime address mismatch: READ_ONLY_SPACE_START\n");
             }
             break;
+#ifdef LISP_FEATURE_IMMOBILE_SPACE
+         // Immobile space is subdivided into fixed-size and variable-size.
+         // There is no margin between the two, though for efficiency
+         // they are written separately to eliminate waste in the core file.
+        case IMMOBILE_FIXEDOBJ_CORE_SPACE_ID:
+            if (addr != (os_vm_address_t)IMMOBILE_SPACE_START) {
+                fprintf(stderr, "in core: %p - in runtime: %p\n",
+                        (void*)addr, (void*)IMMOBILE_SPACE_START);
+                lose("core/runtime address mismatch: IMMOBILE_SPACE_START\n");
+            }
+            immobile_space_coreparse(IMMOBILE_SPACE_START, len);
+            break;
+        case IMMOBILE_VARYOBJ_CORE_SPACE_ID:
+            if (addr != (os_vm_address_t)IMMOBILE_VARYOBJ_SUBSPACE_START) {
+                fprintf(stderr, "in core: %p - in runtime: %p\n",
+                        (void*)addr, (void*)IMMOBILE_VARYOBJ_SUBSPACE_START);
+                lose("core/runtime address mismatch: IMMOBILE_VARYOBJ_SUBSPACE_START\n");
+            }
+            immobile_space_coreparse(IMMOBILE_VARYOBJ_SUBSPACE_START, len);
+            break;
+#endif
         default:
             lose("unknown space ID %ld addr %p\n", id, addr);
         }
@@ -414,7 +438,7 @@ load_core_file(char *file, os_vm_offset_t file_offset)
     header = calloc(os_vm_page_size, 1);
 
     count = read(fd, header, os_vm_page_size);
-    if (count < os_vm_page_size) {
+    if (count < (ssize_t) os_vm_page_size) {
         lose("premature end of core file\n");
     }
     SHOW("successfully read first page of core");
@@ -516,7 +540,7 @@ load_core_file(char *file, os_vm_offset_t file_offset)
                      * The low bits of each word are allocation flags.
                      */
                     if ((word=data[i])) {
-                        page_table[offset].scan_start_offset = word & ~0x03;
+                        set_page_scan_start_offset(offset, word & ~0x03);
                         page_table[offset].allocated = word & 0x03;
                     }
                     i++;
@@ -537,6 +561,40 @@ load_core_file(char *file, os_vm_offset_t file_offset)
     }
     SHOW("about to free(header)");
     free(header);
+    close(fd);
     SHOW("returning from load_core_file(..)");
     return initial_function;
 }
+
+#include "genesis/hash-table.h"
+#include "genesis/vector.h"
+os_vm_address_t get_asm_routine_by_name(const char* name)
+{
+    lispobj routines = SYMBOL(ASSEMBLER_ROUTINES)->value;
+    if (lowtag_of(routines) == INSTANCE_POINTER_LOWTAG) {
+        struct hash_table* ht = (struct hash_table*)native_pointer(routines);
+        struct vector* table = (struct vector*)native_pointer(ht->table);
+        lispobj sym;
+        int i;
+        for (i=2 ; i < fixnum_value(table->length) ; i += 2) {
+          sym = table->data[i];
+          if (lowtag_of(sym) == OTHER_POINTER_LOWTAG
+              && widetag_of(SYMBOL(sym)->header) == SYMBOL_WIDETAG
+              && !strcmp(name,
+                         (char*)((struct vector*)
+                                 native_pointer(SYMBOL(sym)->name))->data))
+              return (os_vm_address_t)fixnum_value(table->data[i+1]);
+        }
+        // Something is wrong if we have a hashtable but find nothing.
+        fprintf(stderr, "WARNING: get_asm_routine_by_name(%s) failed\n",
+                name);
+    }
+    return NULL;
+}
+
+void asm_routine_poke(const char* routine, int offset, char byte)
+{
+    char *address = (char *)get_asm_routine_by_name(routine);
+    if (address)
+        address[offset] = byte;
+}
