@@ -24,6 +24,7 @@
 #include "sbcl.h"
 #include "print.h"
 #include "runtime.h"
+#include "gc-internal.h"
 #include <stdarg.h>
 #include "thread.h"              /* genesis/primitive-objects.h needs this */
 #include <errno.h>
@@ -461,26 +462,16 @@ static void brief_struct(lispobj obj)
 }
 
 #include "genesis/layout.h"
-static boolean untagged_slot_p(struct layout * layout,
+static boolean tagged_slot_p(struct layout * layout,
                                int slot_index)
 {
-#ifdef LISP_FEATURE_INTERLEAVED_RAW_SLOTS
-  extern boolean positive_bignum_logbitp(int,struct bignum*);
-  lispobj bitmap = layout->untagged_bitmap;
+  lispobj bitmap = layout->bitmap;
+  sword_t fixnum = (sword_t)bitmap >> N_FIXNUM_TAG_BITS; // optimistically
   return fixnump(bitmap)
-         ? (fixnum_value(bitmap) >> slot_index) & 1
+         ? bitmap == make_fixnum(-1) ||
+            (slot_index < N_WORD_BITS && ((fixnum >> slot_index) & 1) != 0)
          : positive_bignum_logbitp(slot_index,
                                    (struct bignum*)native_pointer(bitmap));
-#else
-  // STANDARD-OBJECT has layout-length = number of defined slots, but
-  // the primitive object always occupies 4 physical words. The guard on
-  // n_untagged_slots ensures that for classes with no slots, we don't
-  // wrongly show all words of the primitive object as untagged,
-  // because the second half of the expression reduces to slot_index>=1.
-  return layout->n_untagged_slots > 0
-         && slot_index >= (fixnum_value(layout->length)|1)
-                          - fixnum_value(layout->n_untagged_slots);
-#endif
 }
 
 static void print_struct(lispobj obj)
@@ -496,11 +487,12 @@ static void print_struct(lispobj obj)
         struct layout * layout = (struct layout*)native_pointer(layout_obj);
         for (i=INSTANCE_DATA_START; i<instance_length(instance->header); i++) {
             sprintf(buffer, "slot %d: ", i);
-            if (layout==NULL || untagged_slot_p(layout, i)) {
+            if (layout != NULL && tagged_slot_p(layout, i)) {
+                print_obj(buffer, instance->slots[i]);
+            } else {
                 newline(NULL);
                 printf("\n\t    %s0x%"OBJ_FMTX" [raw]", buffer, instance->slots[i]);
-            } else
-                print_obj(buffer, instance->slots[i]);
+            }
         }
     }
 }
@@ -540,11 +532,12 @@ void show_lstring(struct vector * string, int quotes, FILE *s)
 
 static void brief_otherptr(lispobj obj)
 {
+    extern void safely_show_lstring(struct vector*, int, FILE*);
     lispobj *ptr, header;
     int type;
     struct symbol *symbol;
 
-    ptr = (lispobj *) native_pointer(obj);
+    ptr = native_pointer(obj);
 
     if (!is_valid_lisp_addr((os_vm_address_t)obj)) {
             printf("(invalid address)");
@@ -554,7 +547,7 @@ static void brief_otherptr(lispobj obj)
     header = *ptr;
     type = widetag_of(header);
     switch (type) {
-        case SYMBOL_HEADER_WIDETAG:
+        case SYMBOL_WIDETAG:
             symbol = (struct symbol *)ptr;
             if (symbol->package == NIL)
                 printf("#:");
@@ -572,6 +565,16 @@ static void brief_otherptr(lispobj obj)
         default:
             printf("#<ptr to ");
             brief_otherimm(header);
+            if (type == FDEFN_WIDETAG) {  // Try to print name, if a symbol
+                // FIXME: more address validity checks perhaps?
+                lispobj name = ((struct fdefn*)ptr)->name;
+                if (lowtag_of(name) == OTHER_POINTER_LOWTAG
+                    && widetag_of(*native_pointer(name)) == SYMBOL_WIDETAG) {
+                  printf(" for ");
+                  struct vector* str = symbol_name(native_pointer(name));
+                  safely_show_lstring(str, 0, stdout);
+                }
+            }
             putchar('>');
     }
 }
@@ -598,15 +601,43 @@ static char *symbol_slots[] = {"value: ", "hash: ",
     NULL};
 static char *ratio_slots[] = {"numer: ", "denom: ", NULL};
 static char *complex_slots[] = {"real: ", "imag: ", NULL};
-static char *code_slots[] = {"words: ", "entry: ", "debug: ", NULL};
-static char *fn_slots[] = {
-    "self: ", "next: ", "name: ", "arglist: ", "type: ", "info: ", NULL};
+static char *code_slots[] = {"bytes: ", "debug: ",
+#ifndef LISP_FEATURE_64_BIT
+                             "n_entries: ",
+#endif
+                             NULL};
+static char *simple_fun_slots[] = {
+    "self: ", "name: ", "arglist: ", "type: ", "info: ", NULL};
 static char *closure_slots[] = {"fn: ", NULL};
-static char *funcallable_instance_slots[] = {"raw_fn: ", "fn: ", "layout: ", NULL};
+static char *funcallable_instance_slots[] = {"raw_fn: ", "fn: ",
+#ifndef LISP_FEATURE_COMPACT_INSTANCE_HEADER
+                                             "layout: ",
+#endif
+                                             NULL};
 static char *weak_pointer_slots[] = {"value: ", NULL};
 static char *fdefn_slots[] = {"name: ", "function: ", "raw_addr: ", NULL};
 static char *value_cell_slots[] = {"value: ", NULL};
 
+static lispobj symbol_function(lispobj* symbol)
+{
+    lispobj info = ((struct symbol*)symbol)->info;
+    if (lowtag_of(info) == LIST_POINTER_LOWTAG)
+        info = CONS(info)->cdr;
+    if (lowtag_of(info) == OTHER_POINTER_LOWTAG) {
+        struct vector* v = (struct vector*)native_pointer(info);
+        int len = fixnum_value(v->length);
+        if (len != 0) {
+            lispobj elt = v->data[0];  // Just like INFO-VECTOR-FDEFN
+            if (fixnump(elt) && (fixnum_value(elt) & 07777) >= 07701) {
+                lispobj fdefn = v->data[len-1];
+                if (lowtag_of(fdefn) == OTHER_POINTER_LOWTAG)
+                    return FDEFN(fdefn)->fun;
+            }
+        }
+    }
+    return NIL;
+}
+
 static void print_otherptr(lispobj obj)
 {
     if (!is_valid_lisp_addr((os_vm_address_t)obj)) {
@@ -624,7 +655,7 @@ static void print_otherptr(lispobj obj)
         int count, type, index;
         char buffer[16];
 
-        ptr = (lispobj*) native_pointer(obj);
+        ptr = native_pointer(obj);
         if (ptr == NULL) {
                 printf(" (NULL Pointer)");
                 return;
@@ -667,10 +698,12 @@ static void print_otherptr(lispobj obj)
                 print_slots(complex_slots, count, ptr);
                 break;
 
-            case SYMBOL_HEADER_WIDETAG:
+            case SYMBOL_WIDETAG:
                 // Only 1 byte of a symbol header conveys its size.
                 // The other bytes may be freely used by the backend.
                 print_slots(symbol_slots, count & 0xFF, ptr);
+                if (symbol_function(ptr-1) != NIL)
+                    print_obj("fun: ", symbol_function(ptr-1));
                 break;
 
 #if N_WORD_BITS == 32
@@ -746,8 +779,9 @@ static void print_otherptr(lispobj obj)
                 break;
 
             // FIXME: This case looks unreachable. print_struct() does it
-            case INSTANCE_HEADER_WIDETAG:
+            case INSTANCE_WIDETAG:
                 NEWLINE_OR_RETURN;
+                count &= SHORT_HEADER_MAX_WORDS;
                 printf("length = %ld", (long) count);
                 index = 0;
                 while (count-- > 0) {
@@ -757,28 +791,41 @@ static void print_otherptr(lispobj obj)
                 break;
 
             case CODE_HEADER_WIDETAG:
+                count &= SHORT_HEADER_MAX_WORDS;
+                // ptr was already bumped up
+                for_each_simple_fun(fun_index, fun, (struct code*)(ptr-1), 0, {
+                    sprintf(buffer, "f[%d]: ", fun_index);
+                    print_obj(buffer, make_lispobj(fun,FUN_POINTER_LOWTAG));
+                });
                 print_slots(code_slots, count-1, ptr);
                 break;
 
-            case SIMPLE_FUN_HEADER_WIDETAG:
-                print_slots(fn_slots, 6, ptr);
+            case SIMPLE_FUN_WIDETAG:
+                print_obj("code: ",
+                          make_lispobj(native_pointer((lispobj)(ptr-1))
+                                       -(HeaderValue(header)&0xFFFF),
+                                       OTHER_POINTER_LOWTAG));
+                print_slots(simple_fun_slots,
+                            sizeof simple_fun_slots/sizeof(char*)-1, ptr);
                 break;
 
 #if !defined(LISP_FEATURE_X86) && !defined(LISP_FEATURE_X86_64)
-            case RETURN_PC_HEADER_WIDETAG:
+            case RETURN_PC_WIDETAG:
                 print_obj("code: ", obj - (count * 4));
                 break;
 #endif
 
-            case CLOSURE_HEADER_WIDETAG:
-                print_slots(closure_slots, count, ptr);
+            case CLOSURE_WIDETAG:
+                print_slots(closure_slots,
+                            count & SHORT_HEADER_MAX_WORDS, ptr);
                 break;
 
-            case FUNCALLABLE_INSTANCE_HEADER_WIDETAG:
-                print_slots(funcallable_instance_slots, count, ptr);
+            case FUNCALLABLE_INSTANCE_WIDETAG:
+                print_slots(funcallable_instance_slots,
+                            count & SHORT_HEADER_MAX_WORDS, ptr);
                 break;
 
-            case VALUE_CELL_HEADER_WIDETAG:
+            case VALUE_CELL_WIDETAG:
                 print_slots(value_cell_slots, 1, ptr);
                 break;
 
@@ -802,7 +849,12 @@ static void print_otherptr(lispobj obj)
                 break;
 
             case FDEFN_WIDETAG:
-                print_slots(fdefn_slots, count, ptr);
+#ifdef LISP_FEATURE_IMMOBILE_CODE
+                print_slots(fdefn_slots, 2, ptr);
+                print_obj("entry: ", fdefn_raw_referent((struct fdefn*)(ptr-1)));
+#else
+                print_slots(fdefn_slots, count & SHORT_HEADER_MAX_WORDS, ptr);
+#endif
                 break;
 
             default:
@@ -911,6 +963,63 @@ void brief_print(lispobj obj)
     putchar('\n');
 }
 
+// The following accessors, which take a valid native pointer as input
+// and return a Lisp string, are designed to be foolproof during GC,
+// hence all the forwarding checks.
+
+#include "forwarding-ptr.h"
+#include "genesis/classoid.h"
+struct vector * symbol_name(lispobj * sym)
+{
+  if (forwarding_pointer_p(sym))
+    sym = native_pointer(forwarding_pointer_value(sym));
+  if (lowtag_of(((struct symbol*)sym)->name) != OTHER_POINTER_LOWTAG)
+      return NULL;
+  lispobj * name = native_pointer(((struct symbol*)sym)->name);
+  if (forwarding_pointer_p(name))
+      name = native_pointer(forwarding_pointer_value(name));
+  return (struct vector*)name;
+}
+struct vector * classoid_name(lispobj * classoid)
+{
+  if (forwarding_pointer_p(classoid))
+      classoid = native_pointer(forwarding_pointer_value(classoid));
+  lispobj sym = ((struct classoid*)classoid)->name;
+  return lowtag_of(sym) != OTHER_POINTER_LOWTAG ? NULL
+    : symbol_name(native_pointer(sym));
+}
+struct vector * layout_classoid_name(lispobj * layout)
+{
+  if (forwarding_pointer_p(layout))
+      layout = native_pointer(forwarding_pointer_value(layout));
+  lispobj classoid = ((struct layout*)layout)->classoid;
+  return lowtag_of(classoid) != INSTANCE_POINTER_LOWTAG ? NULL
+    : classoid_name(native_pointer(classoid));
+}
+struct vector * instance_classoid_name(lispobj * instance)
+{
+  if (forwarding_pointer_p(instance))
+      instance = native_pointer(forwarding_pointer_value(instance));
+  lispobj layout = instance_layout(instance);
+  return lowtag_of(layout) != INSTANCE_POINTER_LOWTAG ? NULL
+    : layout_classoid_name(native_pointer(layout));
+}
+void safely_show_lstring(struct vector * string, int quotes, FILE *s)
+{
+  extern void show_lstring(struct vector*, int, FILE*);
+  if (forwarding_pointer_p((lispobj*)string))
+      string = (struct vector*)forwarding_pointer_value((lispobj*)string);
+  if (
+#ifdef SIMPLE_CHARACTER_STRING_WIDETAG
+      widetag_of(string->header) == SIMPLE_CHARACTER_STRING_WIDETAG ||
+#endif
+      widetag_of(string->header) == SIMPLE_BASE_STRING_WIDETAG)
+    show_lstring(string, quotes, s);
+  else {
+    fprintf(s, "#<[widetag=%02X]>", widetag_of(string->header));
+  }
+}
+
 #else
 
 void
